@@ -12,10 +12,10 @@
  * The Original Code is the Sablotron XSLT Processor.
  * 
  * The Initial Developer of the Original Code is Ginger Alliance Ltd.
- * Portions created by Ginger Alliance are Copyright (C) 2000 Ginger
- * Alliance Ltd. All Rights Reserved.
+ * Portions created by Ginger Alliance are Copyright (C) 2000-2003
+ * Ginger Alliance Ltd. All Rights Reserved.
  * 
- * Contributor(s): Christian Lefebvre
+ * Contributor(s): Christian Lefebvre, Stefan Behnel
  * 
  * Alternatively, the contents of this file may be used under the
  * terms of the GNU General Public License Version 2 or later (the
@@ -40,11 +40,9 @@
 defines
 *****************************************************************/
 
-#define SAB_VERSION "0.90"
-#define SAB_DATE "Mar 8, 2002"
 #define BARE_SIGN '@'
 
-#define MAX_BARE        3
+#define MAX_BARE        64
 #define MAX_EQ          16
 #define MAX_ARG_OR_PAR  MAX_EQ
 
@@ -61,7 +59,11 @@ includes
 #include <time.h>      // needed by <sys/timeb.h> for definition of time_t
 
 #if defined(HAVE_CONFIG_H)
-#include <autocfg.h>
+#include <config.h>
+#endif
+
+#ifdef SABLOT_DEBUGGER
+#include <sabdbg.h>
 #endif
 
 #if defined(HAVE_SYS_TIMEB_H) || defined(WIN32)
@@ -80,13 +82,14 @@ includes
 vars to hold the results of switch parsing
 *****************************************************************/
 
-int 
-    numberTimes = 0,
-    numberMode = 0,
-    numberMeasure = 0;
+int batchMode = 0,
+  numberTimes = 0,
+  numberMode = 0,
+  numberMeasure = 0,
+  commandFlags = 0;
 
-char *stringBase = NULL,
-    *stringLog = NULL;
+char * stringBase = NULL,
+    * stringLog = NULL;
 
 /*****************************************************************
 switch definitions
@@ -94,18 +97,20 @@ switch definitions
 
 typedef enum 
 {
-    ID_NONE,
-    ID_SPS, ID_SPF, ID_SPS_ON_FILES, ID_BASE, ID_TIMES,
-    ID_LOG, ID_HELP, ID_DEBUG, ID_DBG_HELP,
-    ID_VERSION, ID_MEASURE
+  ID_NONE,
+  ID_SPS, ID_SPF, ID_SPS_ON_FILES, ID_BASE, ID_TIMES,
+  ID_LOG, ID_HELP, ID_DEBUG, ID_DEBUGGER, ID_DBG_HELP, ID_FLAGS, 
+  ID_VERSION, ID_MEASURE, ID_BATCH_XML, ID_BATCH_XSL,
+  ID_CHAIN_XSL
 } SwitchId;
 
 struct SwitchData
 {
-    SwitchId id;
-    char *longName, shortName;
-    char type;  // Integer, String, Ordinal
-    void *destination;
+  SwitchId id;
+  const char *longName;
+  char shortName;
+  char type;  // Integer, String, Ordinal
+  void *destination;
 };
 
 struct SwitchData switches[] =
@@ -118,9 +123,14 @@ struct SwitchData switches[] =
     {ID_LOG, "log-file", 'L', 'S', &stringLog},         
     {ID_HELP, "help", '?', 'O', &numberMode},            
     {ID_DEBUG, "debug", 0, 'O', &numberMode},
+    {ID_DEBUGGER, "debugger", 0, 'O', &numberMode},
     {ID_DBG_HELP, "debug-options", 0, 'O', &numberMode},
+    {ID_FLAGS, "flags", 'f', 'N', &commandFlags},
     {ID_VERSION, "version", 'v', 'O', &numberMode},
     {ID_MEASURE, "measure", 'm', 'O', &numberMeasure},
+    {ID_BATCH_XML, "batch-xml", 'x', 'O', &batchMode},
+    {ID_BATCH_XSL, "batch-xsl", 's', 'O', &batchMode},
+    {ID_CHAIN_XSL, "chain-xsl", 'c', 'O', &batchMode},
     {ID_NONE, NULL, 0, 0, NULL }
 };
 
@@ -129,34 +139,47 @@ texts
 *****************************************************************/
 
 char usage[] = 
-    "\nUsage: \n\tsabcmd [options] <stylesheet> [<input> [<output>]] [assignments]\n"\
-    "Options:\n"\
-    "\t--base=NAME, -b\t\tset the hard base URI to NAME\n"\
-    "\t--debug-options\t\tdisplay information on debugging options\n"\
-    "\t--help, -?\t\tdisplay this help message\n"\
-    "\t--log-file=NAME, -L\tset the log file, turn logging on\n"\
-    "\t--measure, -m\t\tmeasure the processing time\n"\
-    "\t--version, -v\t\tdisplay version information\n"\
-    "Defaults:\n\t<input> = stdin, <output> = stdout\n"\
-    "\tlogging off, no hard base URI\n"\
-    "Notes:\n"\
-    "\t- assignments define named buffers (name=value)\n"\
-    "\t  and top-level params ($name=value).\n"\
-    "\t- to specify value in an option, use -b=NAME or -b NAME\n"\
-    "\t  (correspondingly for the equivalent long options)";
+"\nUsage for single XSLT:\n"\
+"\tsabcmd [options] <stylesheet> [<input> [<output>]] [assignments]\n"\
+"\nUsage in batch processing:\n"\
+"    chain multiple stylesheets, starting with a single input file:\n"\
+"    \tsabcmd [options] --chain-xsl <input> <output> [<stylesheet>]+ [assignments]\n"\
+"    run multiple stylesheets on a single input file:\n"\
+"    \tsabcmd [options] --batch-xml <input> [<stylesheet> <output>]+ [assignments]\n"\
+"    run a single stylesheet on multiple input files:\n"\
+"    \tsabcmd [options] --batch-xsl <stylesheet> [<input> <output>]+ [assignments]\n"\
+"Options:\n"\
+"\t--chain-xsl, -c      single input file, multiple chained stylesheets\n"\
+"\t--batch-xml, -x      single input file, multiple stylesheets\n"\
+"\t--batch-xsl, -s      multiple input files, single stylesheet\n"\
+"\t--base=NAME, -b      set the hard base URI to NAME\n"\
+"\t--debug-options      display information on debugging options\n"\
+"\t--help, -?           display this help message\n"\
+"\t--log-file=NAME, -L  set the log file, turn logging on\n"\
+"\t--measure, -m        measure the processing time\n"\
+"\t--version, -v        display version information\n"\
+"Defaults:\n\t<input> = stdin, <output> = stdout\n"\
+"\tlogging off, no hard base URI\n"\
+"Notes:\n"\
+"\t- assignments define named buffers (name=value)\n"\
+"\t  and top-level params ($name=value).\n"\
+"\t- to specify value in an option, use -b=NAME or -b NAME\n"\
+"\t  (correspondingly for the equivalent long options)";
 
 char askhelp[] = "Type sabcmd --help to display a help message.\n";
 char msgConflict[] = "conflict with preceding switches: ";
 char version_txt[] = "\nsabcmd "SAB_VERSION" ("SAB_DATE")\n"\
-    "copyright (C) 2000 - 2002 Ginger Alliance (www.gingerall.com)\n";
+    "copyright (C) 2000 - 2003 Ginger Alliance (www.gingerall.com)\n";
 char dbg_usage[] = "\nDebugging options:\n"\
     "\t--debug\t\t\tdisplay results of the command line parse\n"\
+    "\t--debugger\t\trun the xslt debugger\n"\
     "\t--times=COUNT, -t\trun sabcmd the specified number of times\n"\
+    "\t--flags=FLAGS, -f\tset processor flags for the processing\n"\
     "\t--use-SPF, -F\t\tuse SablotProcessFiles().\n"\
     "\t--use-SPS, -S\t\tuse SablotProcessStrings(). Give 2 args\n"\
-    "\t\t\t\t(stylesheet, input). Precede each by @.\n"
+    "\t\t\t\t(stylesheet, input). Precede each by @.\n" \
     "\t--use-SPS-on-files\tuse SablotProcessStrings() on the contents\n"\
-    "\t\t\t\tof the given files.";
+    "\t\t\t\tof the given files.\n";
 /* removed the GPL notice: 
 char startup_licence[]="sabcmd and Sablotron come with ABSOLUTELY NO WARRANTY. "\
     "They are free software,\n"\
@@ -171,7 +194,7 @@ char startup_licence[] = "The Sablotron XSLT Processor comes with NO WARRANTY.\n
 handy functions
 *****************************************************************/
 
-void saberr(char *msg1, char *msg2)
+void saberr(const char *msg1, const char *msg2)
 {
     fprintf(stderr, "Error: %s", msg1);
     if (msg2) fputs(msg2, stderr);
@@ -179,14 +202,14 @@ void saberr(char *msg1, char *msg2)
     exit(EXIT_FAILURE);
 }
 
-void saberrn(char *msg, int num)
+void saberrn(const char *msg, int num)
 {
     char buf[10];
     sprintf(buf,"%d",num);
     saberr(msg, buf);
 }
 
-int chrpos(char *text, char c)
+int chrpos(const char *text, char c)
 {
     char *p = strchr(text, c);
     return p ? (int)(p - text) : -1;
@@ -215,14 +238,15 @@ void copyAssignment(char **array, int index, char *text, int split)
 
 // globals to hold the names and assignments
 
-char *arrayBare[MAX_BARE], *arrayEq[MAX_EQ]; 
+char* arrayBare[MAX_BARE];
+char* arrayEq[MAX_EQ]; 
 int indexBare = 0, indexEq = 0;
 
 /*****************************************************************
 findSwitch
 *****************************************************************/
 
-int max_strncmp(char *s1, char *s2, int len1)
+int max_strncmp(const char *s1, const char *s2, int len1)
 {
     int len2 = strlen(s2);
     return strncmp(s1, s2, len1 >= len2 ? len1 : len2);
@@ -431,7 +455,7 @@ void full_version_txt()
 }
 
 
-char *readToBuffer(char *filename)
+char* readToBuffer(char *filename)
 {
     long length;
     int count;
@@ -450,34 +474,223 @@ char *readToBuffer(char *filename)
 }
 
 
+int runSingleXSLT(const char **arrayParams, const char **arrayArgs, 
+		  char **resultArg)
+{
+   void *theProcessor;
+   char *resultURI = NULL;
+   int ecode;
+   void *situa = NULL;
+   
+   if (ecode = SablotCreateSituation(&situa))
+     return ecode;
+
+   if (commandFlags)
+     SablotSetOptions(situa, commandFlags);
+
+   if (ecode = SablotCreateProcessorForSituation(situa, &theProcessor))
+      return ecode;
+
+   if (stringLog)
+      SablotSetLog(theProcessor, stringLog, 0);
+
+   if (stringBase)
+      SablotSetBase(theProcessor, stringBase);
+
+   if (indexBare <= 2)
+      resultURI = (char*)"file:///__stdout";
+   else
+      resultURI = arrayBare[2];
+
+   ecode = SablotRunProcessor(theProcessor,
+	arrayBare[0],
+	indexBare <= 1  ? (char*)"file:///__stdin" : arrayBare[1],
+	resultURI,
+	arrayParams, arrayArgs);
+
+   if (ecode == 0)
+      SablotGetResultArg(theProcessor, resultURI, resultArg);
+                
+   if (theProcessor)
+      SablotDestroyProcessor(theProcessor);
+
+   if (situa)
+     SablotDestroySituation(situa);
+
+   return ecode;
+}
+
+int runBatchXSLT(const char **arrayParams, const char **arrayArgs, 
+		 char **resultArg)
+{
+   void *theProcessor = NULL, *sabSit = NULL, *preparsed = NULL;
+   char *resultURI = (char*)"file:///__stdout";
+   char *input = NULL;
+   int ecode;
+   int argcount = 2, firstarg = 1;
+
+   if ((batchMode != ID_BATCH_XML) &&
+       (batchMode != ID_BATCH_XSL) &&
+       (batchMode != ID_CHAIN_XSL))
+      return runSingleXSLT(arrayParams, arrayArgs, resultArg);
+
+   ecode = SablotCreateSituation(&sabSit);
+   if (ecode != 0)
+      return ecode;
+
+   if (commandFlags)
+     SablotSetOptions(sabSit, commandFlags);
+
+   ecode = SablotCreateProcessorForSituation(sabSit, &theProcessor);
+
+   if (ecode == 0)
+   {
+      if (stringLog)
+	 SablotSetLog(theProcessor, stringLog, 0);
+
+      if (stringBase)
+	 SablotSetBase(theProcessor, stringBase);
+
+      switch (batchMode)
+      {
+	 case ID_BATCH_XSL:
+	    ecode = SablotParseStylesheet(sabSit, arrayBare[0], &preparsed);
+	    break;
+	 case ID_BATCH_XML:
+	    ecode = SablotParse(sabSit, arrayBare[0], &preparsed);
+	    break;
+	 case ID_CHAIN_XSL:
+	    input = arrayBare[0];
+	    argcount = 1;
+	    firstarg = 2;
+	    break;
+      }
+   }
+
+   for (int t = firstarg; (t <= indexBare-argcount) && (ecode == 0); t += argcount)
+      {
+	 if (arrayParams)
+	 {
+	    const char **params = arrayParams;
+	    while ((params[0] != NULL) && (ecode == 0))
+	    {
+	       ecode = SablotAddParam(sabSit, theProcessor, params[0], params[1]);
+	       params += 2;
+	    }
+	 }
+
+	 if (arrayArgs)
+	 {
+	    const char **args = arrayArgs;
+	    while ((args[0] != NULL) && (ecode == 0))
+	    {
+	       ecode = SablotAddArgBuffer(sabSit, theProcessor, args[0], args[1]);
+	       args += 2;
+	    }
+	 }
+
+	 if (ecode == 0)
+	 {
+	 switch (batchMode)
+	    {
+	    case ID_BATCH_XSL:
+	    {
+	       const char *document = arrayBare[t];
+	       const char *output = arrayBare[t+1];
+	       SablotAddArgTree(sabSit, theProcessor, "stylesheet", preparsed);
+	       ecode = SablotRunProcessorGen(sabSit, theProcessor, "arg:stylesheet", document, output);
+	    } break;
+	    case ID_BATCH_XML:
+	    {
+	       const char *stylesheet = arrayBare[t];
+	       const char *output = arrayBare[t+1];
+	       SablotAddArgTree(sabSit, theProcessor, "source", preparsed);
+	       ecode = SablotRunProcessorGen(sabSit, theProcessor, stylesheet, "arg:source", output);
+	    } break;
+	    case ID_CHAIN_XSL:
+	    {
+	       const char *stylesheet = arrayBare[t];
+	       char *output;
+	       if (t < indexBare-argcount)
+		  output = (char*)"arg:output";
+	       else
+		  output = arrayBare[1];
+
+	       ecode = SablotRunProcessorGen(sabSit, theProcessor, stylesheet, input, output);
+
+	    if (ecode == 0)
+	       {
+		  if (input == arrayBare[0])
+		     input = "arg:buffer";
+		  else
+		     SablotFree(*resultArg);
+
+		  SablotGetResultArg(theProcessor, output, resultArg);
+		  SablotAddArgBuffer(sabSit, theProcessor, "buffer", *resultArg);
+	       }
+	    } break;
+	 }
+      }
+   }
+
+   if ((ecode == 0) && (batchMode != ID_CHAIN_XSL))
+      SablotGetResultArg(theProcessor, resultURI, resultArg);
+
+   if (preparsed)
+      SablotDestroyDocument(sabSit, preparsed);
+
+   if (theProcessor)
+      SablotDestroyProcessor(theProcessor);
+
+   if (sabSit)
+      SablotDestroySituation(sabSit);
+
+   return ecode;
+}
+
+void runDebugger()
+{
+#ifdef SABLOT_DEBUGGER
+  debuggerInit();
+  debuggerEnterIdle();
+  debuggerDone();
+#else
+  printf("Debugegr is not supported in this build\n");
+#endif
+}
+
 int main(int argc, char *argv[])
 {
     int indexParams, indexArgs, ecode, i;
     char *arrayParams[2 * MAX_ARG_OR_PAR + 2],
-        *arrayArgs[2 * MAX_ARG_OR_PAR + 2],
-        *resultArg = NULL;
+      *arrayArgs[2 * MAX_ARG_OR_PAR + 2];
+    char *resultArg = NULL;
     double timeZero = 0;
 
     // the following set the globals arrayBare and arrayEq
     // we get the number of words/assignments in indexBare/indexEq
-
+    
     readSwitches(argc, argv);
-
+    
     switch(numberMode)
-    {
-    case ID_HELP:
+      {
+      case ID_HELP:
         {
-            puts(usage); return 0;
+	  puts(usage); return 0;
         }; break;
-    case ID_DBG_HELP:
+      case ID_DBG_HELP:
         {
-            puts(dbg_usage); return 0;
+	  puts(dbg_usage); return 0;
         }; break;
-    case ID_VERSION:
+      case ID_VERSION:
         {
-            full_version_txt(); return 0;
+	  full_version_txt(); return 0;
         }; break;
-    }
+      case ID_DEBUGGER:
+	{
+	  runDebugger(); return 0;
+	}; break;
+      }
 
     if (!indexBare && numberMode != ID_DEBUG)
     {
@@ -505,34 +718,14 @@ int main(int argc, char *argv[])
         switch(numberMode)
         {
         case 0: 
-            {
-                void *theProcessor;
-                char *resultURI = NULL;
-
-                if (ecode = SablotCreateProcessor(&theProcessor))
-                    break;
-
-                if (stringLog)
-                    SablotSetLog(theProcessor, stringLog, 0);
-
-		if (stringBase)
-		    SablotSetBase(theProcessor, stringBase);
-
-                if (indexBare <= 2)
-                    resultURI = (char*)"file://stdout";
-                else
-                    resultURI = arrayBare[2];
-
-                if (0 != (ecode = SablotRunProcessor(theProcessor,
-                    arrayBare[0],
-                    indexBare <= 1  ? (char*)"file://stdin" : arrayBare[1],
-                    resultURI,
-                    arrayParams, arrayArgs)))
-                    SablotGetResultArg(theProcessor, resultURI, &resultArg);
-                
-                if (theProcessor)
-                    SablotDestroyProcessor(theProcessor);
-            }; break;
+	  {
+	     if ((batchMode == ID_BATCH_XML) || (batchMode == ID_BATCH_XSL) || (batchMode == ID_CHAIN_XSL))
+	      ecode = runBatchXSLT((const char**)arrayParams, 
+		(const char**)arrayArgs, &resultArg);
+	    else
+	      ecode = runSingleXSLT((const char**)arrayParams, 
+				    (const char**)arrayArgs, &resultArg);
+	  }; break;
         case ID_SPS:
             {
                 if (indexBare != 2)
@@ -552,8 +745,8 @@ int main(int argc, char *argv[])
             }; break;
         case ID_SPF: 
             ecode = SablotProcessFiles(arrayBare[0],
-                indexBare <= 1  ? (char*)"file://stdin" : arrayBare[1],
-                indexBare <= 2  ? (char*)"file://stdout" : arrayBare[2]);
+                indexBare <= 1  ? (char*)"file:///__stdin" : arrayBare[1],
+                indexBare <= 2  ? (char*)"file:///__stdout" : arrayBare[2]);
             break;
         }
         
