@@ -12,8 +12,8 @@
  * The Original Code is the Sablotron XSLT Processor.
  * 
  * The Initial Developer of the Original Code is Ginger Alliance Ltd.
- * Portions created by Ginger Alliance are Copyright (C) 2000 Ginger
- * Alliance Ltd. All Rights Reserved.
+ * Portions created by Ginger Alliance are Copyright (C) 2000-2002
+ * Ginger Alliance Ltd. All Rights Reserved.
  * 
  * Contributor(s): Han Qi
  * 
@@ -30,15 +30,17 @@
  * GPL.
  */
 
+#include "jsdom.h"
+
+#ifdef ENABLE_JS
+
 #define SablotAsExport
 #include "base.h"
 #include "sdom.h"
-#include "jsdom.h"
 #include "verts.h"
 
-#ifdef HAVE_JSAPI_H
-
 #define declPRIV  NodePrivate *np = (NodePrivate*)JS_GetPrivate(cx, obj)
+#define NPDOM np->situa->dom()
 
 /************************ domex  *********************/
 
@@ -57,7 +59,7 @@ JSClass domexClass = {
   JS_PropertyStub,
   JS_PropertyStub,
   JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub,
-  domexFinalize
+  domexFinalize, 0, 0, NULL, NULL, NULL, NULL, 0, 0
 };
 
 JSBool domexGetProperty (JSContext *cx, JSObject *obj, jsval id, jsval *vp)
@@ -73,16 +75,6 @@ JS_PROP(domexGetCode)
   return TRUE;
 }
 
-JS_METHOD(domexToString)
-{
-  DomExPrivate *priv = (DomExPrivate*)JS_GetPrivate(cx, obj);
-  char buff[200];
-  sprintf(buff, "[DOMException] code: %d", priv -> code);
-  JSString *str = JS_NewStringCopyZ(cx, buff);
-  *rval = STRING_TO_JSVAL(str);
-  return TRUE;
-}
-
 JSPropertySpec domexProtoProps[] =
 {
   {"INDEX_SIZE_ERR", 1, PROP_OPT, domexGetProperty, NULL},
@@ -100,19 +92,38 @@ JSPropertySpec domexProtoProps[] =
   {"INVALID_MODIFICATION_ERR", 13, PROP_OPT, domexGetProperty, NULL},
   {"NAMESPACE_ERR", 14, PROP_OPT, domexGetProperty, NULL},
   {"INVALID_ACCESS_ERR", 15, PROP_OPT, domexGetProperty, NULL},
-  NULL
+  {NULL, 0, 0, 0, 0}
 };
 
+JS_METHOD(domexToString)
+{
+  DomExPrivate *priv = (DomExPrivate*)JS_GetPrivate(cx, obj);
+  char buff[128];
+  //find the entry
+  JSPropertySpec* aux = domexProtoProps;
+  while (aux -> name)
+    {
+      if (aux -> tinyid == priv -> code) break;
+      aux++;
+    }
+
+  sprintf(buff, "[DOMException %d = %s]", priv -> code,
+	  aux -> name ? aux -> name : "unknown code");
+  JSString *str = JS_NewStringCopyZ(cx, buff);
+  *rval = STRING_TO_JSVAL(str);
+  return TRUE;
+}
+
 JSPropertySpec domexProps[] =
 {
   {"code", 0, PROP_OPT, domexGetCode, NULL},
-  NULL
+  {NULL, 0, 0, 0, 0}
 };
 
 JSFunctionSpec domexFunctions[] =
 {
-  {"toString", domexToString, 0, 0},
-  NULL
+  {"toString", domexToString, 0, 0, 0},
+  {NULL, 0, 0, 0, 0}
 };
 
 /***************nodelist array **********************/
@@ -164,11 +175,10 @@ JS_METHOD(nlistRemove)
 }
 
 JSFunctionSpec nlistFunctions[] = {
-  {"item", nlistItem, 0, 0},
-  //{"debug", nlistDebug, 0, 0},
-  {"append", nlistAppend, 0, 0},
-  //{"remove", nlistRemove, 0, 0}, //BUGGY
-  NULL
+  {"item", nlistItem, 0, 0, 0},
+  {"append", nlistAppend, 0, 0, 0},
+  //{"remove", nlistRemove, 0, 0, 0}, //BUGGY
+   {NULL, 0, 0, 0, 0}
 };
 
 JS_METHOD(nlistConstructor)
@@ -202,10 +212,10 @@ JSClass nlistClass = {
   JS_PropertyStub,
   JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub,
   JS_FinalizeStub,
+  0,
+  0,
   NULL,
-  NULL,
-  NULL,
-  nlistConstructor
+  nlistConstructor, NULL, NULL, 0, 0
 };
 
 /******************** dom implementation ************/
@@ -216,7 +226,7 @@ JSClass domimplClass = {
   JS_PropertyStub,
   JS_PropertyStub,
   JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub,
-  JS_FinalizeStub
+  JS_FinalizeStub, 0, 0, NULL, NULL, NULL, NULL, 0, 0
 };
 
 JS_METHOD(domimplHasFeature)
@@ -238,14 +248,14 @@ JS_METHOD(domimplCreateDocument)
 }
 
 JSFunctionSpec domimplFunctions[] = {
-  {"hasFeature", domimplHasFeature, 0, 0},
-  {"createDocumentType", domimplCreateDocumentType, 0, 0},
-  {"createDocument", domimplCreateDocument, 0, 0},
-  NULL
+  {"hasFeature", domimplHasFeature, 0, 0, 0},
+  {"createDocumentType", domimplCreateDocumentType, 0, 0, 0},
+  {"createDocument", domimplCreateDocument, 0, 0, 0},
+  {NULL, 0, 0, 0, 0}
 };
 
 /************************ node  *********************/
-void *_jsdom_getNodePrivate(Sit S, NodeHandle node)
+void* _jsdom_getNodePrivate(Sit S, NodeHandle node)
 {
   NodePrivate *np = new NodePrivate;
   np -> situa = &S;
@@ -273,10 +283,34 @@ JS_PROP(nodeGetNodeName)
   declPRIV;
   if (np) {
     char* val;
-    DOM_EX( SDOM_getNodeName(np->situa, np->node, &val) );
+    Bool doFree = FALSE;
+    switch(NPDOM.getNodeType(np->node)) {
+    case ELEMENT_NODE:
+    case ATTRIBUTE_NODE:
+    case PROCESSING_INSTRUCTION_NODE:
+      val = (char *)NPDOM.getNodeName(np->node);
+      doFree = TRUE;
+      break;
+    case TEXT_NODE:
+      val = "#text";
+      break;
+    case COMMENT_NODE:
+      val = "#comment";
+      break;
+    case DOCUMENT_NODE:
+      val = "#document";
+      break;
+    default:
+      /* NAMESPACE_NODE - not in DOM 2;
+       * CDATASection, DocumentFragment, DocumentType,
+       * Entity,EntityReference, Notation - not in Sablotron
+       */
+      val = (char *)NPDOM.getNodeName(np->node);
+      doFree = TRUE;
+    }
     JSString *str = JS_NewStringCopyZ(cx, val);
+    if (doFree) NPDOM.freeName(np->node, val);
     *rval = STRING_TO_JSVAL(str);
-    SablotFree(val);
     return TRUE;
   } 
   else {
@@ -288,11 +322,10 @@ JS_PROP(nodeGetNodeValue)
 {
   declPRIV;
   if (np) {
-    char* val;
-    DOM_EX( SDOM_getNodeValue(np->situa, np->node, &val) );
+    char* val = (char *)NPDOM.getNodeValue(np->node);
     JSString *str = JS_NewStringCopyZ(cx, val);
+    NPDOM.freeValue(np->node, val);
     *rval = STRING_TO_JSVAL(str);
-    SablotFree(val);
     return TRUE;
   } 
   else {
@@ -304,9 +337,7 @@ JS_PROP(nodeGetNodeType)
 {
   declPRIV;
   if (np) {
-    SDOM_NodeType val;
-    DOM_EX( SDOM_getNodeType(np->situa, np->node, &val) );
-    *rval = INT_TO_JSVAL(val);
+    *rval = INT_TO_JSVAL(NPDOM.getNodeType(np->node));
     return TRUE;
   } 
   else {
@@ -318,9 +349,7 @@ JS_PROP(nodeGetParentNode)
 {
   declPRIV;
   if (np) {
-    NodeHandle n;
-    DOM_EX( SDOM_getParentNode(np->situa, np->node, &n) );
-    JSObject *o = jsdom_createNode(cx, np, n);
+    JSObject *o = jsdom_createNode(cx, np, NPDOM.getParent(np->node));
     *rval = o ? OBJECT_TO_JSVAL(o) : JSVAL_NULL;
     return TRUE;
   }
@@ -335,16 +364,12 @@ JS_PROP(nodeGetChildNodes)
   if (np) {
     JSObject *arr = jsdom_createNodeList(cx, 0);
     *rval = OBJECT_TO_JSVAL(arr);
-    NodeHandle node;
-    DOM_EX( SDOM_getFirstChild(np->situa, np->node, &node) );
-    int i = 0;
-    while (node) 
+    int count = NPDOM.getChildCount(np->node);
+    for (int i = 0; i < count; i++)
       {
-	JSObject *x = jsdom_createNode(cx, np, node);
+	JSObject *x = jsdom_createNode(cx, np, NPDOM.getChildNo(np->node, i));
 	jsval xx = OBJECT_TO_JSVAL(x);
 	JS_SetElement(cx, arr, i, &xx);
-	SDOM_getNextSibling(np->situa, node, &node);
-	i++;
       }
     return TRUE;
   }
@@ -357,9 +382,7 @@ JS_PROP(nodeGetFirstChild)
 {
   declPRIV;
   if (np) {
-    NodeHandle n;
-    DOM_EX( SDOM_getFirstChild(np->situa, np->node, &n) );
-    JSObject *o = jsdom_createNode(cx, np, n);
+    JSObject *o = jsdom_createNode(cx, np, NPDOM.getChildNo(np->node, 0));
     *rval = o ? OBJECT_TO_JSVAL(o) : JSVAL_NULL;
     return TRUE;
   }
@@ -372,9 +395,9 @@ JS_PROP(nodeGetLastChild)
 {
   declPRIV;
   if (np) {
-    NodeHandle n;
-    DOM_EX( SDOM_getLastChild(np->situa, np->node, &n) );
-    JSObject *o = jsdom_createNode(cx, np, n);
+    JSObject *o = jsdom_createNode(cx, np, 
+				   NPDOM.getChildNo(np->node, 
+						    NPDOM.getChildCount(np->node) - 1 ));
     *rval = o ? OBJECT_TO_JSVAL(o) : JSVAL_NULL;
     return TRUE;
   }
@@ -387,9 +410,7 @@ JS_PROP(nodeGetPreviousSibling)
 {
   declPRIV;
   if (np) {
-    NodeHandle n;
-    DOM_EX( SDOM_getPreviousSibling(np->situa, np->node, &n) );
-    JSObject *o = jsdom_createNode(cx, np, n);
+    JSObject *o = jsdom_createNode(cx, np, NPDOM.getPreviousSibling(np->node));
     *rval = o ? OBJECT_TO_JSVAL(o) : JSVAL_NULL;
     return TRUE;
   }
@@ -402,9 +423,7 @@ JS_PROP(nodeGetNextSibling)
 {
   declPRIV;
   if (np) {
-    NodeHandle n;
-    DOM_EX( SDOM_getNextSibling(np->situa, np->node, &n) );
-    JSObject *o = jsdom_createNode(cx, np, n);
+    JSObject *o = jsdom_createNode(cx, np, NPDOM.getNextSibling(np->node));
     *rval = o ? OBJECT_TO_JSVAL(o) : JSVAL_NULL;
     return TRUE;
   }
@@ -417,34 +436,20 @@ JS_PROP(nodeGetAttributes)
 {
   declPRIV;
   if (np) {
-    SDOM_NodeType type;
-    SDOM_getNodeType(np->situa, np->node, &type);
-    if (type == SDOM_ELEMENT_NODE) {
+    if (NPDOM.getNodeType(np->node) == ELEMENT_NODE) {
       JSObject *nmap = jsdom_createNamedNodeMap(cx, 0);
       //iterate attributes
-      SDOM_NodeList l;
-      DOM_EX( SDOM_getAttributeList(np->situa, np->node, &l) );
-      int len;
-      DOM_EX( SDOM_getNodeListLength(np->situa, l, &len) );
-      //sdom returns namespaces w/ attributes
-      for (int i = 0; i < len; i++)
+      int attcount = NPDOM.getAttributeCount(np->node);
+      for (int i = 0; i < attcount; i++)
 	{
-	  NodeHandle n;
-  	  SDOM_getNodeListItem(np->situa, l, i, &n);
-	  SDOM_NodeType type;
-	  SDOM_getNodeType(np->situa, n, &type);
-	  if (type == SDOM_ATTRIBUTE_NODE) 
-	    {
-	      JSObject *att = jsdom_wrapNode(*(np->situa), cx, n);
-	      char *name;
-	      SDOM_getNodeName(np->situa, n, &name);
-	      jsval val = OBJECT_TO_JSVAL(att);
-	      JS_SetProperty(cx, nmap, name, &val);
-	      SablotFree(name);
-	    }
+	  SXP_Node n = NPDOM.getAttributeNo(np->node, i);
+	  JSObject *att = jsdom_wrapNode(*(np->situa), cx, n);
+	  jsval val = OBJECT_TO_JSVAL(att);
+	  char *name = (char *)NPDOM.getNodeName(n);
+	  JS_SetProperty(cx, nmap, name, &val);
+	  NPDOM.freeName(n,name);
 	}
       //return 
-      DOM_EX( SDOM_disposeNodeList(np->situa, l) );
       *rval = OBJECT_TO_JSVAL(nmap);
     } else {
       *rval = JSVAL_NULL;
@@ -460,9 +465,7 @@ JS_PROP(nodeGetOwnerDocument)
 {
   declPRIV;
   if (np) {
-    NodeHandle n;
-    DOM_EX( SDOM_getOwnerDocument(np->situa, np->node, &n) );
-    JSObject *o = jsdom_createNode(cx, np, n);
+    JSObject *o = jsdom_createNode(cx, np, NPDOM.getOwnerDocument(np->node));
     *rval = o ? OBJECT_TO_JSVAL(o) : JSVAL_NULL;
     return TRUE;
   }
@@ -475,16 +478,15 @@ JS_PROP(nodeGetNamespaceUri)
 {
   declPRIV;
   if (np) {
-    char* val;
-    DOM_EX( SDOM_getNodeNSUri(np->situa, np->node, &val) );
+    char* val = (char *)NPDOM.getNodeNameURI(np->node);
     if (val) {
       JSString *str = JS_NewStringCopyZ(cx, val);
       *rval = STRING_TO_JSVAL(str);
-      SablotFree(val);
     } 
     else {
       *rval = JSVAL_NULL;
     }
+    NPDOM.freeName(np->node,val);
     return TRUE;
   } 
   else {
@@ -496,16 +498,29 @@ JS_PROP(nodeGetPrefix)
 {
   declPRIV;
   if (np) {
-    char* val;
-    DOM_EX( SDOM_getNodePrefix(np->situa, np->node, &val) );
-    if (val) {
-      JSString *str = JS_NewStringCopyZ(cx, val);
+    char *val = NULL, *colon = NULL;
+    bool doFree = false;
+    switch(NPDOM.getNodeType(np->node)) {
+    case TEXT_NODE:
+    case DOCUMENT_NODE:
+    case COMMENT_NODE:
+      break;
+    case ELEMENT_NODE:
+    case ATTRIBUTE_NODE:
+    case PROCESSING_INSTRUCTION_NODE:
+    default: // see comment within nodeGetNodeName 
+      val = (char *)NPDOM.getNodeName(np->node);
+      colon = strchr(val,':');
+      doFree = true;
+    }
+    if (val && colon) {
+      JSString *str = JS_NewStringCopyN(cx, val, colon - val);
       *rval = STRING_TO_JSVAL(str);
-      SablotFree(val);
     } 
     else {
       *rval = JSVAL_NULL;
     }
+    if (doFree) NPDOM.freeName(np->node, val);
     return TRUE;
   } 
   else {
@@ -517,16 +532,15 @@ JS_PROP(nodeGetLocalName)
 {
   declPRIV;
   if (np) {
-    char* val;
-    DOM_EX( SDOM_getNodeLocalName(np->situa, np->node, &val) );
+    char* val = (char *)NPDOM.getNodeNameLocal(np->node);
     if (val) {
       JSString *str = JS_NewStringCopyZ(cx, val);
       *rval = STRING_TO_JSVAL(str);
-      SablotFree(val);
     } 
     else {
       *rval = JSVAL_NULL;
     }
+    NPDOM.freeName(np->node,val);
     return TRUE;
   } 
   else {
@@ -536,14 +550,12 @@ JS_PROP(nodeGetLocalName)
 
 JS_METHOD(nodeToString)
 {
-  //__fixme__
   declPRIV;
   if (np) {
-    char* val;
-    DOM_EX( SDOM_getNodeValue(np->situa, np->node, &val) );
-    JSString *str = JS_NewStringCopyZ(cx, val);
+    DStr val;
+    NPDOM.constructStringValue(np->node, val);
+    JSString *str = JS_NewStringCopyZ(cx, (char*)val);
     *rval = STRING_TO_JSVAL(str);
-    SablotFree(val);
     return TRUE;
   } 
   else {
@@ -561,9 +573,7 @@ JS_METHOD(nodeHasChildNodes)
 {
   declPRIV;
   if (np) {
-    NodeHandle n;
-    DOM_EX( SDOM_getFirstChild(np->situa, np->node, &n) );
-    if (n) *rval = JSVAL_TRUE;
+    if (NPDOM.getChildCount(np->node)) *rval = JSVAL_TRUE;
     else *rval = JSVAL_FALSE;
     return TRUE;
   }
@@ -589,33 +599,8 @@ JS_METHOD(nodeHasAttributes)
 {
   declPRIV;
   if (np) {
-    SDOM_NodeType type;
-    DOM_EX( SDOM_getNodeType(np->situa, np->node, &type) );
-    //element ?
-    if (type == SDOM_ELEMENT_NODE) {
-      SDOM_NodeList l;
-      DOM_EX( SDOM_getAttributeList(np->situa, np->node, &l) );
-      int len;
-      DOM_EX( SDOM_getNodeListLength(np->situa, l, &len) );
-      //count non ns atts
-      int num = 0;
-      //sdom returns namespaces together w/ attributes
-      for (int i = 0; i < len; i++)
-	{
-	  NodeHandle n;
-  	  SDOM_getNodeListItem(np->situa, l, i, &n);
-	  SDOM_NodeType type;
-	  SDOM_getNodeType(np->situa, n, &type);
-	  if (type == SDOM_ATTRIBUTE_NODE) num++;
-	}
-      //return 
-      if (num) *rval = JSVAL_TRUE;
-      else *rval = JSVAL_FALSE;
-      DOM_EX( SDOM_disposeNodeList(np->situa, l) );
-    } 
-    else { //another type
-      *rval = JSVAL_FALSE;
-    }
+    if (NPDOM.getAttributeCount(np->node)) *rval = JSVAL_TRUE;
+    else *rval = JSVAL_FALSE;
     return TRUE;
   }
   else {
@@ -630,7 +615,7 @@ JSClass nodeClass = {
   JS_PropertyStub,
   JS_PropertyStub,
   JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub,
-  nodeFinalize
+  nodeFinalize, 0, 0, NULL, NULL, NULL, NULL, 0, 0
 };
 
 //node prototype properties
@@ -661,7 +646,7 @@ JSPropertySpec nodeProtoProps[] =
    PROP_OPT, nodeGetProperty, NULL},
   {"NOTATION_NODE",  SDOM_NOTATION_NODE , 
    PROP_OPT, nodeGetProperty, NULL},
-  NULL
+  {NULL, 0, 0, 0, 0}
 };
 
 //'real' node props
@@ -682,25 +667,25 @@ JSPropertySpec nodeProps[] =
   {"namespaceURI", 0, PROP_OPT, nodeGetNamespaceUri, NULL},
   {"prefix", 0, PROP_OPT, nodeGetPrefix, NULL},
   {"localName", 0, PROP_OPT, nodeGetLocalName, NULL},
-  NULL
+  {NULL, 0, 0, 0, 0}  
 };
 
 
 JSFunctionSpec nodeFunctions[] = 
 {
   //unsupported
-  {"toString", nodeToString, 0, 0},
-  {"insertBefore", nodeUnsupported, 0, 0},
-  {"replaceChild", nodeUnsupported, 0, 0},
-  {"removeChild", nodeUnsupported, 0, 0},
-  {"appendChild", nodeUnsupported, 0, 0},
-  {"cloneNode", nodeUnsupported, 0, 0},
+  {"toString", nodeToString, 0, 0, 0},
+  {"insertBefore", nodeUnsupported, 0, 0, 0},
+  {"replaceChild", nodeUnsupported, 0, 0, 0},
+  {"removeChild", nodeUnsupported, 0, 0, 0},
+  {"appendChild", nodeUnsupported, 0, 0, 0},
+  {"cloneNode", nodeUnsupported, 0, 0, 0},
   //supported
-  {"hasChildNodes", nodeHasChildNodes, 0, 0},
-  {"normalize", nodeNormalize, 0, 0},
-  {"isSupported", nodeIsSupported, 0, 0},
-  {"hasAttributes", nodeHasAttributes, 0, 0},
-  NULL
+  {"hasChildNodes", nodeHasChildNodes, 0, 0, 0},
+  {"normalize", nodeNormalize, 0, 0, 0},
+  {"isSupported", nodeIsSupported, 0, 0, 0},
+  {"hasAttributes", nodeHasAttributes, 0, 0, 0},
+  {NULL, 0, 0, 0, 0}
 };
 
 /************************** document ***********************/
@@ -712,7 +697,7 @@ JSClass documentClass = {
   JS_PropertyStub,
   JS_PropertyStub,
   JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub,
-  nodeFinalize
+  nodeFinalize, 0, 0, NULL, NULL, NULL, NULL, 0, 0
 };
 
 JS_PROP(docGetType)
@@ -734,9 +719,7 @@ JS_PROP(docGetDocumentElement)
 {
   declPRIV;
   if (np) {
-    NodeHandle n;
-    DOM_EX( SDOM_getFirstChild(np->situa, np->node, &n) );
-    JSObject *o = jsdom_createNode(cx, np, n);
+    JSObject *o = jsdom_createNode(cx, np, NPDOM.getChildNo(np->node, 0));
     *rval = o ? OBJECT_TO_JSVAL(o) : JSVAL_NULL;
     return TRUE;
   }
@@ -761,28 +744,28 @@ JSPropertySpec documentProps[] = {
   {"doctype", 0, PROP_OPT, docGetType, NULL},
   {"implementation", 0, PROP_OPT, docGetImplementation, NULL},
   {"documentElement", 0, PROP_OPT, docGetDocumentElement, NULL},
-  NULL
+  {NULL, 0, 0, 0, 0}
 };
 
 JSFunctionSpec documentFunctions[] = {
   //not needed
-  {"createElement", nodeUnsupported, 0, 0},
-  {"createDocumentFragment", nodeUnsupported, 0, 0},
-  {"createTextNode", nodeUnsupported, 0, 0},
-  {"createComment", nodeUnsupported, 0, 0},
-  {"createCDATASection", nodeUnsupported, 0, 0},
-  {"createProcessingInstruction", nodeUnsupported, 0, 0},
-  {"createAttribute", nodeUnsupported, 0, 0},
-  {"createEntityReference", nodeUnsupported, 0, 0},
-  {"importNode", nodeUnsupported, 0, 0},
-  {"createElementNS", nodeUnsupported, 0, 0},
-  {"createAttributeNS", nodeUnsupported, 0, 0},
+  {"createElement", nodeUnsupported, 0, 0, 0},
+  {"createDocumentFragment", nodeUnsupported, 0, 0, 0},
+  {"createTextNode", nodeUnsupported, 0, 0, 0},
+  {"createComment", nodeUnsupported, 0, 0, 0},
+  {"createCDATASection", nodeUnsupported, 0, 0, 0},
+  {"createProcessingInstruction", nodeUnsupported, 0, 0, 0},
+  {"createAttribute", nodeUnsupported, 0, 0, 0},
+  {"createEntityReference", nodeUnsupported, 0, 0, 0},
+  {"importNode", nodeUnsupported, 0, 0, 0},
+  {"createElementNS", nodeUnsupported, 0, 0, 0},
+  {"createAttributeNS", nodeUnsupported, 0, 0, 0},
   //we do not support ids
-  {"getElementsById", nodeUnsupported, 0, 0},
+  {"getElementsById", nodeUnsupported, 0, 0, 0},
   //should be supported
-  {"getElementsByTagName", docElementsByTagName, 0, 0},
-  {"getElementsByTagNameNS", docElementsByTagNameNS, 0, 0},
-  NULL
+  {"getElementsByTagName", docElementsByTagName, 0, 0, 0},
+  {"getElementsByTagNameNS", docElementsByTagNameNS, 0, 0, 0},
+  {NULL, 0, 0, 0, 0}
 };
 
 /************************** element ************************/
@@ -794,7 +777,7 @@ JSClass elementClass = {
   JS_PropertyStub,
   JS_PropertyStub,
   JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub,
-  nodeFinalize
+  nodeFinalize, 0, 0, NULL, NULL, NULL, NULL, 0, 0
 };
 
 
@@ -803,10 +786,10 @@ JS_PROP(eleGetTagName)
   declPRIV;
   if (np) {
     char *name;
-    DOM_EX( SDOM_getNodeName(np->situa, np->node, &name) );
+    name = (char *)NPDOM.getNodeName(np->node);
     JSString *jname = JS_NewStringCopyZ(cx, name);
     *rval = STRING_TO_JSVAL(jname);
-    SablotFree(name);
+    NPDOM.freeName(np->node, name);
     return TRUE;
   }
   else {
@@ -815,7 +798,7 @@ JS_PROP(eleGetTagName)
 }
 
 JS_METHOD(eleGetAttribute)
-{
+{ 
   declPRIV;
   if (argc < 1) 
     {
@@ -826,10 +809,25 @@ JS_METHOD(eleGetAttribute)
     JSString *str = JSVAL_TO_STRING(argv[0]);
     char *name = JS_GetStringBytes(str);
     char *value;
-    DOM_EX( SDOM_getAttribute(np->situa, np->node, name, &value) );
-    JSString *jval = JS_NewStringCopyZ(cx, value);
+    int attcount = NPDOM.getAttributeCount(np->node);
+    char *attname;
+    NodeHandle attnode;
+    JSString *jval;
+    for (int i = 0; i < attcount; i++) {
+      attnode = NPDOM.getAttributeNo(np->node, i);
+      attname = (char *)NPDOM.getNodeName(attnode);
+      if ( !strcmp(name, attname) ) {
+	value = (char *)NPDOM.getNodeValue(attnode);
+	jval = JS_NewStringCopyZ(cx, value);
+	*rval = STRING_TO_JSVAL(jval);	
+	NPDOM.freeValue(attnode, value);
+	NPDOM.freeName(attnode, attname);
+	return TRUE;
+      };
+      NPDOM.freeName(attnode, attname);
+    };
+    jval = JS_NewStringCopyZ(cx, "");
     *rval = STRING_TO_JSVAL(jval);
-    SablotFree(value);
     return TRUE;
   }
   else {
@@ -842,16 +840,27 @@ JS_METHOD(eleGetAttributeNode)
   declPRIV;
   if (argc < 1) 
     {
-      JS_ReportError(cx, "Element.getAttributeNode must have one parameter");
+      JS_ReportError(cx, "Element.getAttribute must have one parameter");
       return FALSE;
     }
   if (np) {
     JSString *str = JSVAL_TO_STRING(argv[0]);
     char *name = JS_GetStringBytes(str);
-    SDOM_Node attr;
-    DOM_EX( SDOM_getAttributeNode(np->situa, np->node, name, &attr) );
-    JSObject *o = jsdom_createNode(cx, np, attr);
-    *rval = o ? OBJECT_TO_JSVAL(o) : JSVAL_NULL;
+    int attcount = NPDOM.getAttributeCount(np->node);
+    char *attname;
+    NodeHandle attnode;
+    for (int i = 0; i < attcount; i++) {
+      attnode = NPDOM.getAttributeNo(np->node, i);
+      attname = (char *)NPDOM.getNodeName(attnode);
+      if ( !strcmp(name, attname) ) {
+	JSObject *o = jsdom_createNode(cx, np, attnode);
+	*rval = o ? OBJECT_TO_JSVAL(o) : JSVAL_NULL;
+	NPDOM.freeName(attnode, attname);
+	return TRUE;
+      };
+      NPDOM.freeName(attnode, attname);
+    };
+    *rval = JSVAL_NULL;
     return TRUE;
   }
   else {
@@ -869,12 +878,22 @@ JS_METHOD(eleHasAttribute)
 {
   declPRIV;
   if (np) {
-    SDOM_Node attr;
-    JSString *jname = JSVAL_TO_STRING(argv[0]);
-    DOM_EX( SDOM_getAttributeNode(np->situa, np->node, 
-				  JS_GetStringBytes(jname), &attr) );
-    JSBool rv = attr != 0;
-    *rval = BOOLEAN_TO_JSVAL(rv);
+    JSString *str = JSVAL_TO_STRING(argv[0]);
+    char *name = JS_GetStringBytes(str);
+    int attcount = NPDOM.getAttributeCount(np->node);
+    char *attname;
+    NodeHandle attnode;
+    for (int i = 0; i < attcount; i++) {
+      attnode = NPDOM.getAttributeNo(np->node, i);
+      attname = (char *)NPDOM.getNodeName(attnode);
+      if ( !strcmp(name, attname) ) {
+	NPDOM.freeName(attnode, attname);
+	*rval = JSVAL_TRUE;
+	return TRUE;
+      };
+      NPDOM.freeName(attnode, attname);
+    };
+    *rval = JSVAL_FALSE;
     return TRUE;
   }
   else {
@@ -885,27 +904,27 @@ JS_METHOD(eleHasAttribute)
 JSPropertySpec elementProps[] =
 {
   {"tagName", 0, PROP_OPT, eleGetTagName, NULL},
-  NULL
+  {NULL, 0, 0, 0, 0}
 };
 
 JSFunctionSpec elementFunctions[] = 
 {
-  {"getAttribute", eleGetAttribute, 0, 0},
-  {"setAttribute", nodeUnsupported, 0, 0},
-  {"removeAttribute", nodeUnsupported, 0, 0},
-  {"getAttributeNode", eleGetAttributeNode, 0, 0},
-  {"setAttributeNode", nodeUnsupported, 0, 0},
-  {"removeAttributeNode", nodeUnsupported, 0, 0},
-  {"getElementsByTagName", eleGetElementsByTagName, 0, 0},
-  {"getAttributeNS", nodeUnsupported, 0, 0},
-  {"setAttributeNS", nodeUnsupported, 0, 0},
-  {"removeAttributeNS", nodeUnsupported, 0, 0},
-  {"getAttributeNodeNS", nodeUnsupported, 0, 0},
-  {"setAttributeNodeNS", nodeUnsupported, 0, 0},
-  {"getElementsByTagNameNS", nodeUnsupported, 0, 0},
-  {"hasAttribute", eleHasAttribute, 0, 0},
-  {"hasAttributeNS", nodeUnsupported, 0, 0},
-  NULL
+  {"getAttribute", eleGetAttribute, 0, 0, 0},
+  {"setAttribute", nodeUnsupported, 0, 0, 0},
+  {"removeAttribute", nodeUnsupported, 0, 0, 0},
+  {"getAttributeNode", eleGetAttributeNode, 0, 0, 0},
+  {"setAttributeNode", nodeUnsupported, 0, 0, 0},
+  {"removeAttributeNode", nodeUnsupported, 0, 0, 0},
+  {"getElementsByTagName", eleGetElementsByTagName, 0, 0, 0},
+  {"getAttributeNS", nodeUnsupported, 0, 0, 0},
+  {"setAttributeNS", nodeUnsupported, 0, 0, 0},
+  {"removeAttributeNS", nodeUnsupported, 0, 0, 0},
+  {"getAttributeNodeNS", nodeUnsupported, 0, 0, 0},
+  {"setAttributeNodeNS", nodeUnsupported, 0, 0, 0},
+  {"getElementsByTagNameNS", nodeUnsupported, 0, 0, 0},
+  {"hasAttribute", eleHasAttribute, 0, 0, 0},
+  {"hasAttributeNS", nodeUnsupported, 0, 0, 0},
+  {NULL, 0, 0, 0, 0}
 };
 
 /********************* Attr ************************/
@@ -917,7 +936,7 @@ JSClass attrClass = {
   JS_PropertyStub,
   JS_PropertyStub,
   JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub,
-  nodeFinalize
+  nodeFinalize, 0, 0, NULL, NULL, NULL, NULL, 0, 0
 };
 
 JS_PROP(attrGetSpecified)
@@ -930,9 +949,7 @@ JS_PROP(attrGetOwnerElement)
 {
   declPRIV;
   if (np) {
-    NodeHandle n;
-    DOM_EX( SDOM_getAttributeElement(np->situa, np->node, &n) );
-    JSObject *o = jsdom_createNode(cx, np, n);
+    JSObject *o = jsdom_createNode(cx, np, NPDOM.getParent(np->node));
     *rval = o ? OBJECT_TO_JSVAL(o) : JSVAL_NULL;
     return TRUE;
   }
@@ -946,7 +963,7 @@ JSPropertySpec attrProps[] = {
   {"value", 0, PROP_OPT, nodeGetNodeValue, NULL},
   {"specified", 0, PROP_OPT, attrGetSpecified, NULL},
   {"ownerElement", 0, PROP_OPT, attrGetOwnerElement, NULL},
-  NULL
+  {NULL, 0, 0, 0, 0}
 };
 
 /******************* character data & Co. **********/
@@ -958,22 +975,21 @@ JSClass chardataClass = {
   JS_PropertyStub,
   JS_PropertyStub,
   JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub,
-  nodeFinalize
+  nodeFinalize, 0, 0, NULL, NULL, NULL, NULL, 0, 0
 };
 
 JS_PROP(chardataGetData)
 {
   declPRIV;
   if (np) {
-    char *val;
-    DOM_EX( SDOM_getNodeValue(np->situa, np->node, &val) );
+    char *val = (char *)NPDOM.getNodeValue(np->node);
     if (val) {
       JSString *jval = JS_NewStringCopyZ(cx, val);
       *rval = STRING_TO_JSVAL(jval);
-      SablotFree(val);
     } else {
       *rval = JSVAL_NULL;
     }
+    NPDOM.freeValue(np->node, val);
     return TRUE;
   }
   else {
@@ -981,32 +997,18 @@ JS_PROP(chardataGetData)
   }
 }
 
-int _getDataLen(SDOM_Node n)
+int _getDataLen(const char* data)
 {
-  Vertex * v = (Vertex*)n;
-  int len;
-  switch (v -> vt & VT_BASE) {
-  case VT_TEXT:
-    len = toText(v) -> cont.length();
-    break;
-  case VT_COMMENT:
-    len = toComment(v) -> cont.length();
-    break;
-  default:
-    {
-      len = 0;
-    };
-  }
-  return len;
+  return strlen((char *)data);
 }
 
 JS_PROP(chardataGetLength)
 {
-  //bypassing the dom interface - DANGEROUS!!
   declPRIV;
   if (np) {
-    int len = _getDataLen(np->node);
-    *rval = INT_TO_JSVAL(len);
+    const char* val = NPDOM.getNodeValue(np->node);
+    *rval = INT_TO_JSVAL(_getDataLen(val));
+    NPDOM.freeValue(np->node, (char*)val);
     return TRUE;
   }
   else {
@@ -1016,7 +1018,6 @@ JS_PROP(chardataGetLength)
 
 JS_METHOD(chardataSubstringData)
 {
-  //bypassing the dom interface - DANGEROUS!!
   declPRIV;
   if (argc < 2)
     {
@@ -1025,7 +1026,8 @@ JS_METHOD(chardataSubstringData)
       return FALSE;
     }
   if (np) {
-    int len = _getDataLen(np->node);
+    const char *val = NPDOM.getNodeValue(np->node);
+    int len = _getDataLen(val);
     int off = JSVAL_TO_INT(argv[0]);
     int count = JSVAL_TO_INT(argv[1]);
     JSString *str;
@@ -1036,11 +1038,9 @@ JS_METHOD(chardataSubstringData)
     else 
       {
 	if (off + count > len) count = len - off;
-	char *val;
-	DOM_EX( SDOM_getNodeValue(np->situa, np->node, &val) );
 	str = JS_NewStringCopyN(cx, val + off, count);
-	SablotFree(val);
       }
+    NPDOM.freeValue(np->node, (char*)val);
     *rval = STRING_TO_JSVAL(str);
     return TRUE;
   }
@@ -1053,17 +1053,17 @@ JSPropertySpec chardataProps[] =
 {
   {"data", 0, PROP_OPT, chardataGetData, NULL},
   {"length", 0, PROP_OPT, chardataGetLength, NULL},
-  NULL
+  {NULL, 0, 0, 0, 0}
 };
 
 JSFunctionSpec chardataFunctions[] = 
 {
-  {"substringData", chardataSubstringData, 0, 0},
-  {"appendData", nodeUnsupported, 0, 0},
-  {"insertData", nodeUnsupported, 0, 0},
-  {"deleteData", nodeUnsupported, 0, 0},
-  {"replaceData", nodeUnsupported, 0, 0},
-  NULL
+  {"substringData", chardataSubstringData, 0, 0, 0},
+  {"appendData", nodeUnsupported, 0, 0, 0},
+  {"insertData", nodeUnsupported, 0, 0, 0},
+  {"deleteData", nodeUnsupported, 0, 0, 0},
+  {"replaceData", nodeUnsupported, 0, 0, 0},
+  {NULL, 0, 0, 0, 0}
 };
 
 /********************* text ************************/
@@ -1075,13 +1075,13 @@ JSClass textClass = {
   JS_PropertyStub,
   JS_PropertyStub,
   JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub,
-  nodeFinalize
+  nodeFinalize, 0, 0, NULL, NULL, NULL, NULL, 0, 0
 };
 
 JSFunctionSpec textFunctions[] = 
 {
-  {"splitText", nodeUnsupported, 0, 0},
-  NULL
+  {"splitText", nodeUnsupported, 0, 0, 0},
+  {NULL, 0, 0, 0, 0}
 };
 
 /******************** cdata ************************/
@@ -1093,7 +1093,7 @@ JSClass cdataClass = {
   JS_PropertyStub,
   JS_PropertyStub,
   JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub,
-  nodeFinalize
+  nodeFinalize, 0, 0, NULL, NULL, NULL, NULL, 0, 0
 };
 
 /***************** comment *************************/
@@ -1105,7 +1105,7 @@ JSClass commentClass = {
   JS_PropertyStub,
   JS_PropertyStub,
   JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub,
-  nodeFinalize
+  nodeFinalize, 0, 0, NULL, NULL, NULL, NULL, 0, 0
 };
 
 /********************* named node map **************/
@@ -1149,10 +1149,9 @@ JS_METHOD(nnmSetNamedItem)
   if (proto == cxi->node || cls == &nodeClass)
     {
       NodePrivate *np = (NodePrivate*)JS_GetPrivate(cx, nobj);
-      char *name;
-      DOM_EX( SDOM_getNodeName(np->situa, np->node, &name) );
+      char *name = (char *)NPDOM.getNodeName(np->node);
       JS_SetProperty(cx, obj, name, &(argv[0]));
-      SablotFree(name);
+      NPDOM.freeName(np->node,name);
       return TRUE;
     }
   else
@@ -1179,7 +1178,6 @@ JS_METHOD(nnmItem)
   if (argc >= 1 || ! JSVAL_IS_INT(argv[0]) )
     {
       JSIdArray *arr = JS_Enumerate(cx, obj);
-      
       int idx = JSVAL_TO_INT(argv[0]);
       if (idx < 0 || idx >= arr -> length - 1)
 	{
@@ -1205,18 +1203,18 @@ JS_METHOD(nnmItem)
 
 JSPropertySpec nnmProps[] = {
   {"length", 0, PROP_OPT, nnmGetLength, NULL},
-  NULL
+  {NULL, 0, 0, 0, 0}
 };
 
 JSFunctionSpec nnmFunctions[] = {
-  {"getNamedItem", nnmGetNamedItem, 0, 0},
-  {"setNamedItem", nnmSetNamedItem, 0, 0},
-  {"removeNamedItem", nnmRemoveNamedItem, 0, 0},
-  {"item", nnmItem, 0, 0},
-  {"getNamedItemsNS", nodeUnsupported, 0, 0},
-  {"setNamedItemsNS", nodeUnsupported, 0, 0},
-  {"removeNamedItemsNS", nodeUnsupported, 0, 0},
-  NULL
+  {"getNamedItem", nnmGetNamedItem, 0, 0, 0},
+  {"setNamedItem", nnmSetNamedItem, 0, 0, 0},
+  {"removeNamedItem", nnmRemoveNamedItem, 0, 0, 0},
+  {"item", nnmItem, 0, 0, 0},
+  {"getNamedItemsNS", nodeUnsupported, 0, 0, 0},
+  {"setNamedItemsNS", nodeUnsupported, 0, 0, 0},
+  {"removeNamedItemsNS", nodeUnsupported, 0, 0, 0},
+  {NULL, 0, 0, 0, 0}
 };
 
 JSClass nnmClass = {
@@ -1226,12 +1224,12 @@ JSClass nnmClass = {
   JS_PropertyStub,
   JS_PropertyStub,
   JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub,
-  JS_FinalizeStub,
+  JS_FinalizeStub, 0, 0, NULL, NULL, NULL, NULL, 0, 0
 };
 
 /********************* functions *******************/
 
-JSClass *_jsdom_getNodeClass(SDOM_NodeType type) 
+JSClass* _jsdom_getNodeClass(SDOM_NodeType type) 
 {
   JSClass *cls;
   switch (type) 
@@ -1262,14 +1260,12 @@ JSClass *_jsdom_getNodeClass(SDOM_NodeType type)
   return cls;
 }
 
-JSClass *_jsdom_getNodeClass(Sit S, JSContext *cx, NodeHandle node) 
+JSClass* _jsdom_getNodeClass(Sit S, JSContext *cx, NodeHandle node) 
 {
-  SDOM_NodeType type; 
-  DOM_EX( SDOM_getNodeType(&S, node, &type) );
-  return _jsdom_getNodeClass(type);
+  return _jsdom_getNodeClass((SDOM_NodeType)S.dom().getNodeType(node));
 }
 
-JSObject *jsdom_wrapNode(Sit S, JSContext *cx, NodeHandle node) 
+JSObject* jsdom_wrapNode(Sit S, JSContext *cx, NodeHandle node) 
 {
   JSObject *obj = NULL;
   if (node)
@@ -1277,9 +1273,8 @@ JSObject *jsdom_wrapNode(Sit S, JSContext *cx, NodeHandle node)
       JSContextItem *item = (JSContextItem*)JS_GetContextPrivate(cx);
       assert(item);
       
-      SDOM_NodeType type; 
-      DOM_EX( SDOM_getNodeType(&S, node, &type) );
-      
+      SDOM_NodeType type = (SDOM_NodeType)S.dom().getNodeType(node);
+
       obj = JS_NewObject(cx, _jsdom_getNodeClass(type), 
 				   item->node, NULL);
       JS_SetPrivate(cx, obj, _jsdom_getNodePrivate(S, node));
@@ -1327,7 +1322,7 @@ JSObject *jsdom_wrapNode(Sit S, JSContext *cx, NodeHandle node)
   return obj;
 }
 
-JSObject *jsdom_createNode(JSContext *cx, NodePrivate *np_, NodeHandle node) 
+JSObject* jsdom_createNode(JSContext *cx, NodePrivate *np_, NodeHandle node) 
 {
   return jsdom_wrapNode(*(np_->situa), cx, node);
 }
@@ -1392,7 +1387,7 @@ void jsdom_raiseException(JSContext *cx, int code)
   JS_SetPendingException(cx, OBJECT_TO_JSVAL(obj));
 }
 
-JSObject *jsdom_createNodeList(JSContext *cx, int num)
+JSObject* jsdom_createNodeList(JSContext *cx, int num)
 {
   //JSObject *obj = JS_NewArrayObject(cx, 0, NULL);
   //JS_DefineFunctions(cx, obj, nlistFunctions);
@@ -1401,7 +1396,7 @@ JSObject *jsdom_createNodeList(JSContext *cx, int num)
   return obj;
 }
 
-JSObject *jsdom_createNamedNodeMap(JSContext *cx, int num)
+JSObject* jsdom_createNamedNodeMap(JSContext *cx, int num)
 {
   JSObject *obj = JS_NewObject(cx, &nnmClass, NULL, NULL);
   JS_DefineFunctions(cx, obj, nnmFunctions);
@@ -1410,5 +1405,4 @@ JSObject *jsdom_createNamedNodeMap(JSContext *cx, int num)
   return obj;
 }
 
-
-#endif
+#endif //ENABLE_JS
