@@ -12,8 +12,8 @@
  * The Original Code is the Sablotron XSLT Processor.
  * 
  * The Initial Developer of the Original Code is Ginger Alliance Ltd.
- * Portions created by Ginger Alliance are Copyright (C) 2000 Ginger
- * Alliance Ltd. All Rights Reserved.
+ * Portions created by Ginger Alliance are Copyright (C) 2000-2002
+ * Ginger Alliance Ltd. All Rights Reserved.
  * 
  * Contributor(s): Marc Lehmann <pcg@goof.com>
  * 
@@ -39,7 +39,11 @@
 #include "proc.h"
 
 #ifdef HAVE_CONFIG_H
-#include <autocfg.h>
+#include <config.h>
+#endif
+
+#ifdef SABLOT_DEBUGGER
+#include "debugger.h"
 #endif
 
 #include "guard.h"
@@ -67,6 +71,11 @@ enum OutputterFlags
     { if (mySAXHandler)\
         { mySAXHandler -> FUNC(mySAXUserData, S.getProcessor(), (ARG1), (ARG2)); }}
 
+//macros on internal handler
+#define IF_INT_SAX4( FUNC, ARG1, ARG2, ARG3 )\
+    { if (mySAXHandler)\
+        { ((SAXHandlerInternal*)mySAXHandler) -> FUNC(mySAXUserData, S.getProcessor(), (ARG1), (ARG2), (ARG3)); }}
+
 #define IF_PH1( FUNC )\
     { if (physical)\
         { physical -> FUNC(S); }}
@@ -404,6 +413,7 @@ PhysicalOutputLayerObj::PhysicalOutputLayerObj(CDesc encodingCD_)
     indent = FALSE;
     after_markup = FALSE;  //pc
     level = 0;  //pc
+    defaultNSWas = false;
 }
 
 PhysicalOutputLayerObj::~PhysicalOutputLayerObj()
@@ -436,8 +446,8 @@ eFlag PhysicalOutputLayerObj::outputTrailingNewline (Sit S)
 
 eFlag PhysicalOutputLayerObj::outputDone(Sit S)
 {
-    E( flushBuffer(S) );
-    return OK;
+  E( flushBuffer(S) );
+  return OK;
 }
 
 Str _localName(Str fullname)
@@ -472,9 +482,18 @@ eFlag PhysicalOutputLayerObj::outputElementStart(Sit S,
     const Str* prefix;
     for (i = namespace_index; i < namespaces.number(); i++)
     {
-      prefix = &(namespaces[i] -> prefix);
-      if (! namespaces.isHidden(*prefix))
-       {
+      NamespaceStackObj * nm = namespaces[i];
+      prefix = &(nm -> prefix);
+      //prefix = &(namespaces[i] -> prefix); 
+      //if (!namespaces.isHidden(*prefix))
+
+      //swallow hidden namespaces as well as the empty default
+      //namespace, if no default namespace was output recently
+      if (! nm -> hidden &&
+	  (!(prefix -> isEmpty()) || !(nm -> uri.isEmpty()) || defaultNSWas))
+	{
+	  defaultNSWas = defaultNSWas || prefix -> isEmpty();
+
           sendLit(S, " xmlns");
           if (!prefix -> isEmpty())
           {
@@ -482,7 +501,7 @@ eFlag PhysicalOutputLayerObj::outputElementStart(Sit S,
              E( sendStr(S, *prefix) );
           }
           sendLit(S, "=\"");
-          E( sendStrEsc(S, namespaces[i] -> uri, 
+          E( sendStrEsc(S, nm -> uri, 
                         method == OUTPUT_HTML || method == OUTPUT_XHTML ?
                         ESCAPING_HTML_URI : ESCAPING_URI) );
           sendLit(S, "\"");
@@ -536,7 +555,8 @@ eFlag PhysicalOutputLayerObj::outputElementStart(Sit S,
 	      }
 	    else 
 	      {
-		escapingMode = ESCAPING_HTML_ATTR;
+		if (method == OUTPUT_HTML) 
+		  escapingMode = ESCAPING_HTML_ATTR;
 	      }
 	  }
 
@@ -854,11 +874,10 @@ eFlag PhysicalOutputLayerObj::sendOut(Sit S,
                 case ESCAPING_ATTR:
                 case ESCAPING_LT_AMP:
                     {
-		                if (method == OUTPUT_HTML || method == OUTPUT_XHTML)
-				        {
-                            E( sendLit(S, "&gt;") );
-                            served = TRUE;
-			            }
+		      //now we escape &gt; always to prevent bad
+		      //escaping in cdata sections.
+		      E( sendLit(S, "&gt;") );
+		      served = TRUE;
                     }; break;
                 }
             }; break;
@@ -945,22 +964,23 @@ eFlag PhysicalOutputLayerObj::sendOut(Sit S,
             }; break;
 	    default:
 	        {
-		        // escape non-ASCII values in URIs
-		        if (*data < 0)
-			    {
-                    switch(escapeMode)
-                    {
-                    case ESCAPING_URI:
-                    case ESCAPING_HTML_URI:
-                        {
+		  // escape non-ASCII values in URIs
+		  if (*data & 0x80)
+		    {
+		      switch(escapeMode)
+			{
+			case ESCAPING_URI:
+			case ESCAPING_HTML_URI:
+			  {
                             E( sendOut(S, smallBuf, 
-                                writeCharacterRef(smallBuf, data, escapeMode),
-                                ESCAPING_NONE) );
+				       writeCharacterRef(smallBuf, data, 
+							 escapeMode),
+				       ESCAPING_NONE) );
                             served = TRUE;
-                        }; break;
-                    }
-			    }			        
-		    }
+			  }; break;
+			}
+		    }			        
+		}
         };
 
 
@@ -969,6 +989,14 @@ eFlag PhysicalOutputLayerObj::sendOut(Sit S,
             srcCharLen = utf8SingleCharLength(data);
             assert(srcCharLen > 0);
             // if encoding differs from utf-8 then recode
+
+#ifdef SABLOT_DEBUGGER
+	    if ( debugger )
+	      {
+		debugger -> printOutput(data, srcCharLen);
+	      }
+#endif
+
             if (encodingCD == (void*)-1)
             {
                 memcpy(buffer + curr, data, srcCharLen);
@@ -985,21 +1013,23 @@ eFlag PhysicalOutputLayerObj::sendOut(Sit S,
                 S.recoder().conv(S, encodingCD, data, inleft, outbuf, outleft, result);
                 // data is shifted automatically, set curr
                 curr = outbuf - buffer;
-                // check result, write a character code if couldn't convert (unless ESCAPING_NONE)
+                // check result, write a character code if 
+		//couldn't convert (unless ESCAPING_NONE)
                 assert(result != ENC_EINVAL && result != ENC_E2BIG);
                 // FIXME: in NT, must check differently
                 if (result == ENC_EILSEQ)
                 {
                     // unable to convert
                     destCharLen = writeCharacterRef(smallBuf, data, escapeMode);
-                    if (escapeMode == ESCAPING_NONE)
+                    //if (escapeMode == ESCAPING_NONE)
+		    if (method == OUTPUT_TEXT)
                         Err1(S, E1_BAD_CHAR_IN_ENC, smallBuf)
                     else
-                    {
+		      {
                         E( sendOut(S, smallBuf, destCharLen, ESCAPING_NONE) );
                         data += srcCharLen;
                         served = TRUE;
-			        }
+		      }
                 }
             }
         }   // if (!served)
@@ -1038,6 +1068,16 @@ eFlag PhysicalOutputLayerObj::setMethodByDefault(Sit S, OutputMethod method_)
     return OK;
 }
 
+eFlag PhysicalOutputLayerObj::close(Sit S) 
+{
+  DataLine * dl;
+  if (dl = getDataLine())
+    {
+      E( dl -> close(S) );
+    }
+  return OK;
+}
+
 void PhysicalOutputLayerObj::report(Sit S, MsgType type, MsgCode code, const Str& arg1, const Str& arg2)
 {
     S.message(type, code, arg1, arg2);
@@ -1071,6 +1111,37 @@ void FrontMatter::report(Sit S, MsgType type, MsgCode code, const Str& arg1, con
     S.message(type, code, arg1, arg2);
 }
 
+//
+//
+//  OutputDocument
+//
+//
+
+OutputDocument::~OutputDocument()
+{
+  //we do not delete outputter objects as they are stored with
+  //Processor
+  // if (outputter) delete outputter;
+  if (def) delete def;
+}
+
+eFlag OutputDocument::finish(Sit S)
+{
+  if (state == OUTDOC_ACTIVE)
+    {
+      E( NZ(outputter) -> eventTrailingNewline(S) );
+      E( NZ(outputter) -> eventEndOutput(S, true) );
+      setState(OUTDOC_FINISHED);
+    }
+  return OK;
+}
+
+OutputterObj* OutputDocument::setOutputter(OutputterObj* new_)
+{
+  if (outputter) delete outputter;
+  return outputter = new_;
+}
+
 //
 //
 //  OutputterObj
@@ -1080,24 +1151,25 @@ void FrontMatter::report(Sit S, MsgType type, MsgCode code, const Str& arg1, con
 
 Str* OutputterObj::nameForSAX(Sit S, const EQName& q)
 {
-  if (q.getUri().isEmpty())
-    return new Str(q.getLocal());
-  DStr temp = q.getUri();
-  temp += THE_NAMESPACE_SEPARATOR;
-  temp += q.getLocal();
-  return new Str(temp);
-
-//    Str aliname;
-//    if (S.getProcessor())
-//      aliname = S.getProcessor() -> getAliasedName(q, currNamespaces);
-//    else
-//      q.getname(aliname);
-//    char *p = new char[aliname.length() + 1];
-//    strcpy(p, (char*) aliname);
-//    char *colon = strchr(p, ':');
-//    if (colon) 
-//      *colon = THE_NAMESPACE_SEPARATOR;
-//    return new Str(p);
+  DStr temp;
+  if (mySAXOutputType == SAXOUTPUT_COPY_TREE || 
+      mySAXOutputType == SAXOUTPUT_INT_PHYSICAL)
+    {
+      if (q.getUri().isEmpty())
+	return new Str(q.getLocal());
+      temp = q.getUri();
+      temp += THE_NAMESPACE_SEPARATOR;
+      temp += q.getLocal();
+      temp += THE_NAMESPACE_SEPARATOR;
+      temp += q.getPrefix();
+      return new Str(temp);
+    }
+  else
+    {
+      Str s;
+      q.getname(s);
+      return new Str(s);
+    }
 };
 
 OutputterObj::OutputterObj()
@@ -1107,15 +1179,12 @@ OutputterObj::OutputterObj()
     physical = NULL;
     outDef = NULL;
     mySAXHandler = NULL;
-    mySAXOutputType = SAXOUTPUT_COPY_TREE;
+    mySAXOutputType = SAXOUTPUT_NONE;
     method = OUTPUT_UNKNOWN;
     noElementYet = TRUE;
     noHeadYet = TRUE;
     delayedDTD = FALSE;
-       // to recognize the first HEAD element
-    //this was an old-fashioned way to supress output of namespaces
-    //currNamespaces.appendConstruct("xsl",Str(theXSLTNamespace),TRUE);
-    //currNamespaces.appendConstruct("xml",Str(theXMLNamespace),TRUE);
+    // to recognize the first HEAD element
 }
 
 OutputterObj::~OutputterObj()
@@ -1123,7 +1192,6 @@ OutputterObj::~OutputterObj()
     history.freeall(FALSE);
     front.freeall(FALSE);
     currNamespaces.freeall(FALSE);
-    excludedNS.deppendall();
     cdelete(physical);
 }
 
@@ -1137,7 +1205,12 @@ eFlag OutputterObj::setOptions(Sit S,
         E( outDef -> setDefaults(S) );
 
     if (S.getProcessor())
+      {
         encoding = S.getProcessor() -> getHardEncoding();
+	if ( !encoding.isEmpty() )
+	  outDef -> setItemStr(S, XSLA_ENCODING, encoding, NULL, 
+			       OUTPUT_PRECEDENCE_STRONGEST);
+      }
     else
 	encoding.empty();
     if (encoding.isEmpty())
@@ -1156,12 +1229,14 @@ eFlag OutputterObj::setOptions(Sit S,
         {
             Warn1(S, W1_UNSUPP_OUT_ENCODING, encoding);
             encoding = "UTF-8";
-            E( outDef -> setItemStr(S, XSLA_ENCODING, encoding, NULL, OUTPUT_PRECEDENCE_STRONGEST) );
+            E( outDef -> setItemStr(S, XSLA_ENCODING, encoding, 
+				    NULL, OUTPUT_PRECEDENCE_STRONGEST) );
         }
     }
     else
 	if (!encoding.isEmpty())
-	    E( outDef -> setItemStr(S, XSLA_ENCODING, encoding, NULL, OUTPUT_PRECEDENCE_WEAKEST) );
+	    E( outDef -> setItemStr(S, XSLA_ENCODING, encoding, 
+				    NULL, OUTPUT_PRECEDENCE_WEAKEST) );
 
     if (targetDataLine_)
     {
@@ -1276,20 +1351,26 @@ eFlag OutputterObj::reportFront(Sit S)
 
 eFlag OutputterObj::eventBeginOutput(Sit S)
 {
-    method = outDef ? outDef -> getMethod() : OUTPUT_UNKNOWN;
-    if (physical)
+  const EQName dummy;
+  pushLevel(dummy); //experimental
+  method = outDef ? outDef -> getMethod() : OUTPUT_UNKNOWN;
+  if (physical)
     {
-      if (method != OUTPUT_UNKNOWN)//??JP??both physical & method !=... duplicit with the same in reportXMLDeclIfMust 
-            E( reportXMLDeclIfMust(S) );
+      //??JP??both physical & method !=... 
+      //duplicit with the same in reportXMLDeclIfMust 
+      if (method != OUTPUT_UNKNOWN)
+	E( reportXMLDeclIfMust(S) );
     }
-    // initialize the SAX interface
-    IF_SAX1( startDocument ); 
-    state = STATE_OUTSIDE;
-    return OK;
+  // initialize the SAX interface
+  IF_SAX1( startDocument ); 
+  state = STATE_OUTSIDE;
+  return OK;
 }
 
 eFlag OutputterObj::eventBeginSubtree(Sit S)
 {
+  const EQName dummy;
+  pushLevel(dummy); //experimntal
     method = outDef -> getMethod();
     if (physical)
     {
@@ -1304,51 +1385,52 @@ eFlag OutputterObj::eventBeginSubtree(Sit S)
 
 eFlag OutputterObj::eventElementStart(Sit S, const EQName& name)
 {
-    if (noElementYet)
+  if (noElementYet)
     {
-        noElementYet = FALSE;
-        if (physical)
+      noElementYet = FALSE;
+      if (physical)
         {
-            if (method == OUTPUT_UNKNOWN)
+	  if (method == OUTPUT_UNKNOWN)
             {
-                if (name.getUri().isEmpty() && name.getLocal().eqNoCase("html"))
-                    method = OUTPUT_HTML;
-                else
-                    method = OUTPUT_XML;
-                E( physical -> setMethodByDefault(S, method) );
-                E( reportXMLDeclIfMust(S) );
-                E( reportFront(S) );//??JP??sax
+	      if (name.getUri().isEmpty() && name.getLocal().eqNoCase("html"))
+		method = OUTPUT_HTML;
+	      else
+		method = OUTPUT_XML;
+	      E( physical -> setMethodByDefault(S, method) );
+	      E( reportXMLDeclIfMust(S) );
+	      E( reportFront(S) );//??JP??sax
             };
-            // If this is an included stylesheet, output method cd be TEXT here
-            if (method != OUTPUT_TEXT) {
-	        E( reportDTDIfMust( S, name ) );//??JP??sax
-            }
+	  // If this is an included stylesheet, output method cd be TEXT here
+	  if (method != OUTPUT_TEXT) {
+	    E( reportDTDIfMust( S, name ) );//??JP??sax
+	  }
         };
     }
-    switch(state)
+  switch(state)
     {
     case STATE_OUTSIDE:
     case STATE_IN_MARKUP:
     case STATE_IN_ELEMENT:
-        {
-            // reportStartTag also sends the event to SAX
-            E( reportStartTag(S, NONEMPTY_ELEMENT) );
-            E( reportCurrData(S) );
-            pushLevel(name);
-            // this sets state to STATE_IN_MARKUP
-        };
-        break;
+      {
+	// reportStartTag also sends the event to SAX
+	E( reportStartTag(S, NONEMPTY_ELEMENT) );
+	E( reportCurrData(S) );
+	pushLevel(name);
+	// this sets state to STATE_IN_MARKUP
+      };
+      break;
     case STATE_IN_COMMENT:
     case STATE_IN_PI:
-        Err( S, E_ELEM_IN_COMMENT_PI );
+    case STATE_IN_ATTRIBUTE:
+      Err( S, E_ELEM_IN_COMMENT_PI );
     default:
-        assert(!"eventElementStart");
+      assert(!"eventElementStart");
     };
-
-    // determine whether we're in a HEAD html element so a META tag
-    // needs to be added
-
-    return OK;
+  
+  // determine whether we're in a HEAD html element so a META tag
+  // needs to be added
+  
+  return OK;
 }
 
 eFlag OutputterObj::throwInMeta(Sit S)
@@ -1406,75 +1488,84 @@ eFlag OutputterObj::throwInMeta(Sit S)
 
 eFlag OutputterObj::eventElementEnd(Sit S, const EQName& name)
 {
-    Str temp;
-    if (S.getProcessor())
-      temp = S.getProcessor() -> getAliasedName(name, 
-					      currNamespaces);
-    else name.getname(temp);
-
-    switch(state)
+  assert(!(physical && mySAXOutputType == SAXOUTPUT_INT_PHYSICAL));
+  Str physName;
+  //  if (S.getProcessor())
+  //  temp = S.getProcessor() -> getAliasedName(name, 
+  //					      currNamespaces,
+  //					      mySAXOutputType == SAXOUTPUT_INT_PHYSICAL);
+  //else name.getname(temp);
+  
+  switch(state)
     {
     case STATE_IN_MARKUP:
-        E( reportStartTag(S, EMPTY_ELEMENT) ); 
-        break;
+      E( reportStartTag(S, EMPTY_ELEMENT) ); 
+      break;
     case STATE_IN_ELEMENT:
-        {
-            E( reportCurrData(S) );
-            if (physical)
-            {
-                physical -> outputElementEnd(S, temp, NONEMPTY_ELEMENT );
-            };
-        }; break;
+      {
+	E( reportCurrData(S) );
+	if (physical)
+	  {
+	    Str physName;
+	    name.getname(physName);
+	    physical -> outputElementEnd(S, physName, NONEMPTY_ELEMENT );
+	  };
+      }; break;
     default:
-        assert(!"eventElementEnd");
+      assert(!"eventElementEnd");
     };
-
-    // send the event to SAX
-    switch(mySAXOutputType)
+  
+  // send the event to SAX
+  switch(mySAXOutputType)
     {
     case SAXOUTPUT_COPY_TREE:
-        {
+    case SAXOUTPUT_INT_PHYSICAL:
+      {
         GP( Str ) theSAXname = nameForSAX(S, name);
 	IF_SAX2( endElement, (const char*) *theSAXname );
 	theSAXname.del();
 	// report namespace scope ends
-	while (currNamespaces.number() > getFirstOwnNS())
+	//while (currNamespaces.number() > getFirstOwnNS())
+	for (int ii = currNamespaces.number() - 1; ii >= getFirstOwnNS(); ii--)
 	  {
-	    IF_SAX2( endNamespace, currNamespaces.last() -> prefix );
-	    currNamespaces.freelast(FALSE);
+	    IF_SAX2( endNamespace, currNamespaces[ii] -> prefix );
+	    //debug
 	  }
-	};
-	break;
+      };
+      break;
     case SAXOUTPUT_AS_PHYSICAL:
-        {
-        IF_SAX2( endElement, (const char*) temp );
+      {
+	Str physName;
+	name.getname(physName);
+        IF_SAX2( endElement, (const char*) physName );
 	// report namespace scope ends
 	Str* prefix;
-	while (currNamespaces.number() > getFirstOwnNS())
+	//while (currNamespaces.number() > getFirstOwnNS())
+	for (int ii = currNamespaces.number() - 1; ii >= getFirstOwnNS(); ii--)
 	  {
-	    prefix = &(currNamespaces.last() -> prefix);
+	    prefix = &(currNamespaces[ii] -> prefix);
 	    if (! currNamespaces.isHidden(*prefix))
-	      IF_SAX2( endNamespace, currNamespaces.last() -> prefix );
-	    currNamespaces.freelast(FALSE);
+	      IF_SAX2( endNamespace, currNamespaces[ii] -> prefix );
 	  }
-	};
-        break;
+      };
+      break;
+    case SAXOUTPUT_NONE: break;
     default:
-        assert(!"eventElementEnd");
+      assert(!"eventElementEnd");
     };
+  
+  //remove our namespaces
+  while (currNamespaces.number() > getFirstOwnNS())
+    currNamespaces.freelast(FALSE);
 
-    while (excludedNS.number() > getFirstOwnExNS()) {
-      excludedNS.deppend();
-    }
-
-    // pop one history level
-    history.freelast(FALSE);
-    if (getLevel())
-        state = STATE_IN_ELEMENT;
-    else
-        state = STATE_OUTSIDE;
-
-    return OK;
+  // pop one history level
+  history.freelast(FALSE);
+  if (getLevel()) //experimental
+    state = STATE_IN_ELEMENT;
+  else
+    state = STATE_OUTSIDE;
+  
+  return OK;
 }
 
 
@@ -1500,7 +1591,8 @@ eFlag OutputterObj::eventAttributeStart(Sit S, const EQName& name)
 	    };
         break;
     default:
-        assert(!"eventAttributeStart");
+      Err1(S, E1_ATTRIBUTE_MISPLACED, fullName);
+      //assert(!"eventAttributeStart");
     };
     return OK;
 }
@@ -1552,7 +1644,7 @@ eFlag OutputterObj::eventCommentEnd(Sit S)
         IF_SAX2( comment, currData );
     }
     currData.empty();
-    state = (getLevel()? STATE_IN_ELEMENT : STATE_OUTSIDE);
+    state = (getLevel() ? STATE_IN_ELEMENT : STATE_OUTSIDE); //experimental
     return OK;
 }
 
@@ -1593,22 +1685,12 @@ eFlag OutputterObj::eventPIEnd(Sit S)
     }
     currData.empty();
     currPIName.empty();
-    state = (getLevel()? STATE_IN_ELEMENT : STATE_OUTSIDE);
+    state = (getLevel() ? STATE_IN_ELEMENT : STATE_OUTSIDE);
     return OK;
 }
 
-eFlag OutputterObj::eventExcludeNS(Sit S, UriList &exNS) {
-  //append all prefixes to be excluded
-
-  //!!! the outputter status should be asserted here
-
-  for (int i = 0; i < exNS.number(); i++) {
-    excludedNS.append(exNS[i]);
-  }
-  return OK;
-}
-
-eFlag OutputterObj::eventNamespace(Sit S, const Str& prefix, const Str& uri)
+eFlag OutputterObj::eventNamespace(Sit S, const Str& prefix, const Str& uri,
+				   Bool hidden)
 {
     assert(state == STATE_IN_MARKUP);
     // GP: OK, not allocated
@@ -1621,21 +1703,48 @@ eFlag OutputterObj::eventNamespace(Sit S, const Str& prefix, const Str& uri)
     //namespaces with the same prefix may be added (e.g. for
     //xsl:element insrtuction) in this case (existing_ndx >=
     //getFirstOwnNS()) we need override former definition in place.
+
+    /*
+    if (uri == theXSLTNamespace)
+      {
+	//we need to reveal the xslt namespace if it is aliased
+	Processor *proc = S.getProcessor();
+	if (proc)
+	  { 
+	    Bool aliased;
+	    E( proc -> prefixIsAliasTarget(S, prefix, aliased) );
+	    hidden = hidden && ! aliased;
+	  }
+      }
+    */
+
     if (!existing_namespace)
-      currNamespaces.appendConstruct(prefix, uri, nsExcluded(search));
+      {
+	//if the namespace doesn't exist yet, we add it
+	currNamespaces.appendConstruct(prefix, uri, hidden);
+      }
     else
-      // the following condition was asserted here for some reason
-      if (!(*existing_namespace == uri))
-	{
-	  if (existing_ndx >= getFirstOwnNS())
-	    {
-	      currNamespaces[existing_ndx] -> uri = uri;
-	    }
-	  else
-	    {
-	      currNamespaces.appendConstruct(prefix, uri, nsExcluded(search));
-	    }
-	}
+      {
+	Bool exIsHidden = currNamespaces[existing_ndx]->hidden;
+	if (!(*existing_namespace == uri))
+	  {
+	    if (existing_ndx >= getFirstOwnNS())
+	      {
+		currNamespaces[existing_ndx] -> uri = uri;
+		currNamespaces[existing_ndx] -> hidden = hidden;
+	      }
+	    else
+	      {
+		currNamespaces.appendConstruct(prefix, uri, hidden);
+	      }
+	  }
+	//may be we found excluded one, but in current subtree
+	//it sould be kept
+	else if (exIsHidden && !hidden)
+	  {
+	    currNamespaces.appendConstruct(prefix, uri, hidden);
+	  }
+      }
     return OK;
 }
 
@@ -1648,9 +1757,9 @@ eFlag OutputterObj::eventDisableEscapingForNext(Sit S)
         case STATE_IN_ATTRIBUTE:
         case STATE_IN_PI:
         case STATE_IN_COMMENT:
-            Warn(S, W_DISABLE_OUTPUT_ESC);
+	  Warn(S, W_DISABLE_OUTPUT_ESC); break;
         default:
-            outputEscaping = FALSE;
+	  outputEscaping = FALSE;
         };
     }
     return OK;
@@ -1684,7 +1793,8 @@ eFlag OutputterObj::eventData(Sit S, const Str& data, Bool hardCData /* = FALSE
                 int htmlScriptOrStyle = ( getFlags() & HTML_SCRIPT_OR_STYLE );
 
                 if (physical)
-                    E( physical -> outputText(S, data, !outputEscaping, htmlScriptOrStyle) );
+                    E( physical -> outputText(S, data, !outputEscaping, 
+					      htmlScriptOrStyle) );
             }
             // reset outputEscaping to default after use
             // even if we just save a part of cdata section
@@ -1723,7 +1833,7 @@ eFlag OutputterObj::eventCDataSection(Sit S, const Str& data)
     return OK;
 }
 
-eFlag OutputterObj::eventEndOutput(Sit S)
+eFlag OutputterObj::eventEndOutput(Sit S, Bool closePhysical /* = false */)
 {
     assert(state == STATE_OUTSIDE);
     E( reportCurrData(S) );
@@ -1737,6 +1847,15 @@ eFlag OutputterObj::eventEndOutput(Sit S)
     IF_SAX1( endDocument );
     state = STATE_DONE;
 
+    //pop fake history item
+    history.freelast(FALSE);
+
+    //close physical (used for multiple doc output)
+    if (physical && closePhysical) 
+      {
+	E( physical -> close(S) );
+      }
+    
     return OK;
 }
 
@@ -1788,130 +1907,151 @@ eFlag OutputterObj::reportCurrData(Sit S, Bool hardCData /* = FALSE */)
 
 eFlag OutputterObj::reportStartTag(Sit S, Bool isEmpty)
 {
-    // FIXME: couldn't improve reportStartTag from GP point of view?
-    if (state == STATE_OUTSIDE || currElement.isEmpty())
-        return OK;
+  assert(!(physical && mySAXOutputType == SAXOUTPUT_INT_PHYSICAL));
+  // FIXME: couldn't improve reportStartTag from GP point of view?
+  if (state == STATE_OUTSIDE || currElement.isEmpty())
+    return OK;
+  
+  int doThrowMeta = method == OUTPUT_HTML && noHeadYet && 
+    currElement.getUri().isEmpty() && currElement.getLocal().eqNoCase("head");
+  int i;
+  //get attr aliased names
+  StrStrList attAliased;
+  // the temporary string is a workaround for VisualAge on AIX
+  //Str temp;
 
-    int doThrowMeta = method == OUTPUT_HTML && noHeadYet && 
-      currElement.getUri().isEmpty() && currElement.getLocal().eqNoCase("head");
-    int i;
-    //get attr aliased names
-    StrStrList attAliased;
-    // the temporary string is a workaround for VisualAge on AIX
-    Str temp;
 
-    if (physical || (mySAXHandler && mySAXOutputType == SAXOUTPUT_AS_PHYSICAL))
-      {
-	if (S.getProcessor())
-	  temp = S.getProcessor() -> getAliasedName(currElement, currNamespaces);
-	else
-	  currElement.getname(temp);
+  Str physName;
+  Bool intPhys = mySAXOutputType == SAXOUTPUT_INT_PHYSICAL;
 
-	for (i = 0; i < currAtts.number(); i++)
+  if (physical || (mySAXHandler && (mySAXOutputType == SAXOUTPUT_AS_PHYSICAL)))
+    {
+      //if (S.getProcessor())
+      //temp = S.getProcessor() -> getAliasedName(currElement, 
+      //					  currNamespaces, intPhys);
+      //else
+      //currElement.getname(temp);
+      currElement.getname(physName);
+      
+      for (i = 0; i < currAtts.number(); i++)
+	{
+	  Str aliname;
+	  //if (S.getProcessor())
+	  //  aliname = S.getProcessor() -> getAliasedName(currAtts[i] -> key,
+	  //					 currNamespaces, 
+	  //					 intPhys);
+	  //else
+	  //  currAtts[i] -> key.getname(aliname); 
+	  currAtts[i] -> key.getname(aliname);
+	  attAliased.appendConstruct(aliname, currAtts[i] -> value);
+	}
+      
+      if (physical) {
+	if ( delayedDTD )
 	  {
-	    Str aliname;
-	    if (S.getProcessor())
-	      aliname = S.getProcessor() -> getAliasedName(currAtts[i] -> key,
-							   currNamespaces);
-	    else
-	      currAtts[i] -> key.getname(aliname); 
-	    attAliased.appendConstruct(aliname, currAtts[i] -> value);
+	    const Str& 
+	      DTSystem = outDef -> getValueStr(XSLA_DOCTYPE_SYSTEM),
+	      DTPublic = outDef -> getValueStr(XSLA_DOCTYPE_PUBLIC);        
+	    E( physical -> outputDTD( S, physName, DTPublic, DTSystem) );
+	    delayedDTD = FALSE;
 	  }
 
-	if (physical) {
-	  if ( delayedDTD )
-	    {
-	      const Str& 
-		DTSystem = outDef -> getValueStr(XSLA_DOCTYPE_SYSTEM),
-		DTPublic = outDef -> getValueStr(XSLA_DOCTYPE_PUBLIC);        
-	      E( physical -> outputDTD( S, temp, DTPublic, DTSystem) );
-	      delayedDTD = FALSE;
-	    }
-	  
-	  E( physical -> outputElementStart(S,
-					    temp, currNamespaces, 
-					    getFirstOwnNS(), attAliased, 
-					    isEmpty && !doThrowMeta) );
-	}
+	E( physical -> outputElementStart(S, physName, currNamespaces, 
+					  getFirstOwnNS(), attAliased, 
+					  isEmpty && !doThrowMeta) );
       }
-
-    if (mySAXHandler)
+    }
+  
+  if (mySAXHandler)
     {
-        GPA( ConstCharP ) attsTable = new const char*[(currAtts.number() * 2) + 1];
-        // GP: the allocations are OK: no exit routes
-
-        PList<Str*> stringsForDeletion;
-        attsTable[currAtts.number() * 2] = NULL;
-
-	switch(mySAXOutputType)
+      const char** attsTable;
+      GPA( ConstCharP ) attsArr = attsTable = new const char*[(currAtts.number() * 2) + 1];
+      // GP: the allocations are OK: no exit routes
+      
+      PList<Str*> stringsForDeletion;
+      attsTable[currAtts.number() * 2] = NULL;
+      
+      switch(mySAXOutputType)
+	{
+	case SAXOUTPUT_COPY_TREE:
+	case SAXOUTPUT_INT_PHYSICAL:
 	  {
-	  case SAXOUTPUT_COPY_TREE:
-	    {
-	      // report only the new namespaces
-	      for (i = getFirstOwnNS(); i < currNamespaces.number(); i++)
-		{
-		  IF_SAX3( startNamespace, currNamespaces[i] -> prefix, 
-			   currNamespaces[i] -> uri );
-		}
-	      // create the attribute table
-	      for (i = 0; i < currAtts.number(); i++)
-		{
-		  Str *expatName = nameForSAX(S, currAtts[i] -> key);
-		  stringsForDeletion.append(expatName);
-		  attsTable[i * 2] = *expatName;
-		  attsTable[(i * 2) + 1] = currAtts[i] -> value;
-		};
-	      Str* theSAXname = nameForSAX(S, currElement);
-	      stringsForDeletion.append(theSAXname);
-	      IF_SAX3(startElement, *theSAXname, attsTable);
-	      stringsForDeletion.freeall(FALSE);
-	    };
-	    break;
-	  case SAXOUTPUT_AS_PHYSICAL:
-	    // report only the new not hidden namespaces
+	    // report only the new namespaces
 	    Str* prefix;
 	    for (i = getFirstOwnNS(); i < currNamespaces.number(); i++)
 	      {
 		prefix = &(currNamespaces[i] -> prefix);
-		if (! currNamespaces.isHidden(*prefix)) 
+		if (intPhys)
+		  {
+		    IF_INT_SAX4( startNamespace2, currNamespaces[i] -> prefix, 
+				 currNamespaces[i] -> uri, 
+				 currNamespaces.isHidden(*prefix) );
+		  }
+		else
 		  IF_SAX3( startNamespace, currNamespaces[i] -> prefix, 
 			   currNamespaces[i] -> uri );
 	      }
 	    // create the attribute table
 	    for (i = 0; i < currAtts.number(); i++)
 	      {
-		attsTable[i * 2] = (char *)attAliased[i] -> key;
-		attsTable[(i * 2) + 1] = (char *)attAliased[i] -> value;
+		Str *expatName = nameForSAX(S, currAtts[i] -> key);
+		stringsForDeletion.append(expatName);
+		attsTable[i * 2] = *expatName;
+		attsTable[(i * 2) + 1] = currAtts[i] -> value;
 	      };
-	    
-	    IF_SAX3(startElement, temp, attsTable);
-	    break;
-	  default:
-	    assert(!"eventElementEnd");
-	  };
-	attsTable.delArray();	
-    }
-
-    if (physical || (mySAXHandler && mySAXOutputType == SAXOUTPUT_AS_PHYSICAL))
-      {
-	attAliased.freeall(FALSE);
-      }
-
-    eFlag result = OK;
-    if (doThrowMeta) {
-        result = throwInMeta(S);
-	//__PH__ close head if needed
-	if (isEmpty) {
-	  if (physical) {
-	    E( physical ->outputElementEnd(S, temp, FALSE) );
+	    Str* theSAXname = nameForSAX(S, currElement);
+	    stringsForDeletion.append(theSAXname);
+	    IF_SAX3(startElement, *theSAXname, attsTable);
+	    stringsForDeletion.freeall(FALSE);
 	  };
+	  break;
+	case SAXOUTPUT_AS_PHYSICAL:
+	  // report only the new not hidden namespaces
+	  Str* prefix;
+	  for (i = getFirstOwnNS(); i < currNamespaces.number(); i++)
+	    {
+	      prefix = &(currNamespaces[i] -> prefix);
+	      if (! currNamespaces.isHidden(*prefix)) 
+		IF_SAX3( startNamespace, currNamespaces[i] -> prefix, 
+			 currNamespaces[i] -> uri );
+	    }
+	  // create the attribute table
+	  for (i = 0; i < currAtts.number(); i++)
+	    {
+	      attsTable[i * 2] = (char *)attAliased[i] -> key;
+	      attsTable[(i * 2) + 1] = (char *)attAliased[i] -> value;
+	    };
+	  
+	  IF_SAX3(startElement, physName, attsTable);
+	  break;
+	case SAXOUTPUT_NONE:
+	  assert(!"saxoutput == NONE");
+	default:
+	  assert(!"eventElementEnd");
 	};
+      //attsTable.delArray();	
     }
-
-    currElement.empty();
-    currAtts.freeall(FALSE);
-    currData.empty();
-    return result;
+  
+  if (physical || (mySAXHandler && mySAXOutputType == SAXOUTPUT_AS_PHYSICAL))
+    {
+      attAliased.freeall(FALSE);
+    }
+  
+  eFlag result = OK;
+  if (doThrowMeta) {
+    result = throwInMeta(S);
+    //__PH__ close head if needed
+    if (isEmpty) {
+      if (physical) {
+	E( physical ->outputElementEnd(S, physName, FALSE) );
+      };
+    };
+  }
+  
+  currElement.empty();
+  currAtts.freeall(FALSE);
+  currData.empty();
+  return result;
 }
 
 void OutputterObj::pushLevel(const EQName& name)
@@ -1919,9 +2059,18 @@ void OutputterObj::pushLevel(const EQName& name)
     currElement = name;
     GP( OutputHistoryItem ) newItem = new OutputHistoryItem;
     if (history.number())
+      {
         *newItem = *(history.last());
+	//next line is speed optimalization (very few)
+	(*newItem).useDocument = (*(history.last())).document;
+      }
     else
+      {
         (*newItem).flags = 0;
+	(*newItem).useDocument = NULL;
+      }
+    //definition is used for multiple document output
+    (*newItem).document = NULL; //in any case
     if (physical)
     {
         if (outDef -> askEQNameList(XSLA_CDATA_SECT_ELEMS, name))
@@ -1941,7 +2090,6 @@ void OutputterObj::pushLevel(const EQName& name)
 	  }
     };
     (*newItem).firstOwnNS = currNamespaces.number();
-    (*newItem).firstOwnExNS = excludedNS.number();
     history.append(newItem.keep());
     state = STATE_IN_MARKUP;
 }
@@ -1951,7 +2099,18 @@ void OutputterObj::report(Sit S, MsgType type, MsgCode code, const Str& arg1, co
     S.message(type, code, arg1, arg2);
 }
 
-Bool OutputterObj::nsExcluded(Str & uri) {
-  return excludedNS.findNdx(uri) != -1;
+eFlag OutputterObj::setDocumentForLevel(Sit S, OutputDocument *doc)
+{
+  if (history.number())
+    (*(history.last())).document = doc;
+  return OK;
 }
 
+OutputDocument* OutputterObj::getDocumentForLevel(Bool forElement)
+{
+  if (history.number())
+    return forElement ? (*(history.last())).useDocument :
+    (*(history.last())).document;
+  else 
+    return NULL;
+}
