@@ -12,8 +12,8 @@
  * The Original Code is the Sablotron XSLT Processor.
  * 
  * The Initial Developer of the Original Code is Ginger Alliance Ltd.
- * Portions created by Ginger Alliance are Copyright (C) 2000 Ginger
- * Alliance Ltd. All Rights Reserved.
+ * Portions created by Ginger Alliance are Copyright (C) 2000-2002
+ * Ginger Alliance Ltd. All Rights Reserved.
  * 
  * Contributor(s): Tim Crook <tcrook@accelio.com>
  *                 Christian Lefebvre <christian.lefebvre@atosorigin.com>
@@ -50,7 +50,7 @@
 //
 //
 
-SAXHandler TreeConstructer::myHandlerRecord =
+SAXHandlerInternal TreeConstructer::myHandlerRecord =
 {
     saxStartDocument,
     saxStartElement,
@@ -60,7 +60,8 @@ SAXHandler TreeConstructer::myHandlerRecord =
     saxComment,
     saxPI,
     saxCharacters,
-    saxEndDocument
+    saxEndDocument,
+    saxStartNamespace2
 };
 
 TreeConstructer::TreeConstructer(Sit S_)
@@ -71,6 +72,7 @@ TreeConstructer::TreeConstructer(Sit S_)
     theDataLine = NULL;
     theLineNumber = 0;
     inSAXForAWhile = FALSE;
+    nsCount.append(0);
 }
 
 TreeConstructer::~TreeConstructer()
@@ -84,6 +86,8 @@ eFlag TreeConstructer::parseDataLineUsingGivenExpat(Sit S,
     theTree = t;
     theDataLine = d;
     theParser = theParser_;
+
+    S.setCurrFile(d -> fullUri);
     E( feedDocumentToParser(S, this) );
     E( t -> parseFinished(S) );
     if (t -> XSLTree)
@@ -97,7 +101,8 @@ eFlag TreeConstructer::parseDataLineUsingGivenExpat(Sit S,
 
 
 eFlag TreeConstructer::parseDataLineUsingExpat(Sit S, 
-    Tree *t, DataLine *d)
+					       Tree *t, DataLine *d, 
+					       char *base_ /*=NULL*/)
 {
     theParser = XML_ParserCreateNS(NULL, THE_NAMESPACE_SEPARATOR);
 
@@ -122,11 +127,17 @@ eFlag TreeConstructer::parseDataLineUsingExpat(Sit S,
     //     NULL);
 
     XML_SetExternalEntityRefHandler(theParser,
-        tcExternalEntityRef);
+				    tcExternalEntityRef);
+    XML_SetEntityDeclHandler(theParser, tcEntityDecl);
+
+
     XML_SetUserData(theParser, this);
     if (S.getProcessor())
-    XML_SetBase(theParser, S.getProcessor() -> findBaseURI( t -> getURI() ));
-	// wasn't this done before?
+      XML_SetBase(theParser, S.getProcessor() -> findBaseURI(S, t ->getURI()));
+    else if (base_)
+      XML_SetBase(theParser, base_);
+    
+    // wasn't this done before?
     XML_SetParamEntityParsing(theParser, XML_PARAM_ENTITY_PARSING_ALWAYS);
 
     //we want to get all info (including prefixes)
@@ -142,34 +153,47 @@ eFlag TreeConstructer::parseDataLineUsingExpat(Sit S,
     return OK;
 };
 
-eFlag TreeConstructer::parseUsingSAX(Sit S, Tree *t, OutputterObj& source)
+eFlag TreeConstructer::parseUsingSAX(Sit S, Tree *t, OutputterObj& source,
+				     SAXOutputType ot)
 {
     theTree = t;
     theDataLine = NULL;
     // register the handler with the outputter
-    E( source.setOptionsSAX(S, &myHandlerRecord, this, SAXOUTPUT_COPY_TREE) );
+    E( source.setOptionsSAX(S, (SAXHandler*)&myHandlerRecord, this, ot) );
     E( source.eventBeginOutput(S) );
     return OK;
 }
 
 eFlag TreeConstructer::parseUsingSAXForAWhile(Sit S, OutputterObj& source,
-					      Str &saxUri)
+					      Str &saxUri, 
+					      Bool resetNamespaces,
+					      Tree* srcTree,
+					      NSList &swallowNS)
 // it's assumed that this is a tree constructer over a dataline
 // (not a SAX one)
 {
     assert(theTree && theDataLine);
     inSAXForAWhile = TRUE;
     theSAXUri = saxUri;
+
+    if (resetNamespaces) 
+      E( theTree -> pushPendingNS(S, srcTree, swallowNS) );
+
     // register the handler with the outputter
-    E( source.setOptionsSAX(S, &myHandlerRecord, this, SAXOUTPUT_COPY_TREE) );
+    E( source.setOptionsSAX(S, (SAXHandler*)&myHandlerRecord, this, SAXOUTPUT_COPY_TREE) );
     E( source.eventBeginOutput(S) );
     return OK;
 }
 
-eFlag TreeConstructer::parseUsingSAXForAWhileDone(Sit S, OutputterObj& source)
+eFlag TreeConstructer::parseUsingSAXForAWhileDone(Sit S, OutputterObj& source,
+						  Bool resetNamespaces)
 {
     E( source.eventEndOutput(S) );
     inSAXForAWhile = FALSE;
+
+    if (resetNamespaces) 
+      E( theTree -> popPendingNS(S) );
+
     return OK;
 }
 
@@ -222,31 +246,36 @@ eFlag TreeConstructer::getDocEncoding(Sit S, const char *buf, Str& theEncoding,
         {
 	// look for "<?xml " to distinguish from "<?xml-stylesheet"
 	// or any other similar processing instructions
+	  Bool found = FALSE;
 	  if (buf[0] == '<' && buf[1] == '?' && buf[2] == 'x' 
 	      && buf[3] == 'm' && buf[4] == 'l' && buf[5] == ' ')
             {
-                const char *p = buf;
-                p = strpbrk(p + 2,"=?");
-                if (p && *p == '=')
-                {
-                    // found '=' in version, search on
-                    p = strpbrk(p + 1, "=?");
-                    if (p && *p == '=')
-                    {
-                        p++;
-                        skipWhite(p);
-
-                        const char *q = strpbrk(p + 1, "?\'\"");
-                        if (q && *q != '?' && *q == *p)
-                        {
-                            theEncoding.nset(p + 1, q - p - 1);
-                            break;
-                        }
-                    }
-                }
+	      const char *p = buf;
+	      p = strpbrk(p + 2,"=?");
+	      while (! found && p && *p == '=')
+		{
+		  //now find the pseudo-attribute name
+		  const char *beg = p - 1;
+		  while ( isWhite(*beg) ) beg--;
+		  while ( (!isWhite(*beg)) && *beg != '?') beg--;
+		  if ( strncmp(beg + 1, "encoding", 8) )
+		    p = strpbrk(p + 1, "=?");
+		  else
+		    {
+		      p++;
+		      skipWhite(p);
+		  
+		      const char *q = strpbrk(p + 1, "?\'\"");
+		      if (q && *q != '?' && *q == *p)
+			{
+			  theEncoding.nset(p + 1, q - p - 1);
+			  found = TRUE;
+			}
+		    }
+		}
             }
-            theEncoding = "UTF-8";
-        };
+	  if (! found) theEncoding = "UTF-8";
+	};
     }  
     return OK;
 }
@@ -260,8 +289,9 @@ eFlag TreeConstructer::feedDocumentToParser(Sit S, void* constructer)
     //char convBuffer[PARSE_CONV_BUFSIZE];
     //char rawBuffer[PARSE_BUFSIZE + 1];
 
-    GPA( GChar ) convBuffer = new char[PARSE_CONV_BUFSIZE];
-    GPA( GChar ) rawBuffer = new char[PARSE_BUFSIZE + 1];
+    char *convBuffer, *rawBuffer;
+    GPA( GChar ) convAux = convBuffer = new char[PARSE_CONV_BUFSIZE];
+    GPA( GChar ) rawAux = rawBuffer = new char[PARSE_BUFSIZE + 1];
 
     char *outbuf = (char *) convBuffer;
     const char *inbuf = (char *) rawBuffer;
@@ -400,9 +430,9 @@ eFlag TreeConstructer::feedDocumentToParser(Sit S, void* constructer)
         if (!(firstTime && rawBuffer[0] == (char) 0xEF && 
             rawBuffer[1] == (char) 0xBB && rawBuffer[2] == (char) 0xBF))
             S.setCurrLine(XML_GetCurrentLineNumber(parser));
-        int code = XML_GetErrorCode(parser); 
+        XML_Error code = XML_GetErrorCode(parser); 
         Str eCodeStr, eNameStr;
-        eCodeStr = code;
+        eCodeStr = (int)code;
         eNameStr = (char*) XML_ErrorString(code);
         // XML_ParserFree(parser); -- done later
         Err2T(this_, S, E_XML, eCodeStr, eNameStr);
@@ -423,6 +453,12 @@ eFlag TreeConstructer::feedDocumentToParser(Sit S, void* constructer)
 void TreeConstructer::tcStartDocument(
     void* constructer)
 {
+    TreeConstructer *this_ =
+        (TreeConstructer*) constructer;
+    if (this_ -> S.isError())
+        return;
+    // new slot for count of pending namespaces
+    this_ -> nsCount.append(0);
 };
 
 
@@ -453,44 +489,107 @@ void TreeConstructer::tcStartElement(
     XSL_OP opCode;
     XSL_ATT attCode;
     BOOL itsXSL = FALSE;
+    BOOL itsExtension = FALSE;
     Vertex *v;  // GP: OK (goes to tree)
     Attribute *a;   // GP: OK (goes to tree)
-    QName q;
+    QName el;
     int elemLine;
+    int xmlSpace = -1;
+    AttsCache auxAtts;
 
     if (this_ -> S.isError())
         return;
 
     this_ -> S.setCurrLine(elemLine = this_ -> getCurrentLineNumber());
 
-    if (setQNameFromExpat(this_ -> S, this_, q, elName))
+    //get elName to QName q - DO NOT CHANGE LATER!!
+    if (setQNameFromExpat(this_ -> S, this_, el, elName))
         return;
 
-	// DETACH: optimize this comparison ###
-    if ((t -> XSLTree) && (t -> expand(q.getUri()) == theXSLTNamespace)) //t -> stdPhrase(PHRASE_XSL_NAMESPACE)))
-    {
-        itsXSL = TRUE;
-        opCode = (XSL_OP) lookup((char*)(t -> expand(q.getLocal())),xslOpNames);
-        if (opCode == XSL_NONE)
-        {
-	  if (this_ -> theDataLine) {
-	    if (this_ -> inSAXForAWhile)
-	      this_ -> S.setCurrFile(this_ -> theSAXUri);
-	    else
-	      this_ -> S.setCurrFile(this_ -> theDataLine -> fullUri);
-	  }
-	  this_ -> S.message(MT_ERROR, ET_BAD_XSL, (char*)NULL, (char*)NULL);
-	  return;
-        };
-        v = new(&(t -> getArena())) XSLElement(*t, q, opCode);
-    }
-    else if ((t -> XSLTree) && t -> isExtensionUri(q.getUri())) { 
-      //extension element
-      v = new(&(t -> getArena())) ExtensionElement(*t, q);
-    }
-    else
-        v = new(&(t -> getArena())) Element(*t, q);
+    itsXSL = (t -> XSLTree) && (t -> expand(el.getUri()) == theXSLTNamespace);
 
+    //first we have to create all attributes, we need them for extension
+    //and excluded namespace definitions
+    t -> markNamespacePrefixes(this_ -> S); //IMPORTANT
+
+    while(*p)
+    {
+      QName q;
+      if (setQNameFromExpat(this_ -> S, this_, q, (char *)p[0]))
+	return;
+      
+      //check xml attributes (currently xml:space)
+      if (q.getUri() == t -> stdPhrase(PHRASE_XML_NAMESPACE) && 
+	  t -> expand(q.getLocal()) == (const char*)"space")
+	{
+	  if (!strcmp(p[1], "preserve"))
+	    xmlSpace = 1;
+	  else
+	    xmlSpace = 0;
+	}
+      
+      //xsl attributes may occure in literal ements!!
+      //attCode = (itsXSL ? 
+      attCode = (itsXSL || (t -> expand(q.getUri()) == theXSLTNamespace) ?
+		 (XSL_ATT) lookup(t -> expand(q.getLocal()),xslAttNames) : 
+		 XSLA_NONE);
+      //check duplicities
+      if (auxAtts.find(q))
+	{
+	  Str name;
+	  t -> expandQStr(q, name);
+	  this_ -> S.message(MT_ERROR, E_DUPLICIT_ATTRIBUTE, 
+			     (char*)name, (char*)NULL);
+	}
+      a = new(&(t -> getArena())) Attribute(*t, q, p[1],attCode);
+      a -> lineno = this_ -> getCurrentLineNumber();
+      auxAtts.append(a);
+
+      //process extension and excluded prefixes
+      if (attCode == XSLA_EXT_ELEM_PREFIXES || 
+	  attCode == XSLA_EXCL_RES_PREFIXES)
+	{
+	  t -> pushNamespacePrefixes(this_ -> S, a -> cont, attCode);
+	}
+      //move on
+      p += 2;
+    };
+
+    // DETACH: optimize this comparison ###
+    if (itsXSL) 
+      //t -> stdPhrase(PHRASE_XSL_NAMESPACE)))
+      {
+        opCode = (XSL_OP) lookup((char*)(t->expand(el.getLocal())),xslOpNames);
+        if (opCode == XSL_NONE)
+	  {
+	    if (this_ -> theDataLine) {
+	      if (this_ -> inSAXForAWhile)
+		this_ -> S.setCurrFile(this_ -> theSAXUri);
+	      else
+		this_ -> S.setCurrFile(this_ -> theDataLine -> fullUri);
+	    }
+	    this_ -> S.message(MT_ERROR, ET_BAD_XSL, (char*)NULL, (char*)NULL);
+	    return;
+	  };
+        v = new(&(t -> getArena())) XSLElement(*t, el, opCode);
+      }
+    else if ((t->XSLTree) && t -> isExtensionUri(el.getUri())) 
+      { 
+	//extension element
+	v = new(&(t -> getArena())) ExtensionElement(*t, el);
+	itsExtension = TRUE;
+      }
+    else
+      {
+        v = new(&(t -> getArena())) Element(*t, el);
+	Processor *proc = this_->S.getProcessor();
+	if ( proc && proc->outputter() ) 
+	  {
+	    OutputDocument *doc = proc->outputter()->getDocumentForLevel(TRUE);
+	    toE(v)->setOutputDocument(doc);
+	  }
+      }
+    
     //check whether it is top level foreign
     if (t -> XSLTree) 
       {
@@ -503,60 +602,76 @@ void TreeConstructer::tcStartElement(
 	    v -> vt = (VTYPE)(v -> vt | VT_TOP_FOREIGN);
 	  }
       }
-    
+
     v -> lineno = elemLine;
     t -> appendVertex(this_ -> S, v);
-    t -> pendingNS.giveCurrent(this_ -> S, toE(v) -> namespaces, t);
+
+    //if we are looking at the document element we need to check,
+    // whether the default namespace was declared
+    if (isRoot(v -> parent) && t -> pendingNS().findNdx(UNDEF_PHRASE) == -1)
+      tcStartNamespace(constructer, NULL, ""); //empty default namespace
+
+    t -> pendingNS().giveCurrent(this_ -> S, toE(v) -> namespaces, t,
+				 this_ -> nsCount.last());
 #ifndef EXPAT_SUPPORTS_TRIPLETS
     toE(v) -> namespaces.findPrefix(toE(v) -> name);
 #endif
       //_dump(t, toE(v) -> name);
 
-    while(*p)
-    {
-        if (setQNameFromExpat(this_ -> S, this_, q, (char *)p[0]))
-            return;
-	
-	/*
-	if (q.getUri() == UNDEF_PHRASE)
-	  q.setUri(v -> getName().getUri());
-	*/
+    toE(v) -> namespaces.incPrefixUsage(toE(v) -> name.prefix);
 
+    //add attributes to the element
+    for (int n = 0; n < auxAtts.number(); n++)
+      {
+	const QName *auxName = &(auxAtts[n] -> getName());
 #ifndef EXPAT_SUPPORTS_TRIPLETS
-	toE(v) -> namespaces.findPrefix(q);
+	//force the prefix!!
+	toE(v) -> namespaces.findPrefix(*(const_cast<QName*>(auxName)));
 #endif
-	//xsl attributes may occure in literal ements!!
-        //attCode = (itsXSL ? 
-	attCode = (itsXSL || (t -> expand(q.getUri()) == theXSLTNamespace) ?
-            (XSL_ATT) lookup(t -> expand(q.getLocal()),xslAttNames) : XSLA_NONE);
-        a = new(&(t -> getArena())) Attribute(*t, q, p[1],attCode);
-        a -> lineno = this_ -> getCurrentLineNumber();
-        t -> appendVertex(this_ -> S, a);
-        p += 2;
-	//_dump(t, q);
-    };
-    
-    if (itsXSL)
-    {
-        t -> extensionNamespaces(this_ -> S, toE(v));
-        if (toX(v) -> checkAtts(this_ -> S)) return; 
-        // also check if a toplevel element does not have a non-stylesheet parent
-        if (toX(v) -> checkToplevel(this_ -> S)) return; 
-    }
-    else
-    {
-        if (t -> XSLTree)
-        {
-            int k, 
-                kLimit = toE(v) -> atts.number();
-            for (k = 0; k < kLimit; k++)
-                if (toA(toE(v) -> atts[k]) -> buildExpr(this_ -> S,
-                    TRUE, EX_NONE)) return;
-            // this calls situation.error() on error
-        }
-    }
-}
+	t -> appendVertex(this_ -> S, auxAtts[n]);
+	if (auxName -> getPrefix() != UNDEF_PHRASE)
+	  toE(v) -> namespaces.incPrefixUsage(auxName -> getPrefix());
+      }
+    auxAtts.keepThem();
 
+     //process xml space
+    if (xmlSpace == -1)
+      {
+ 	Bool val = this_ -> preserveSpace.number() ? 
+ 	  this_ -> preserveSpace.last() : FALSE;
+ 	this_ -> preserveSpace.append(val);
+      }
+    else
+      this_ -> preserveSpace.append(xmlSpace);
+    toE(v) -> preserveSpace = this_ -> preserveSpace.last();
+    //xml-space end
+
+    if (itsXSL)
+      {
+        //t -> extensionNamespaces(this_ -> S, toE(v));
+        if (toX(v) -> checkAtts(this_ -> S)) return; 
+        // also check if a toplevel element does not 
+	//have a non-stylesheet parent
+        if (toX(v) -> checkToplevel(this_ -> S)) return; 
+      }
+    else if (itsExtension)
+      {
+        if (toExtension(v) -> checkAtts(this_ -> S)) return; 
+      }
+    else
+      {
+        if (t -> XSLTree)
+	  {
+	    int k;
+	    int kLimit = toE(v) -> atts.number();
+	    for (k = 0; k < kLimit; k++)
+	      if (toA(toE(v) -> atts[k]) -> buildExpr(this_ -> S,
+						      TRUE, EX_NONE)) return;
+	    // this calls situation.error() on error
+        }
+      }
+    this_ -> nsCount.append(0);
+}
 
 //
 //  tcEndElement
@@ -572,6 +687,8 @@ void TreeConstructer::tcEndElement(
     if (this_ -> S.isError())
         return;
     Tree *t = this_ -> theTree;
+    // destroyed slot for count of pending namespaces
+    this_ -> nsCount.deppend();
 
     Vertex *v = NZ( t -> stackTop );
     t -> flushPendingText();
@@ -589,11 +706,18 @@ void TreeConstructer::tcEndElement(
     } 
     else if ( isExtension(v) ) 
       {
-        if (toX(v) -> checkChildren(this_ -> S))
+        if (toExtension(v) -> checkChildren(this_ -> S))
             return;
       }
 
-    t -> processVertexAfterParse(this_ -> S, v, this_);        
+    //IMPORTANT STUFF
+    t -> processVertexAfterParse(this_ -> S, v, this_);
+
+    //cleanup
+    t -> popNamespacePrefixes(this_ -> S);
+    
+    //pop preserver space
+    this_ -> preserveSpace.deppend();
 }
 
 
@@ -605,6 +729,12 @@ void TreeConstructer::tcEndElement(
 /* static */
 void TreeConstructer::tcStartNamespace(
     void* constructer, const char* prefix, const char* uri)
+{
+  tcStartNamespace2(constructer, prefix, uri, false);
+}
+
+void TreeConstructer::tcStartNamespace2(
+    void* constructer, const char* prefix, const char* uri, Bool hidden)
 {
     TreeConstructer *this_ =
         (TreeConstructer*) constructer;
@@ -614,31 +744,39 @@ void TreeConstructer::tcStartNamespace(
 
     Vertex *newv;   // GP: OK (goes to tree)
 
+    //XML_Parser parser = this_ -> theParser;
+
     Phrase prefixPh, uriPh;
-    if (prefix)
+    if (prefix && *prefix) //not empty/default prefix
+      {
         t -> dict().insert(prefix, prefixPh);
+      }
     else
+      {
         prefixPh = UNDEF_PHRASE;
+      }
     t -> dict().insert(uri, uriPh);
 
-    // t -> pendingNS.appendAndSetOrdinal(
-    t -> pendingNS.append(
-        newv = new(&(t -> getArena())) NmSpace(*t, prefixPh, uriPh));
-    newv -> lineno = this_ -> getCurrentLineNumber();
+    // t -> pendingNS().appendAndSetOrdinal(
+    t -> pendingNS().append(
+        newv = new(&(t -> getArena())) 
+	    NmSpace(*t, prefixPh, uriPh, hidden));
     
+    newv -> lineno = this_ -> getCurrentLineNumber();
+    // count of pending namespaces increased
+    this_ -> nsCount[this_ -> nsCount.number()-1]++;
+
     // warn on obsolete namespace
     if (uri && !strcmp(oldXSLTNamespace, uri)) /* _PH_ */
         this_ -> Warn1(this_ -> S, W1_OLD_NS_USED, (char*)uri)
     else
-    {
+      {
         if (prefix && !strcmp(prefix, "xsl") && 
 	      uri && strcmp(theXSLTNamespace, uri)) /* _PH_ */
 	  this_ -> Warn1(this_ -> S, W1_XSL_NOT_XSL, (char*) uri);
-    }
+      }
 };
 
-
-
 //
 //  tcEndNamespace
 //  callback for the namespace scope end event
@@ -657,13 +795,15 @@ void TreeConstructer::tcEndNamespace(
 #ifdef _DEBUG
     // hope this works
     Phrase prefixPh;
-    if (prefix)
+    if (prefix && *prefix)
         prefixPh = t -> dict().lookup(prefix);
     else
         prefixPh = UNDEF_PHRASE;
-    assert(toNS(t -> pendingNS.last()) -> prefix == prefixPh);
+    //assert(toNS(t -> pendingNS().last()) -> prefix == prefixPh);
 #endif
-    t -> pendingNS.freelast(FALSE);
+    t -> pendingNS().freelast(FALSE);
+    // count of pending namespaces decreased
+    this_ -> nsCount[this_ -> nsCount.number()-1]--;
 };
 
 
@@ -688,6 +828,14 @@ void TreeConstructer::tcComment(
     Comment *newNode;   // GP: OK
     newNode = new(&(t -> getArena())) Comment(*t, contents);
     newNode -> lineno = this_ -> getCurrentLineNumber();
+
+    Processor *proc = this_ -> S.getProcessor();
+    if ( proc && proc->outputter() ) 
+      {
+	OutputDocument *doc = proc->outputter()->getDocumentForLevel(FALSE);
+	newNode -> setOutputDocument(doc);
+      }
+
     t -> appendVertex(this_ -> S, newNode);
 };
 
@@ -717,6 +865,14 @@ void TreeConstructer::tcPI(
     newNode = new(&(t -> getArena()))
         ProcInstr(*t, targetPh, contents);
     newNode -> lineno = this_ -> getCurrentLineNumber();
+
+    Processor *proc = this_ -> S.getProcessor();
+    if ( proc && proc->outputter() ) 
+      {
+	OutputDocument *doc = proc->outputter()->getDocumentForLevel(FALSE);
+	newNode -> setOutputDocument(doc);
+      }
+    
     t -> appendVertex(this_ -> S, newNode);
 };
 
@@ -752,6 +908,12 @@ void TreeConstructer::tcCharacters(
 void TreeConstructer::tcEndDocument(
     void* constructer)
 {
+    TreeConstructer *this_ =
+        (TreeConstructer*) constructer;
+    if (this_ -> S.isError())
+        return;
+    // slot for namespaces counted for childs of this element destroyed 
+    this_ -> nsCount.deppend();
 };
 
 
@@ -802,8 +964,9 @@ int TreeConstructer::tcExternalEntityRef(
     TreeConstructer *this_ =
         (TreeConstructer*) XML_GetUserData(parser);
 
-    //PH
-    if ( publicId && ! this_ -> S.hasFlag(SAB_PARSE_PUBLIC_ENTITIES)) 
+    //ignore external entities at all or just public ones
+    if (this_ -> S.hasFlag(SAB_NO_EXTERNAL_ENTITIES) ||
+	(publicId && ! this_ -> S.hasFlag(SAB_PARSE_PUBLIC_ENTITIES)))
       return 1;
     
     if (this_ -> S.isError())
@@ -814,36 +977,62 @@ int TreeConstructer::tcExternalEntityRef(
     XML_Parser newParser =
         XML_ExternalEntityParserCreate(parser, context, /* encoding= */ NULL);
 
-    XML_SetBase(newParser, XML_GetBase(parser)); //Christian Lefebvre
-
     if (!newParser)
         return 0;
 
-    DataLine *newDL;
     Str absolute;
-    makeAbsoluteURI(systemId, base, absolute);
-    if (NZ(this_ -> S.getProcessor()) -> addLineNoTree(this_ -> S, newDL, absolute, t -> XSLTree) || !newDL)
-    {
+    makeAbsoluteURI(this_ -> S,systemId, base, absolute);
+    XML_SetBase(newParser, absolute); //Christian + PH
+
+    GP(DataLine) newDL = new DataLine;
+    if ( (*newDL).open(this_ -> S, absolute, DLMODE_READ, NULL) )
+      {
         XML_ParserFree(newParser);
         return 0;
-    }
+      }
+
+    /*
+    if (NZ(this_->S.getProcessor())->addLineNoTree(this_ -> S, newDL, absolute, 
+						   t -> XSLTree) || !newDL)
+      {
+        XML_ParserFree(newParser);
+        return 0;
+      }
+    */
 
     TreeConstructer *newTC;
     MT( this_, this_ -> S, newTC = new TreeConstructer(this_ -> S));
 
     // record subtree information for the external entity
     // XSL_NONE stands for 'external entity'
-    eFlag code = t -> startSubtree(this_ -> S, absolute, XSL_NONE);
+    eFlag code = t -> startSubtree(this_ -> S, absolute, XSL_NONE, TRUE);
     if (!code)
     {
 	code = newTC -> parseDataLineUsingGivenExpat(this_ -> S, t, newDL, newParser);
-	t -> endSubtree(this_ -> S); 
+	t -> endSubtree(this_ -> S, XSL_NONE); 
     }
+    (*newDL).close(this_ -> S);
     XML_ParserFree(newParser);
     delete newTC;
     return code ? 0 : 1;
 }
 
+void TreeConstructer::tcEntityDecl(void *userData, const char *entityName,
+				   int is_parameter_entity, const char *value,
+				   int value_length, const char *base,
+				   const char *systemId, const char *publicId,
+				   const char *notationName)
+{
+  TreeConstructer *this_ = (TreeConstructer*) userData;
+  if (notationName)
+    {
+      Str name = entityName;
+      Str absolute;
+      makeAbsoluteURI(this_ -> S, systemId, base, absolute);
+      this_ -> theTree -> setUnparsedEntityUri( name, absolute );
+    }
+}
+
 /* methods for internal sax handler */
 void TreeConstructer::saxStartDocument(void* userData, 
 				       SablotHandle processor_)
@@ -872,6 +1061,14 @@ void TreeConstructer::saxStartNamespace(void* userData,
   tcStartNamespace(userData, prefix, uri);
 }
 
+void TreeConstructer::saxStartNamespace2(void* userData, 
+					 SablotHandle processor_,
+					 const char* prefix, const char* uri,
+					 Bool hidden)
+{
+  tcStartNamespace2(userData, prefix, uri, hidden);
+}
+
 void TreeConstructer::saxEndNamespace(void* constructer, 
 				      SablotHandle processor_, 
 				      const char* prefix)
@@ -906,7 +1103,7 @@ void TreeConstructer::saxEndDocument(void* constructer,
   tcEndDocument(constructer);
 }
 
-/* static */
+/* static MUST NOT RELY ON EXPAT_SUPPORTS_TRIPLETS*/
 eFlag TreeConstructer::setQNameFromExpat(Sit S, TreeConstructer* this_, QName& qname_, const char* text)
 {
     Tree *t = this_ -> theTree;
@@ -950,7 +1147,8 @@ eFlag TreeConstructer::setQNameFromExpat(Sit S, TreeConstructer* this_, QName& q
         {
             *isColon = 0;
 
-            // fixing what appears a bug in expat - sometimes the xml:lang attr arrives unexpanded
+            // fixing what appears a bug in expat 
+	    //- sometimes the xml:lang attr arrives unexpanded
             // apparently only in XSL and when it's not at the top level
             if (!strEqNoCase(p,"xml"))
                 Err1T(this_, S, ET_BAD_PREFIX,p)
