@@ -12,8 +12,8 @@
  * The Original Code is the Sablotron XSLT Processor.
  * 
  * The Initial Developer of the Original Code is Ginger Alliance Ltd.
- * Portions created by Ginger Alliance are Copyright (C) 2000 Ginger
- * Alliance Ltd. All Rights Reserved.
+ * Portions created by Ginger Alliance are Copyright (C) 2000-2002
+ * Ginger Alliance Ltd. All Rights Reserved.
  * 
  * Contributor(s):
  * 
@@ -54,6 +54,8 @@ i n c l u d e s
 #include "uri.h"
 #include "parser.h"
 #include "guard.h"
+#include "sxpath.h"
+#include "domprovider.h"
 
 //
 //  defines
@@ -132,6 +134,11 @@ int SablotSetOptions(SablotSituation S, int flags)
     return OK;
 }
 
+int SablotGetOptions(SablotSituation S)
+{
+  return ((Situation*)S) -> getFlags();
+}
+
 int SablotClearSituation(SablotSituation S)
 {
     ((Situation*)S) -> clear();
@@ -146,7 +153,13 @@ int SablotCreateDocument(SablotSituation S, SDOM_Document *D)
 {
     // non-XSL tree with empty name
     Tree *t = new Tree((char*)"", FALSE);
-    *D = &(t -> getRoot());
+    SabArena &ar = t -> getArena();
+    NmSpace *nm = new (&ar) NmSpace(*t, t -> unexpand("xml"), 
+				    t -> unexpand(theXMLNamespace),
+				    TRUE, NSKIND_DECLARED);
+    RootNode &root = t -> getRoot();
+    root.namespaces.append(nm);
+    *D = &root;
     return OK;
 }
 
@@ -166,12 +179,14 @@ int SablotParse_(
 
     SIT(S).clear();
     // make absolute address
+    char *parserBase = NULL;
     if (!buffer)
     {
         my_getcwd(base);
         base = Str("file://") + base;
         // should call Processor::findBaseURI() here (but have no Processor)
-        makeAbsoluteURI(uri, base, absolute);
+        makeAbsoluteURI(SIT(S), uri, base, absolute);
+	parserBase = (char*)absolute;
     }
     else
       absolute = "arg:/_parsed_";
@@ -189,7 +204,7 @@ int SablotParse_(
     EC( code, line.open(SIT(S), absolute, DLMODE_READ, &argList) );
     
     Log1(SIT(S), L1_PARSING, absolute);
-    EC( code, tc.parseDataLineUsingExpat(SIT(S), newTree, &line) );        
+    EC( code, tc.parseDataLineUsingExpat(SIT(S), newTree, &line, parserBase) );        
     EC( code, line.close(SIT(S)) );
     *D = &(newTree.keep() -> getRoot());    
     Log1(SIT(S), L1_PARSE_DONE, getMillisecsDiff(time_was));
@@ -239,8 +254,8 @@ int SablotLockDocument(SablotSituation S, SDOM_Document D)
 
 int SablotDestroyDocument(SablotSituation S, SDOM_Document D)
 {
-    delete &((toRoot(toV(D))) -> getOwner());
-    return OK;
+  delete &((toRoot(toV(D))) -> getOwner());
+  return OK;
 }
 
 
@@ -339,9 +354,9 @@ int SablotAddArgTree(
 int SablotRunProcessorGen(
     SablotSituation S,
     void *processor_,
-    char *sheetURI, 
-    char *inputURI, 
-    char *resultURI)
+    const char *sheetURI, 
+    const char *inputURI, 
+    const char *resultURI)
 {
     int code = 0;
 
@@ -373,6 +388,45 @@ int SablotRunProcessorGen(
     return code;
 }
 
+int SablotRunProcessorExt(
+    SablotSituation S,
+    void *processor_,
+    const char *sheetURI, 
+    const char *resultURI, 
+    NodeHandle doc)
+{
+    int code = 0;
+    //mask external node
+    doc = SXP_MASK_LEVEL(doc, SIT(S).getSXPMaskBit());
+
+    void *saveproc = processor_;
+    SIT(S).swapProcessor(saveproc);
+
+    SIT(S).clear();
+    // remove the existing 'arg' buffers
+    if (!(P(processor_) -> getAddedFlag()))
+        EC( code, SablotFreeResultArgs (processor_) );
+    if (!code)
+	P(processor_) -> prepareForRun();
+    // must first open the files, then use global params
+    // because VarsList needs stylesheet's dictionary
+    EC( code, P(processor_) -> open(SIT(S), sheetURI, NULL) );
+    EC( code, P(processor_) -> useGlobalParams(SIT(S)) );
+    EC( code, P(processor_) -> run(SIT(S), resultURI, doc) );
+    code = SIT(S).getError();
+    P(processor_) -> cleanupAfterRun((Situation*)S);
+    if (code)
+    {
+        // GP: if not all files could be opened, those that could are
+        // closed in cleanupAfterRun
+        // code = SIT(S).getError();
+        P(processor_) -> freeResultArgs(SIT(S));
+    };
+    
+    SIT(S).swapProcessor(saveproc);
+    return code;
+}
+
 
 //
 //
@@ -380,11 +434,11 @@ int SablotRunProcessorGen(
 //
 //
 int SablotRunProcessor(void *processor_,
-                       char *sheetURI, 
-                       char *inputURI, 
-                       char *resultURI,
-                       char **params, 
-                       char **arguments)
+                       const char *sheetURI, 
+                       const char *inputURI, 
+                       const char *resultURI,
+                       const char **params, 
+                       const char **arguments)
 {    
     Bool problem = FALSE;
     Sit S = *(NZ(P( processor_ )) -> recallSituation());
@@ -394,7 +448,7 @@ int SablotRunProcessor(void *processor_,
 	    S.clearError();
         E( SablotFreeResultArgs (processor_));
         P(processor_) -> prepareForRun();
-        char **p;
+        const char **p;
         if (arguments)
         {
             for (p = arguments; *p && !problem; p += 2)
@@ -493,7 +547,7 @@ int SablotSetBaseForScheme(void* processor_,
 //
 
 int SablotGetResultArg(void *processor_,
-                       char *argURI,
+                       const char *argURI,
                        char **argValue)
 {
     char *newCopy; // GP: OK
@@ -501,9 +555,11 @@ int SablotGetResultArg(void *processor_,
     int resultNdx_;
     if (argValue)
     {
-        P(processor_) -> copyArg(argURI, &resultNdx_, newCopy);
-        // if output does not go to a buffer, newCopy is NULL
-        *argValue = newCopy;
+      Situation *s = P(processor_) -> recallSituation();
+      assert(s); //we always should have some situation, master or internal
+      P(processor_) -> copyArg(SIT(s), argURI, &resultNdx_, newCopy);
+      // if output does not go to a buffer, newCopy is NULL
+      *argValue = newCopy;
     }
     return OK;
 }
@@ -518,7 +574,7 @@ int SablotGetResultArg(void *processor_,
 int SablotSetLog(void *processor_,
     const char *logFilename, int logLevel)
 {
-    P(processor_) -> recallSituation()->msgOutputFile((char *)"stderr",(char*) logFilename);
+    P(processor_) -> recallSituation()->msgOutputFile((char *)"/__stderr",(char*) logFilename);
     // logLevel ignored so far
     return OK;
 }
@@ -548,8 +604,10 @@ RETURNS
                 point to it (otherwise to NULL). Free it using SablotFree().
 *****************************************************************/
 
-int SablotProcess(char *sheetURI, char *inputURI, char *resultURI,
-    char **params, char **arguments, char **resultArg)
+int SablotProcess(const char *sheetURI, const char *inputURI, 
+		  const char *resultURI,
+		  const char **params, const char **arguments, 
+		  char **resultArg)
 {
     void *theproc;
     EE( SablotCreateProcessor (&theproc) );
@@ -577,9 +635,9 @@ RETURNS
 *****************************************************************/
 
 
-int SablotProcessFiles(char *styleSheetName,
-                       char *inputName,
-                       char *resultName)
+int SablotProcessFiles(const char *styleSheetName,
+                       const char *inputName,
+                       const char *resultName)
 {
     return SablotProcess(
         styleSheetName, inputName, resultName, 
@@ -601,20 +659,20 @@ RETURNS
   *resultStr    pointer to the newly allocated block containing
                 the result 
 *****************************************************************/
-int SablotProcessStrings(char *styleSheetStr,
-                         char *inputStr,
+int SablotProcessStrings(const char *styleSheetStr,
+                         const char *inputStr,
                          char **resultStr)
 {
-  char *argums[] =
+  const char *argums[] =
   {
-    (char *)"/_stylesheet", styleSheetStr,
-    (char *)"/_xmlinput", inputStr,
-    (char *)"/_output", NULL, //was: *resultStr,
+    "/_stylesheet", styleSheetStr,
+    "/_xmlinput", inputStr,
+    "/_output", NULL, //was: *resultStr,
     NULL
   };
-  return SablotProcess((char *) "arg:/_stylesheet",
-                       (char *) "arg:/_xmlinput",
-                       (char *) "arg:/_output",
+  return SablotProcess("arg:/_stylesheet",
+                       "arg:/_xmlinput",
+                       "arg:/_output",
                        NULL, argums, resultStr);
 };
 
@@ -636,28 +694,28 @@ RETURNS
                 the result 
 *****************************************************************/
 
-int SablotProcessStringsWithBase(char *styleSheetStr,
-                         char *inputStr,
+int SablotProcessStringsWithBase(const char *styleSheetStr,
+                         const char *inputStr,
                          char **resultStr,
-                         char *theHardBase)
+                         const char *theHardBase)
 {
-    char *argums[] =
+    const char *argums[] =
     {
-        (char *) "/_stylesheet", styleSheetStr,
-        (char *) "/_xmlinput", inputStr,
-        (char *) "/_output", NULL, //was: *resultStr,
+        "/_stylesheet", styleSheetStr,
+        "/_xmlinput", inputStr,
+        "/_output", NULL, //was: *resultStr,
         NULL
     };
     void *theproc;
     EE( SablotCreateProcessor(&theproc));
     EE_D( SablotSetBase(theproc, theHardBase), theproc );
     EE_D( SablotRunProcessor(theproc,
-        (char *) "arg:/_stylesheet",
-        (char *) "arg:/_xmlinput",
-        (char *) "arg:/_output",
+        "arg:/_stylesheet",
+        "arg:/_xmlinput",
+        "arg:/_output",
         NULL, argums), theproc);
     EE_D( SablotGetResultArg(theproc,
-        (char *) "arg:/_output", resultStr), theproc);
+        "arg:/_output", resultStr), theproc);
     EE( SablotDestroyProcessor(theproc));
     return OK;
 };
@@ -738,7 +796,7 @@ int SablotUnregHandler(void *processor_,
 //
 //
 
-const char *SablotGetMsgText(int code)
+const char* SablotGetMsgText(int code)
 {
     return GetMessage((MsgCode) code) -> text;
 }
@@ -772,7 +830,7 @@ void SablotSetInstanceData(void *processor_, void *idata)
 //
 //
 
-void *SablotGetInstanceData(void *processor_)
+void* SablotGetInstanceData(void *processor_)
 {
     return P(processor_) -> getInstanceData();
 }
@@ -791,27 +849,3 @@ void SablotSetEncoding(void *processor_, char *encoding_)
     P(processor_) -> setHardEncoding(
         encoding_ ? encoding_ : (const char*) "");
 }
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
