@@ -12,8 +12,8 @@
  * The Original Code is the Sablotron XSLT Processor.
  * 
  * The Initial Developer of the Original Code is Ginger Alliance Ltd.
- * Portions created by Ginger Alliance are Copyright (C) 2000 Ginger
- * Alliance Ltd. All Rights Reserved.
+ * Portions created by Ginger Alliance are Copyright (C) 2000-2002
+ * Ginger Alliance Ltd. All Rights Reserved.
  * 
  * Contributor(s):
  * 
@@ -35,6 +35,91 @@
 #include "guard.h"
 #include "key.h"
 
+/* ------------------------------------------------------------ */
+/* SpaceNameList */
+/* -------------------------------------------------------------*/
+
+Bool SpaceNameList::findName(EQName &ename, double &prio)
+{
+  int num = number();
+  prio = -10;
+  Bool found = 0;
+  for (int i = 0; i < num; i++)
+    {
+      EQName *aux = operator[](i);
+      if (aux -> getLocal() == (const char*) "*")
+	{
+	  if (aux -> getUri() == (const char*) "")
+	    {
+	      found = 1;
+	      prio = -0.5;
+	    }
+	  else
+	    if (ename.getUri() == aux -> getUri())
+	      {
+		found = 1;
+		prio = -0.25;
+	      }
+	}
+      else
+	{
+	  if (ename.getLocal() == aux -> getLocal() && 
+	      ename.getUri() == aux -> getUri())
+	    {
+	      found = 1;
+	      prio = 0; 
+	      break; //exit loop, we can't get better match
+	    }
+	}
+    }
+  return found;
+}
+
+/* ------------------------------------------------------------ */
+/* StylesheetStructure */
+/* -------------------------------------------------------------*/
+
+//
+//
+//  SubtreeInfo
+//
+//
+
+SubtreeInfo::~SubtreeInfo()
+{
+  excludedNS.deppendall();
+  extensionNS.deppendall();
+  excludedCount.deppendall();
+  extensionCount.deppendall();
+}
+
+void SubtreeInfo::pushNamespaceMarks()
+{
+  excludedCount.append(excludedNS.number());
+  extensionCount.append(extensionNS.number());
+}
+
+void SubtreeInfo::popNamespaceMarks()
+{
+  //excluded
+  int limit = excludedCount.number() ? excludedCount.last() : 0;
+  int idx;
+  for (idx = excludedNS.number() - 1; idx >= limit; idx--)
+    excludedNS.deppend();
+  if (excludedCount.number()) excludedCount.deppend();
+  //extensions
+  limit = extensionCount.number() ? extensionCount.last() : 0;
+  for (idx = extensionNS.number() - 1; idx >= limit; idx--)
+    extensionNS.deppend();
+  if (extensionCount.number()) extensionCount.deppend();
+}
+
+//
+//
+//  StylesheetSructure
+//
+//
+
 eFlag StylesheetStructure::findBestRule(
     Sit S, 
     XSLElement *&ret, 
@@ -107,13 +192,84 @@ eFlag StylesheetStructure::findBestRule(
 XSLElement* StylesheetStructure::findRuleByName(Tree &t, QName &q)
 {
     XSLElement* ret = rulesList.findByName(t,q);
-    int importCount = importChildren.number(),
-	i;
-    for (i = 0; !ret && i < importCount; i++)
-	ret = importChildren[i] -> rulesList.findByName(t,q);
+    if (! ret)
+      {
+	int importCount = importChildren.number(),
+	  i;
+	for (i = 0; !ret && i < importCount; i++)
+	  ret = importChildren[i] -> findRuleByName(t,q);
+	  //ret = importChildren[i] -> rulesList.findByName(t,q);
+      }
     return ret;
 }
 
+Bool StylesheetStructure::hasAnyStripped()
+{
+  Bool ret = strippedNamesList.number();
+  if (!ret)
+    {
+      int importCount = importChildren.number();
+      for (int i = 0; i < importCount && !ret; i++)
+	{
+	  ret = importChildren[i] -> hasAnyStripped();
+	}
+    }
+  return ret;
+}
+
+Bool StylesheetStructure::hasAnyPreserved()
+{
+  Bool ret = preservedNamesList.number();
+  if (!ret)
+    {
+      int importCount = importChildren.number();
+      for (int i = 0; i < importCount && !ret; i++)
+	{
+	  ret = importChildren[i] -> hasAnyPreserved();
+	}
+    }
+  return ret;
+}
+
+Bool StylesheetStructure::findStrippedName(EQName &ename, int &prec, double &pri)
+{
+  Bool ret = FALSE;
+  if (strippedNamesList.findName(ename, pri)) {
+    prec = importPrecedence;
+    ret = TRUE;
+  }
+  else
+    {
+      int importCount = importChildren.number();
+      for (int i = 0; i < importCount && !ret; i++)
+	{
+	  // call findBestRule recursively, setting importsOnly to FALSE
+	  ret = importChildren[i] -> findStrippedName(ename, prec, pri);
+	}
+    }
+  return ret;
+}
+
+Bool StylesheetStructure::findPreservedName(EQName &ename, int &prec, double &pri)
+{
+  Bool ret = FALSE;
+  if (preservedNamesList.findName(ename, pri)) {
+    prec = importPrecedence;
+    ret = TRUE;
+  }
+  else
+    {
+      int importCount = importChildren.number();
+      for (int i = 0; i < importCount && !ret; i++)
+	{
+	  // call findBestRule recursively, setting importsOnly to FALSE
+	  ret = importChildren[i] -> findPreservedName(ename, prec, pri);
+	}
+    }
+  return ret;
+}
+
+
 //
 //
 //  SubtreeList
@@ -256,6 +412,7 @@ eFlag AttSet::execute(Sit S, Context *c, Tree& sheet, QNameList& history, Bool r
   for (i = 0; i < number(); i++)
     {
       XSLElement *def = (*this)[i] -> getAttributeDef();
+      //execute the content
       E( def -> execute(S, c, resolvingGlobals) );
     }
   return OK;
@@ -310,7 +467,7 @@ eFlag AttSetList::checkRedefinitions(Sit S)
     return OK;
 }
 
-AttSet *AttSetList::insert(QName &name)
+AttSet* AttSetList::insert(QName &name)
 {
     int ndx = findNdx(name);
     AttSet *ptr = NULL;
@@ -349,7 +506,8 @@ int AttSetList::findNdx(const QName &what) const
 //
 //
 
-void AliasItem::set(Phrase newKey, Phrase newValue, int newPrecedence, XSLElement *source)
+void AliasItem::set(Phrase newKey, Phrase newValue, Phrase newPrefix,
+		    int newPrecedence, XSLElement *source)
 {
     assert(newPrecedence >= 0);
     if (key == UNDEF_PHRASE) key = newKey;
@@ -359,6 +517,7 @@ void AliasItem::set(Phrase newKey, Phrase newValue, int newPrecedence, XSLElemen
     {
 	value = newValue;
 	precedence = newPrecedence;
+	prefix = newPrefix;
     }
     if (newPrecedence < precedence)
 	redefinition = NULL;
@@ -383,7 +542,8 @@ Phrase AliasList::find(Phrase key_) const
     return (ndx != -1) ? (*this)[ndx] -> getValue() : PHRASE_NOT_FOUND;
 }
 
-void AliasList::insertAlias(Phrase key, Phrase value, int precedence, XSLElement *source)
+void AliasList::insertAlias(Phrase key, Phrase value, Phrase prefix,
+			    int precedence, XSLElement *source)
 {
     AliasItem *newItem;
     int ndx = findNdx(key);
@@ -391,7 +551,7 @@ void AliasList::insertAlias(Phrase key, Phrase value, int precedence, XSLElement
 	append(newItem = new AliasItem);
     else
 	newItem = (*this)[ndx];
-    newItem -> set(key, value, precedence, source);
+    newItem -> set(key, value, prefix, precedence, source);
 }
 
 eFlag AliasList::checkRedefinitions(Sit S, Tree &sheet)
@@ -423,27 +583,39 @@ Tree::Tree(const Str &aname, BOOL aXSLTree)
     pendingTextNode = NULL;
     getRoot().stamp = 0;
     vcount = 1;
-    stripped = 0;
     QName dummyName;
     theDummyElement = new(&theArena) Element(*this, dummyName);
     initDict();
 
     // append entry for 'the whole tree' to subtrees
     subtrees.push(new SubtreeInfo(
-	aname, XSL_NONE, &structure));
+	aname, XSL_NONE, &structure, FALSE));
+    getRoot().setSubtreeInfo(subtrees.last());
+
+    excludeStdNamespaces();
+
+    pendingNSList.append(new(&theArena) NSList());
+
+    stripped = aXSLTree;
+    hasAnyStripped = hasAnyPreserved = -1;// means 'who knows?'
+    importUnique = 0xFFFF;
 };
 
 Tree::~Tree()
 {
     getRoot().~RootNode();
     // this is just in case there's a processing error:
-    pendingNS.freeall(FALSE);
-    preservedElements.freeall(FALSE);
-    strippedElements.freeall(FALSE);
+    //pendingNSList.freeall(FALSE);
     delete theDummyElement;
     // clear subtree information
     subtrees.freeall(FALSE);
     aliasesList.freeall(FALSE);
+    unparsedEntities.freeall(FALSE);
+    
+    pendingNSList.freelast(FALSE);
+    //assert(!pendingNSList.number());
+	//tmpList must be empty before the arena gets down automatically
+	tmpList.freeall(FALSE);
 }
 
 void Tree::initDict()
@@ -453,11 +625,23 @@ void Tree::initDict()
     theDictionary.insert("xsl", stdPhrases[PHRASE_XSL]);
     theDictionary.insert(theXSLTNamespace, stdPhrases[PHRASE_XSL_NAMESPACE]);
     theDictionary.insert(theXMLNamespace, stdPhrases[PHRASE_XML_NAMESPACE]);
+    theDictionary.insert(theSabExtNamespace, stdPhrases[PHRASE_SABEXT_NAMESPACE]);
     theDictionary.insert("*", stdPhrases[PHRASE_STAR]);
     theDictionary.insert("xmlns", stdPhrases[PHRASE_XMLNS]);
     theDictionary.insert("lang", stdPhrases[PHRASE_LANG]);
 }
 
+void Tree::excludeStdNamespaces()
+{
+  NZ(getCurrentInfo()) -> 
+    getExcludedNS().addUri(stdPhrase(PHRASE_XML_NAMESPACE));
+  if (XSLTree)
+    {
+      NZ(getCurrentInfo()) -> 
+	getExcludedNS().addUri(stdPhrase(PHRASE_XSL_NAMESPACE));
+    }
+}
+
 Element &Tree::dummyElement() const
 {
     return *theDummyElement;
@@ -496,13 +680,20 @@ void Tree::dropCurrentElement(Vertex *v)
     toE(stackTop) -> contents.deppend();
 }
 
-Vertex *Tree::appendText(Sit S, char *string, int len)
+Vertex* Tree::appendText(Sit S, char *string, int len)
 {
     Vertex *txt = NULL;
     if (!pendingTextNode)
     {
         // the initializing text does not matter
         txt = new(&getArena()) Text(*this, string, len);
+	Processor *proc = S.getProcessor();
+	if ( proc && proc->outputter() ) 
+	  {
+	    OutputDocument *doc=proc->outputter()->getDocumentForLevel(FALSE);
+	    txt -> setOutputDocument(doc);
+	  }
+
         appendVertex(S, txt);
         pendingTextNode = toText(txt);
     }
@@ -510,7 +701,7 @@ Vertex *Tree::appendText(Sit S, char *string, int len)
     return txt;
 }
 
-Vertex *Tree::popVertex()
+Vertex* Tree::popVertex()
 {
     Vertex *v = NZ( stackTop );
     stackTop = v -> parent;
@@ -528,7 +719,7 @@ HashTable& Tree::dict()
     return theDictionary;
 }
 
-Arena& Tree::getArena()
+SabArena& Tree::getArena()
 {
     return theArena;
 }
@@ -669,14 +860,15 @@ void Tree::makeStamps()
     vcount = stamp_;
 }
 
-Bool getWhDelimString(char *&list, Str& firstPart)
+void Tree::updateImportStatus() 
 {
-    skipWhite(list);
-    if (!*list) return FALSE;
-    char *list_was = list;
-    for(; *list && !isWhite(*list); list++);
-    firstPart.nset(list_was, (int)(list - list_was));
-    return TRUE;
+  if (! subtrees.getCurrent() -> getStructure() -> getTopLevelFound() )
+    {
+      StylesheetStructure * stru = subtrees.getCurrent() -> getStructure();
+      stru -> setTopLevelFound(true);
+      stru -> setImportPrecedence(importUnique);
+      importUnique--;
+    }
 }
 
 /*****************************************************************
@@ -708,7 +900,11 @@ eFlag Tree::processVertexAfterParse(Sit S, Vertex *v, TreeConstructer* tc)
   if (isXSLElement(v))
     {
       XSLElement *x = toX(v);
-      switch(theOp = x -> op)
+      theOp = x -> op;
+
+      if (theOp != XSL_IMPORT) updateImportStatus();
+
+      switch(theOp)
         {
 	  //catch xsl:use-attribute-sets
 	case XSL_ELEMENT:
@@ -719,20 +915,11 @@ eFlag Tree::processVertexAfterParse(Sit S, Vertex *v, TreeConstructer* tc)
 	  }; break;
 	case XSL_IMPORT:
 	  {
-	    // first check for ill-positioned xsl:import,
-	    // then continue as for include
-	    if (v -> parent)
+	    if (subtrees.getCurrent() -> getStructure() -> getTopLevelFound())
 	      {
-		Vertex *w;
-		int pos = v -> ordinal - 1;
-		for (; pos >= 0; pos--)
-		  {
-		    if (isElement(w = toD(v -> parent) -> contents[pos]) &&
-			(!isXSL(w) || toX(w) -> op != XSL_IMPORT))
-		      Err2(S, E_ELEM_CONTAINS_ELEM, 
-			   xslOpNames[XSL_STYLESHEET], 
-			   xslOpNames[XSL_IMPORT]);
-		  }
+		Err2(S, E_ELEM_CONTAINS_ELEM, 
+		     xslOpNames[XSL_STYLESHEET], 
+		     xslOpNames[XSL_IMPORT]);
 	      }
 	  }; // no break
 	case XSL_INCLUDE:
@@ -743,7 +930,7 @@ eFlag Tree::processVertexAfterParse(Sit S, Vertex *v, TreeConstructer* tc)
 					    getBaseURI());		
 	    
 	    Str absolute;
-	    makeAbsoluteURI(a -> cont, base, absolute);
+	    makeAbsoluteURI(S, a -> cont, base, absolute);
 	    if (S.getProcessor())
 	      {
 		E( S.getProcessor() -> readTreeFromURI(S, srcTree, 
@@ -764,20 +951,43 @@ eFlag Tree::processVertexAfterParse(Sit S, Vertex *v, TreeConstructer* tc)
 	    
 	    Element *theSheet=(*srcTree).findStylesheet((*srcTree).getRoot());
 	    if (!theSheet)
-	      Warn1(S, W_NO_STYLESHEET, (char*)(a -> cont));
+		Warn1(S, W_NO_STYLESHEET, (char*)(a -> cont));
 	    dropCurrentElement(v);
 	    
 	    OutputterObj source;
-	    // we start a subtree to record where the nodes come from
-	    // when including, we use the old structure
-	    // when importing, Tree creates a new one
+	    //we start a subtree to record where the nodes come from
+	    //when including, we use the old structure
+	    //when importing, Tree creates a new one
 	    E( startSubtree(S, (*srcTree).getURI(), theOp) );
-	    E( tc -> parseUsingSAXForAWhile(S, source, absolute) );
+	    //set extension namespaces for subtree
+
+	    //(*srcTree).speakDebug();
+
+	    //merge it into the current tree
+	    E( tc -> parseUsingSAXForAWhile(S, source, absolute, 
+					    TRUE, 
+					    (Tree*)srcTree,
+					    theSheet -> namespaces) );
+
+	    //first we have to deal with ext. and excl. namespaces
+	    Attribute *attr;
+	    QName q;
+	    //exclusions
+	    q.setLocal((*srcTree).unexpand("exclude-result-prefixes"));
+	    attr = theSheet->atts.find(q);
+	    if (attr)
+	      E(pushNamespacePrefixes(S, attr->cont, XSLA_EXCL_RES_PREFIXES));
+	    //extensions
+	    q.setLocal((*srcTree).unexpand("extension-element-prefixes"));
+	    attr = theSheet->atts.find(q);
+	    if (attr)
+	      E(pushNamespacePrefixes(S, attr->cont, XSLA_EXT_ELEM_PREFIXES));
+
 	    if (theSheet)
 	      E( theSheet -> contents.copy(S, source) );
-	    E( tc -> parseUsingSAXForAWhileDone(S, source) );
+	    E( tc -> parseUsingSAXForAWhileDone(S, source, TRUE) );
 	    // end the subtree
-	    E( endSubtree(S) );		        
+	    E( endSubtree(S, theOp) );		
 	  }; break;
 	case XSL_OUTPUT:
 	  {
@@ -822,6 +1032,8 @@ eFlag Tree::processVertexAfterParse(Sit S, Vertex *v, TreeConstructer* tc)
 			}
 		      while (someRemains);
 		    }; break;
+		  case XSLA_NONE: //skip other namespaces
+		    break;
 		  default:
 		    {
 		      E( outputDef.setItemStr(S, theAtt -> op, theAtt -> cont, 
@@ -835,14 +1047,29 @@ eFlag Tree::processVertexAfterParse(Sit S, Vertex *v, TreeConstructer* tc)
 	  }; break;
 	case XSL_NAMESPACE_ALIAS:
 	  {
-	    Phrase style,result;
+	    Phrase style, result, sUri, rUri;
 	    Attribute *sp = NZ( x -> atts.find(XSLA_STYLESHEET_PREFIX) );
 	    Attribute *rp = NZ( x -> atts.find(XSLA_RESULT_PREFIX) );
 	    if (sp -> cont == "#default") style = UNDEF_PHRASE;
 	    else dict().insert(sp -> cont, style);
 	    if (rp -> cont == "#default") result = UNDEF_PHRASE;
 	    else dict().insert(rp -> cont, result);
-	    aliases().insertAlias(style, result, v -> getImportPrecedence(), x);
+
+	    int i;
+	    i = pendingNS().findNdx(style);
+	    if (i != -1)
+	      sUri = toNS(pendingNS().operator[](i)) -> uri;
+	    else
+	      Err1(S, E_EX_NAMESPACE_UNKNOWN, (char*) sp -> cont);
+
+	    i = pendingNS().findNdx(result);
+	    if (i != -1)
+	      rUri = toNS(pendingNS().operator[](i)) -> uri;
+	    else
+	      Err1(S, E_EX_NAMESPACE_UNKNOWN, (char*) rp -> cont);
+
+	    aliases().insertAlias(sUri, rUri, result, 
+				  v -> getImportPrecedence(), x);
 	    popVertex();
 	  }; break;
 	case XSL_TEMPLATE:
@@ -880,8 +1107,6 @@ eFlag Tree::processVertexAfterParse(Sit S, Vertex *v, TreeConstructer* tc)
 	case XSL_STYLESHEET:
 	case XSL_TRANSFORM:
 	  {
-	    E( excludeNamespaces(S, toE(v), TRUE) );
-	    /* E( extensionNamespaces(S, toE(v)) ); */
 	    popVertex();
 	  }; break;
 	case XSL_VARIABLE:
@@ -905,17 +1130,31 @@ eFlag Tree::processVertexAfterParse(Sit S, Vertex *v, TreeConstructer* tc)
 	      }
 	    popVertex();
 	  }; break;
+	case XSL_STRIP_SPACE:
+	  {
+	    SpaceNameList &foo = 
+	      subtrees.getCurrent() -> getStructure() -> strippedNames();
+	    E( getSpaceNames(S, *x, NZ(x -> atts.find(XSLA_ELEMENTS)) -> cont, 
+			     foo) );
+	    popVertex();
+	  }; break;
+	case XSL_PRESERVE_SPACE:
+	  {
+	    SpaceNameList &foo = 
+	      subtrees.getCurrent() -> getStructure() -> preservedNames();
+	    E( getSpaceNames(S, *x, NZ(x -> atts.find(XSLA_ELEMENTS)) -> cont, 
+			     foo) );
+	    popVertex();
+	  }; break;
 	default:
 	  popVertex();
 	}
 	//the literal output element may have some xsl features
     }
-    else {
+  else 
+    { //isXSLElement
+      updateImportStatus();
       if (XSLTree) {
-	Bool isRoot = (getRoot().contents.number()) > 0 && 
-	  (v == getRoot().contents[0]);
-	E( excludeNamespaces(S, toE(v), isRoot) );
-	
 	E( extractUsedSets(S, toE(v)) );
 	popVertex();
       }
@@ -923,80 +1162,106 @@ eFlag Tree::processVertexAfterParse(Sit S, Vertex *v, TreeConstructer* tc)
 	popVertex();
       }
     }
-    return OK;
+  return OK;
 }
 
-eFlag Tree::parsePrefixes(Sit S, UriList &ul, NSList &namespaces, Str &tokens)
+eFlag Tree::getSpaceNames(Sit S, Element &e, Str &str, SpaceNameList &where)
 {
   char *p, *q;
-
-  //ul.deppendall();
-  q = (char*)tokens;
+  q = (char*)str;
   skipWhite(q);
   p = q;
   int i = strcspn(q, theWhitespace);
-  while (*q && i) {
-    q += i;
-    char save = *q;
-    *q = 0;
+  while (*q && i) 
+    {
+      q += i;
+      char save = *q;
+      *q = 0;
+      
+      Str token = p;
+      QName name;
+      E( e.setLogical(S, name, token, FALSE) );
+      GP(EQName) ename = new EQName;
+      expandQ(name, *ename);
+      where.append(ename.keep());
 
-    Str prefix = p;
-    int idx = namespaces.findNdx(prefix == "#default" ? UNDEF_PHRASE : 
-				 unexpand(prefix));
-    if (idx == -1)
-      Err1(S, E_EX_NAMESPACE_UNKNOWN, p);
+      *q = save;
+      skipWhite(q);
+      p = q;
+      i = strcspn(q, theWhitespace);
+    }
 
-    ul.append(new Str(expand(toNS(namespaces[idx]) -> uri)));
-    *q = save;
-    skipWhite(q);
-    p = q;
-    i = strcspn(q, theWhitespace);
-  }
   return OK;
 }
 
-eFlag Tree::excludeNamespaces(Sit S, Element *e, Bool isRoot) {
-  //exclude result prefixes
-  Attribute *a = e -> atts.find(XSLA_EXCL_RES_PREFIXES);
-  if (a) E( parsePrefixes(S, *(e->excludedNS(TRUE)), e -> namespaces, 
-			a -> cont) );
-  //exclude extension element prefixes
-  a = e -> atts.find(XSLA_EXT_ELEM_PREFIXES);
-  if (a) E( parsePrefixes(S, *(e->excludedNS(TRUE)), e -> namespaces, 
-			a -> cont) );
-  if (isRoot) {
-    //add xml prefix
-    Str xml = theXMLNamespace;
-    Str xsl = theXSLTNamespace;
-    Bool forceXSL = e -> excludedNS(TRUE) -> findNdx(xsl) != -1;
-    
-    e -> excludedNS(TRUE) -> addUri(xml);
-    e -> excludedNS(TRUE) -> addUri(xsl);
-    
-    //if the target alias is bound to xslt nemespace,
-    //we remove theXSLTNamespace from the exclusion list to
-    //eliminate specification leak 
-    //(xslt namespace sould never be output)
-    for (int i = 0; i < aliases().number(); i++) {
-      Phrase res = (aliases()[i]) -> getValue();
-      NmSpace *nm;
-      if ((nm = e -> namespaces.find(res)) && 
-	  (expand(nm -> uri) == theXSLTNamespace) && !forceXSL)
+eFlag Tree::markNamespacePrefixes(Sit S)
+{
+  if (XSLTree)
+    getCurrentInfo() -> pushNamespaceMarks();
+  return OK;
+}
+
+eFlag Tree::pushNamespacePrefixes(Sit S, Str& prefixes, XSL_ATT att)
+{
+  if (!XSLTree) return OK;
+
+  PList<Str*> tokens;
+  char *p, *q;
+  q = (char*) prefixes;
+  skipWhite(q);
+  p = q;
+  int i = strcspn(q, theWhitespace);
+  while (*q && i) 
+    {
+      q += i;
+      char save = *q;
+      *q = 0;
+      
+      Str* aux = new Str(p);
+      tokens.append(aux);
+
+      *q = save;
+      skipWhite(q);
+      p = q;
+      i = strcspn(q, theWhitespace);
+    }
+  //add to list
+  SubtreeInfo *info = getCurrentInfo();
+  for (i = 0; i < tokens.number(); i++)
+    {
+      Str tok = *(tokens[i]);
+      Phrase prefix = 
+	tok == (const char*) "#default" ? UNDEF_PHRASE : unexpand(tok);
+      int idx = pendingNS().findNdx(prefix);
+      if (idx != -1)
 	{
-	  e -> excludedNS(TRUE) -> removeUri(xsl);
+	  switch (att) {
+	  case XSLA_EXT_ELEM_PREFIXES:
+	    {
+	      info -> getExtensionNS().append(toNS(pendingNS()[idx]) -> uri);
+	    }; //no break
+	  case XSLA_EXCL_RES_PREFIXES:
+	    {
+	      info -> getExcludedNS().append(toNS(pendingNS()[idx]) -> uri);
+	    }; break;
+	  }
+	}
+      else
+	{
+	  Str aux = *(tokens[i]);
+	  tokens.freeall(FALSE);
+	  Err1(S, E_EX_NAMESPACE_UNKNOWN, (char*) aux);
 	}
     }
-  }
+  tokens.freeall(FALSE);
+  //printf("----------------------------\n");
   return OK;
 }
 
-eFlag Tree::extensionNamespaces(Sit S, Element *e) {
-  //exclude extension element prefixes
-  Attribute *a = e -> atts.find(XSLA_EXT_ELEM_PREFIXES);
-  if (a) {
-    E( parsePrefixes(S, *(e->extensionNS(TRUE)), e -> namespaces, 
-		     a -> cont) );
-  }
+eFlag Tree::popNamespacePrefixes(Sit S)
+{
+  if (XSLTree)
+    getCurrentInfo() -> popNamespaceMarks();
   return OK;
 }
 
@@ -1039,7 +1304,7 @@ finds a xsl:stylesheet child of the given daddy. Returns NULL
 if not found.
 *****************************************************************/
 
-Element *Tree::findStylesheet(Daddy& d)
+Element* Tree::findStylesheet(Daddy& d)
 {
     Vertex *w;
     int dContentsNumber = d.contents.number();
@@ -1073,25 +1338,26 @@ double Tree::defaultPriorityLP(Expression *lpath)
         {
         case EXNODE_COMMENT:
         case EXNODE_TEXT:
-        case EXNODE_PI:  //// pi('literal') should have priority 0.5
         case EXNODE_NODE:
-            return -0.5;
-            break;
+	  return -0.5;
+	  break;
+        case EXNODE_PI:
+	  return lpath->args[0]->step->piname == "" ? -0.5 : 0;
         case EXNODE_NONE:
-            {
-                QName &qn = lpath -> args[0] -> step -> ntest;
-                if (qn.getLocal() != lpath -> getOwnerTree().stdPhrase(PHRASE_STAR))
-                    return 0.0;
-                else
-                {   
-                    if (qn.getPrefix() == UNDEF_PHRASE)
-                        return -0.5;
-                    else
-                        return -0.25;
-                };
-            }; break;
+	  {
+	    QName &qn = lpath -> args[0] -> step -> ntest;
+	    if (qn.getLocal() != lpath -> getOwnerTree().stdPhrase(PHRASE_STAR))
+	      return 0.0;
+	    else
+	      {   
+		if (qn.getPrefix() == UNDEF_PHRASE)
+		  return -0.5;
+		else
+		  return -0.25;
+	      };
+	  }; break;
         default:
-            return 0.5;
+	  return 0.5;
         };
     };
     return 0;   // BCC thinks we don't return enough
@@ -1152,49 +1418,43 @@ eFlag Tree::getMatchingList(Sit S, Expression& match, Context& result)
 
 Bool Tree::isExtensionUri(Phrase uri) 
 {
-  Str search = expand(uri);
-  for (Vertex *v = stackTop; v; v = v -> parent) 
-    {
-      UriList *ulist;
-      if (isElement(v) && (ulist = toE(v) -> extensionNS(FALSE)) &&
-	  (ulist -> findNdx(search) != -1)) {
-	return TRUE;
-      }
-    }
-  return FALSE;
+  return getCurrentInfo() -> getExtensionNS().findNdx(uri) != -1;
 }
 
 eFlag Tree::insertRule(Sit S, XSLElement *tmpl)
 {
-    double prio;
-    Attribute *a = tmpl -> atts.find(XSLA_PRIORITY);
-    if (!a)
-        prio = defaultPriority(tmpl);
-    else
+  double prio;
+  Attribute *a = tmpl -> atts.find(XSLA_PRIORITY);
+  if (!a)
+    prio = defaultPriority(tmpl);
+  else
     {
-        if (a -> cont.toDouble(prio))
-            Err(S, ET_BAD_NUMBER);
+      if (a -> cont.toDouble(prio))
+	Err(S, ET_BAD_NUMBER);
     };
-    QName q; 
-    GP( QName ) mode = NULL;
-    if (!!(a = tmpl -> atts.find(XSLA_NAME)))
-        E( tmpl -> setLogical(S, q, a -> cont, FALSE) );
-    if (q.getLocal() != UNDEF_PHRASE && 
-        subtrees.getCurrent() -> getStructure() -> rules().findByName(
-	    *this, q))
+  QName q; 
+  GP( QName ) mode = NULL;
+  
+  if (!!(a = tmpl -> atts.find(XSLA_NAME)))
+    E( tmpl -> setLogical(S, q, a -> cont, FALSE) );
+  
+  if (q.getLocal() != UNDEF_PHRASE && 
+      subtrees.getCurrent() -> getStructure() -> 
+      rules().findByName(*this, q))
     {
-	Str fullName;
-	expandQStr(q, fullName);
-
-	Err1(S, ET_DUPLICATE_RULE_NAME, fullName);
+      Str fullName;
+      expandQStr(q, fullName);
+      
+      Err1(S, ET_DUPLICATE_RULE_NAME, fullName);
     };
-    if (!!(a = tmpl -> atts.find(XSLA_MODE)))
-        E( tmpl -> setLogical(S, *(mode = new QName), 
-			      a -> cont, FALSE) );
-    
-    subtrees.getCurrent() -> getStructure() -> rules().insert(
-	new RuleItem(tmpl,prio,q,mode.keep()));
-    return OK;
+  
+  if (!!(a = tmpl -> atts.find(XSLA_MODE)))
+    E( tmpl -> setLogical(S, *(mode = new QName), 
+			  a -> cont, FALSE) );
+  
+  subtrees.getCurrent() -> getStructure() -> 
+    rules().insert(new RuleItem(tmpl,prio,q,mode.keep()));
+  return OK;
 }
 
 eFlag Tree::insertAttSet(Sit S, XSLElement *tmpl)
@@ -1217,7 +1477,8 @@ eFlag Tree::insertAttSet(Sit S, XSLElement *tmpl)
 
 
 
-eFlag Tree::startSubtree(Sit S, const Str& baseURI, XSL_OP dependency)
+eFlag Tree::startSubtree(Sit S, const Str& baseURI, XSL_OP dependency,
+			 Bool isInline /* = FALSE */)
 {
     // look if the URI is on the way to root of the include tree
     if (subtrees.findAmongPredecessors(baseURI))
@@ -1228,30 +1489,57 @@ eFlag Tree::startSubtree(Sit S, const Str& baseURI, XSL_OP dependency)
     else
 	// find structure of current subtree (always defined since Tree::Tree)
 	structure = NZ(subtrees.getCurrent()) -> getStructure();
+
     subtrees.push(
 	new SubtreeInfo(
 	    baseURI, 
 	    dependency,
-	    structure));
+	    structure,
+	    isInline));
+
+    excludeStdNamespaces();
+    //set parent tree (closest non-inline)
+    if (isInline)
+      for (SubtreeInfo *nfo = subtrees.getCurrent();
+	   nfo;
+	   nfo = nfo -> getParentSubtree())
+	{
+	  if (!nfo -> isInline())
+	    {
+	      subtrees.getCurrent() -> setMasterSubtree(nfo);
+	      break;
+	    }
+	}
+
     return OK;
 }
 
-eFlag Tree::endSubtree(Sit S)
+eFlag Tree::endSubtree(Sit S, XSL_OP dependency)
 {
-    // move current subtree
-    subtrees.pop();
-    return OK;
+  /*
+  subtrees.getCurrent()->getStructure()->setImportPrecedence(importUnique);
+  if (dependency == XSL_IMPORT)
+    {
+      importUnique--;
+    }
+  */
+
+  // move current subtree
+  subtrees.pop();
+  return OK;
 }
 
-StylesheetStructure *Tree::createStylesheetStructure(Sit S)
+StylesheetStructure* Tree::createStylesheetStructure(Sit S)
 {
     // current subtree is created in Tree::Tree() so must be NZ
     SubtreeInfo *currSubtree = NZ( subtrees.getCurrent() );
     // there is always a structure for an existing subtree
     StylesheetStructure *currStruct  = NZ( currSubtree -> getStructure() );
     // create new structure with precedence one higher
-    StylesheetStructure *newStruct = new StylesheetStructure(
-	currStruct -> getImportPrecedence() + 1);
+    //StylesheetStructure *newStruct = new StylesheetStructure(
+    //currStruct -> getImportPrecedence() + 1);
+    StylesheetStructure *newStruct = new StylesheetStructure(0);
+
     // file newStruct into existing tree if stylesheet structures
     currStruct -> addImportStructure(S, newStruct);
     return newStruct;    
@@ -1262,17 +1550,46 @@ eFlag Tree::findBestRule(
     XSLElement *&ret, 
     Context *c,
     QName *currMode,
-    Bool importsOnly)
+    Bool importsOnly,
+    SubtreeInfo *subtree /* NULL */)
 {
-    return NZ(subtrees[0]) -> getStructure() -> findBestRule(
+  SubtreeInfo *start = (importsOnly && subtree) ? subtree : subtrees[0];
+    return NZ(start) -> getStructure() -> findBestRule(
 	S, ret, c, currMode, importsOnly);
 }
 
-XSLElement *Tree::findRuleByName(QName &q)
+Bool Tree::hasAnyStrippedName() {
+  if (hasAnyStripped == -1)
+    hasAnyStripped = subtrees[0] -> getStructure() -> hasAnyStripped();
+  return hasAnyStripped;
+}
+
+Bool Tree::hasAnyPreservedName() {
+  if (hasAnyPreserved == -1)
+    hasAnyPreserved = subtrees[0] -> getStructure() -> hasAnyPreserved();
+  return hasAnyPreserved;
+}
+
+Bool Tree::findStrippedName(EQName &name, int &prec, double &prio)
+{
+  return 
+    NZ(subtrees[0]) -> getStructure() -> 
+        findStrippedName(name, prec, prio);
+}
+
+Bool Tree::findPreservedName(EQName &name, int &prec, double &prio)
+{
+  return 
+    NZ(subtrees[0]) -> getStructure() -> 
+        findPreservedName(name, prec, prio);
+}
+
+XSLElement* Tree::findRuleByName(QName &q)
 {
     return NZ( subtrees[0] ) -> getStructure() -> findRuleByName(*this, q);
 }
 
+/*
 eFlag Tree::resolveGlobals(Sit S, Context *c, Processor *proc)
 {
     assert(proc);
@@ -1286,9 +1603,60 @@ eFlag Tree::resolveGlobals(Sit S, Context *c, Processor *proc)
     }
     return OK;
 }
+*/
 
+void Tree::setUnparsedEntityUri(Str &name, Str &uri)
+{
+  unparsedEntities.appendConstruct(name, uri);
+}
 
+Str* Tree::getUnparsedEntityUri(Str &name)
+{
+  return unparsedEntities.find(name);
+}
+
+eFlag Tree::pushPendingNS(Sit S, Tree* srcTree, NSList &other) 
+{
+  NSList *lst = new(&theArena) NSList();
+  lst -> swallow(S, other, srcTree, this);
+  pendingNSList.append(lst);
+  return OK;
+}
     
+eFlag Tree::popPendingNS(Sit S)
+{
+  pendingNSList.freelast(FALSE);
+  return OK;
+}
+
+void Tree::speakDebug()
+{
+  DStr foo;
+  getRoot().speak(foo, (SpeakMode)(SM_OFFICIAL | SM_INS_SPACES));
+  printf("--------------------\n%s\n--------------------\n", (char*)foo);
+}
+
+void Tree::dumpStructure(Sit S)
+{
+  S.message(MT_LOG, L_SHEET_STRUCTURE, "", "");
+  for (int i = 0; i < subtrees.number(); i++)
+    {
+      //count level
+      DStr out = "";
+      int level = 0;
+      SubtreeInfo * si = subtrees[i];
+      while (si)
+	{
+	  if (level > 0) out += "  ";
+	  level++;
+	  si = si -> getParentSubtree();
+	}
+      
+      out += subtrees[i] -> getBaseURI();
+      S.message(MT_LOG, L_SHEET_ITEM, out, "");
+    }
+}
+
 //_TH_ v
 /****************************************
 T m p L i s t   methods
@@ -1335,6 +1703,7 @@ void TmpList::freeall(Bool a)
 
 void TmpList::dump(int p)
 { 
+  /*
   Str vname;
   if(p==0) printf("TmpList.dump (%d item(s))\n", nItems);
   else printf("TmpList.dump %d (%d item(s))\n", p, nItems);
@@ -1342,6 +1711,7 @@ void TmpList::dump(int p)
     ((*this)[i])->getOwner().expandQStr(toE((*this)[i]) -> getName(),vname);
     printf("%d:>%d< >%x< (%s)\n",i,((*this)[i])->ordinal,((*this)[i]),(char*)vname);
   }
+  */
 };
 
 int TmpList::findNum(void *p) const
@@ -1356,3 +1726,53 @@ void TmpList::rmP(void *p)
   rm(((Vertex *)p)->ordinal);
 }
 //_TH_ ^
+
+Attribute* AttsCache::find(XSL_ATT what)
+{
+    int i;
+    Attribute *a;
+    int num = number();
+    for (i = 0; i < num; i++)
+    {
+	// need to use a temporary variable
+	// to get around Solaris template problem
+        Vertex * pTemp = (*this)[i];
+        a = cast(Attribute *, pTemp);
+        if (a -> op == what)
+            return a;
+    };
+    return NULL;
+}
+
+int AttsCache::findNdx(const QName& attName)
+{
+    int i;
+    Attribute *a;
+    int num = number();
+    for (i = 0; i < num; i++)
+    {
+	// need to use a temporary variable
+	// to get around Solaris template problem
+        Vertex * pTemp = (*this)[i];
+        a = toA(pTemp);
+        if (attName == a -> getName())
+            return i;
+    };
+    return -1;
+}
+
+Attribute* AttsCache::find(const QName& attName)
+{
+    int ndx = findNdx(attName);
+    // need to use a temporary variable
+    // to get around Solaris template problem
+    if (ndx != -1)
+      {
+	Vertex * pTemp = (*this)[ndx];
+	return toA(pTemp);
+      } 
+    else
+      {
+	return NULL;
+      }
+}
