@@ -12,8 +12,8 @@
  * The Original Code is the Sablotron XSLT Processor.
  * 
  * The Initial Developer of the Original Code is Ginger Alliance Ltd.
- * Portions created by Ginger Alliance are Copyright (C) 2000 Ginger
- * Alliance Ltd. All Rights Reserved.
+ * Portions created by Ginger Alliance are Copyright (C) 2000-2002
+ * Ginger Alliance Ltd. All Rights Reserved.
  * 
  * Contributor(s):
  * 
@@ -37,6 +37,7 @@
 #include "uri.h"
 #include <string.h>
 #include "proc.h"
+#include "platform.h"
 
 // GP: clean
 
@@ -59,6 +60,24 @@
 #define slashes "/\\"
 #define isSlash(c) (c == '/' || c == '\\')
 
+//special name reporting under windows
+#ifdef __WIN_TOOLS
+#define winName(name) ((name[0] == '/' && name[2] == ':') ? name + 1 : name)
+#else
+#define winName(name) name
+#endif
+
+/*****************************************************************
+uriHasAuthority
+
+  decides, whether given schema should contain the authority
+*****************************************************************/
+
+Bool uriHasAuthority(Str & scheme)
+{
+  return (scheme == (const char*)"file");
+}
+
 /*****************************************************************
 splitBy
 
@@ -102,8 +121,9 @@ void splitURI(const char *uri, FiveStr &parts)
     if (isSlash(*rest) && isSlash(rest[1]))
         RF( c = splitBy(rest += 2, slashes"?#", parts[U_AUTH]) );
     if (isSlash(c) || c == 'A')
-        // extract the path
-        RF( c = splitBy(rest -= (isSlash(c)), "?#", parts[U_PATH]) );
+      // extract the path
+      RF( c = splitBy(rest -= (isSlash(c)), "?#", parts[U_PATH]) );
+    //query and fragment
     if (c == '?')
         // extract the query
         RF( c = splitBy(rest, "#", parts[U_QUERY]) );
@@ -113,16 +133,17 @@ void splitURI(const char *uri, FiveStr &parts)
 
 void joinURI(DStr &joined, FiveStr &parts, Bool schemeToo)
 {
-    joined.empty();
-    if (schemeToo && !parts[U_SCHEME].isEmpty())
-        joined = parts[U_SCHEME] + ":";
-    if (!parts[U_AUTH].isEmpty())
-        joined += Str("//") + parts[U_AUTH];            // add authority
-    joined += parts[U_PATH];                // add path
-    if (!parts[U_QUERY].isEmpty())          // add query
-        joined += Str("?") + parts[U_QUERY];
-    if (!parts[U_FRAG].isEmpty())           // add fragment
-        joined += Str("#") + parts[U_FRAG];
+  joined.empty();
+  if (schemeToo && !parts[U_SCHEME].isEmpty())
+    joined = parts[U_SCHEME] + ":";
+  //if (!parts[U_AUTH].isEmpty())
+  if (uriHasAuthority(parts[U_SCHEME]))
+    joined += Str("//") + parts[U_AUTH];  // add authority
+  joined += parts[U_PATH];                // add path
+  if (!parts[U_QUERY].isEmpty())          // add query
+    joined += Str("?") + parts[U_QUERY];
+  if (!parts[U_FRAG].isEmpty())           // add fragment
+    joined += Str("#") + parts[U_FRAG];
 }
 
 /*****************************************************************
@@ -133,9 +154,9 @@ schemeToURI_()
   simply returned.
 *****************************************************************/
 
-URIScheme schemeToURI_(Str& scheme)
+URIScheme schemeToURI_(Sit S, Str& scheme)
 {
-    if (scheme.eqNoCase("file"))
+    if (scheme.eqNoCase("file") && !S.hasFlag(SAB_FILES_TO_HANDLER))
         return URI_FILE;
     else
     {
@@ -146,6 +167,7 @@ URIScheme schemeToURI_(Str& scheme)
     }
 }
 
+
 /*****************************************************************
 cutLast()
 
@@ -245,7 +267,8 @@ void joinPaths(Str& relPath, const Str& basePath)
 }
 
 
-URIScheme makeAbsoluteURI2(const char* uri ,const char* base, Str& absolute, Str& scheme)
+URIScheme makeAbsoluteURI2(Sit S, const char* uri,
+			   const char* base, Str& absolute, Str& scheme)
 {
     FiveStr 
         u_parts, 
@@ -290,7 +313,7 @@ URIScheme makeAbsoluteURI2(const char* uri ,const char* base, Str& absolute, Str
     DStr joined = absolute;
     joinURI(joined, u_parts, FALSE);         // join all components into a URI for return (no scheme)
     absolute = joined;
-    return schemeToURI_(scheme = u_parts[U_SCHEME]);
+    return schemeToURI_(S, scheme = u_parts[U_SCHEME]);
 }
 
 
@@ -299,17 +322,18 @@ URIScheme makeAbsoluteURI2(const char* uri ,const char* base, Str& absolute, Str
 //    Merges a (possibly relative) URI reference with a base URI, setting
 //    'absolute' to the result. 
 //
-URIScheme makeAbsoluteURI(const char* uri ,const char* base, Str& absolute)
+URIScheme makeAbsoluteURI(Sit S, const char* uri,
+			  const char* base, Str& absolute)
 {
-    Str scheme;
-    URIScheme temp;
-    temp = makeAbsoluteURI2(uri, base, absolute, scheme);
-    absolute = (scheme + ":") + absolute;
-    return temp;
+  Str scheme;
+  URIScheme temp;
+  temp = makeAbsoluteURI2(S, uri, base, absolute, scheme);
+  absolute = (scheme + ":") + absolute;
+  return temp;
 }
 
 
-URIScheme uri2SchemePath(const char *absolute, Str& scheme, Str& rest)
+URIScheme uri2SchemePath(Sit S, const char *absolute, Str& scheme, Str& rest)
 {
     Bool found = (Bool) !!splitBy(absolute, ":", scheme);
     assert(found);
@@ -320,7 +344,7 @@ URIScheme uri2SchemePath(const char *absolute, Str& scheme, Str& rest)
  *   else
  *       rest = (char*) absolute;
  */
-    return schemeToURI_(scheme);
+    return schemeToURI_(S, scheme);
 }
 
 
@@ -402,12 +426,16 @@ _baseUri    the base URI used in case the reference in _uri is
 _mode       the access mode (DLMODE_READ, DLMODE_WRITE)
 *****************************************************************/
 
-eFlag DataLine::open(Sit S, const char *_uri, DLAccessMode _mode, StrStrList* argList_)
+#define specErr1(S, code, arg) \
+{if (ignoreErr) {Warn1(S,code,arg); return NOT_OK;} else Err1(S,code,arg);}
+
+eFlag DataLine::open(Sit S, const char *_uri, DLAccessMode _mode, 
+		     StrStrList* argList_, Bool ignoreErr /* = FALSE */)
 {
     assert(mode == DLMODE_NONE);  // the buffer must not be open yet
     // combine _uri and _baseUri into one
     Str strScheme, strPath;
-    scheme = uri2SchemePath(_uri, strScheme, strPath);
+    scheme = uri2SchemePath(S, _uri, strScheme, strPath);
     char *name = (char*) strPath;
 
     // mode set in the end
@@ -421,11 +449,11 @@ eFlag DataLine::open(Sit S, const char *_uri, DLAccessMode _mode, StrStrList* ar
                 name += 2;          // skipping the "//" in front
             // try to open the file
 #ifdef _MSC_VER
-            if (!(f = stdopen(name,_mode == DLMODE_WRITE ? "wb" : "rt")))
+            if (!(f = stdopen(name,_mode == DLMODE_WRITE ? "wb" : "rb")))
 #else
             if (!(f = stdopen(name,_mode == DLMODE_WRITE ? "w" : "r")))
 #endif
-                Err1(S, E_FILE_OPEN, name);
+                specErr1(S, E_FILE_OPEN, winName(name));
             // set fileIsStd if filename is "stdin", "stdout" or "stderr"
             fileIsStd = isstd(name);
         }; break;
@@ -435,12 +463,12 @@ eFlag DataLine::open(Sit S, const char *_uri, DLAccessMode _mode, StrStrList* ar
             // plus some extra information
             if (_mode == DLMODE_READ)
             {
-		        Str *value = NULL;
-			    if (argList_)
-			        value = argList_ -> find(name);
-			    if (!value)
-                    Err1(S, E1_ARG_NOT_FOUND, name);
-		        buffer = (char*)*value;
+	      Str *value = NULL;
+	      if (argList_)
+		value = argList_ -> find(name);
+	      if (!value)
+		specErr1(S, E1_ARG_NOT_FOUND, name);
+	      buffer = (char*)*value;
             }
             // if opening for write access, just allocate a new dynamic block
             else
@@ -450,35 +478,35 @@ eFlag DataLine::open(Sit S, const char *_uri, DLAccessMode _mode, StrStrList* ar
         {
             // try the extending scheme handler
             // ask the handler address from the Processor
-	        Processor *proc = S.getProcessor();
-		    if (proc)
-                handler = proc->getSchemeHandler(&handlerUD);
-		    else
-		        handler = NULL;
-            // if there is no handler, report unsupported scheme
-            if (!handler)
-                Err1(S, E1_UNSUPPORTED_SCHEME, strScheme);
-            // try the fast way
-            int count = 0;
-            buffer = NULL;
-            if (handler -> getAll)
-                handler -> getAll(handlerUD, proc,
-                    strScheme, name, &buffer, &count);
-            if (buffer && (count != -1))
+	  Processor *proc = S.getProcessor();
+	  if (proc)
+	    handler = proc->getSchemeHandler(&handlerUD);
+	  else
+	    handler = NULL;
+	  // if there is no handler, report unsupported scheme
+	  if (!handler)
+	    specErr1(S, E1_UNSUPPORTED_SCHEME, strScheme);
+	  // try the fast way
+	  int count = 0;
+	  buffer = NULL;
+	  if (_mode == DLMODE_READ && handler -> getAll)
+	    handler -> getAll(handlerUD, proc,
+			      strScheme, name, &buffer, &count);
+	  if (buffer && (count != -1))
             {
-                gotWholeDocument = TRUE;
-                bufCurr = 0;
+	      gotWholeDocument = TRUE;
+	      bufCurr = 0;
             }
-            else
+	  else
             {
-                // call the handler's open() function, obtaining a handle
-                switch(handler -> open(handlerUD, proc, 
-                    strScheme, name, &handle))
+	      // call the handler's open() function, obtaining a handle
+	      switch(handler -> open(handlerUD, proc, 
+				     strScheme, name, &handle))
                 {
                 case SH_ERR_UNSUPPORTED_SCHEME:     // scheme not supported
-                    Err1(S, E1_UNSUPPORTED_SCHEME, strScheme);
+		  specErr1(S, E1_UNSUPPORTED_SCHEME, strScheme);
                 case SH_ERR_NOT_OK:
-                    Err1(S, E1_URI_OPEN, strScheme + ":" + strPath);
+		  specErr1(S, E1_URI_OPEN, strScheme + ":" + strPath);
                 };
             }
         };
@@ -503,8 +531,8 @@ eFlag DataLine::close(Sit S)
             assert(f);
             if (!fileIsStd)
             {
-                if (fclose(f))
-                    Err1(S, E1_URI_CLOSE, fullUri);
+	      if (fclose(f))
+		Err1(S, E1_URI_CLOSE, fullUri);
             };
             f = NULL;
         }; break;
