@@ -12,8 +12,8 @@
  * The Original Code is the Sablotron XSLT Processor.
  * 
  * The Initial Developer of the Original Code is Ginger Alliance Ltd.
- * Portions created by Ginger Alliance are Copyright (C) 2000 Ginger
- * Alliance Ltd. All Rights Reserved.
+ * Portions created by Ginger Alliance are Copyright (C) 2000-2002
+ * Ginger Alliance Ltd. All Rights Reserved.
  * 
  * Contributor(s):
  * 
@@ -31,7 +31,7 @@
  */
 
 #ifdef HAVE_CONFIG_H
-#include <autocfg.h>
+#include <config.h>
 #endif
 
 #define SablotAsExport
@@ -44,12 +44,19 @@
 #include "parser.h"
 
 #include "guard.h"      // GP: clean
-#include "sdom.h"
 #include "numbering.h"
 #include "utf8.h"
 #include "domprovider.h"
 #include "uri.h"
 
+#ifdef SABLOT_DEBUGGER
+#include "debugger.h"
+#endif
+
+#ifdef ENABLE_DOM
+#include "sdom.h"
+#endif
+
 /*****************************************************************
     Instruction tables
 *****************************************************************/
@@ -334,8 +341,10 @@ Vertex::~Vertex()
   printf("Vertex destructor >%x<\n",this);
 #endif
     // call DOM dispose callback if set
-    if (SDOM_getDisposeCallback())
-        (*(SDOM_getDisposeCallback()))(this);
+#ifdef ENABLE_DOM
+  if (SDOM_getDisposeCallback())
+    (*(SDOM_getDisposeCallback()))(this);
+#endif
 };
 
 eFlag Vertex::execute(Sit S, Context *c, Bool resolvingGlobals)
@@ -362,10 +371,18 @@ eFlag Vertex::endCopy(Sit S, OutputterObj& out)
 
 eFlag Vertex::copy(Sit S, OutputterObj& out)
 {
-    S.setCurrSAXLine(lineno); //for SAX copying - be careful
-    E( startCopy(S, out) );
-    E( endCopy(S, out) );
-    return OK;
+  S.setCurrSAXLine(lineno); //for SAX copying - be careful
+
+  OutputterObj *aux;
+  E( startDocument(S, aux) );
+
+  OutputterObj &useOut = aux ? *aux : out;
+  E( startCopy(S, useOut) );
+  E( endCopy(S, useOut) );
+
+  E( finishDocument(S) );
+
+  return OK;
 }
 
 void Vertex::speak(DStr &ret, SpeakMode mode)
@@ -377,7 +394,7 @@ void Vertex::setParent(Vertex *v)
     parent = v;
 }
 
-Vertex *Vertex::getPreviousSibling()
+Vertex* Vertex::getPreviousSibling()
 {
     if (!parent || !isElement(parent))
         return NULL;
@@ -386,7 +403,7 @@ Vertex *Vertex::getPreviousSibling()
 	return toE(parent) -> contents[ordinal - 1];
 }
 
-Vertex *Vertex::getNextSibling()
+Vertex* Vertex::getNextSibling()
 {
     if (!parent || !isElement(parent))
         return NULL;
@@ -400,7 +417,7 @@ const QName& Vertex::getName() const
     return owner.getTheEmptyQName();
 }
 
-Tree &Vertex::getOwner() const
+Tree& Vertex::getOwner() const
 {    
     return owner;
 }
@@ -429,13 +446,13 @@ void Vertex::makeStamps(int& stamp_)
 
 eFlag Vertex::getMatchingList(Sit S, Expression& match, Context& result)
 {
-    Context aux(NULL);
-    aux.set(this);
-    Bool yes;
-    E( match.matchesPattern(S, &aux, yes) );
-    if (yes)
-        result.append(this);
-	return OK;
+  Context aux(NULL);
+  aux.set(this);
+  Bool yes;
+  E( match.matchesPattern(S, &aux, yes) );
+  if (yes)
+    result.append(this);
+  return OK;
 }
 
 int Vertex::getImportPrecedence()
@@ -443,14 +460,35 @@ int Vertex::getImportPrecedence()
     return NZ( getSubtreeInfo() ) -> getStructure() -> getImportPrecedence(); 
 }
 
+eFlag Vertex::startDocument(Sit S, OutputterObj*& out)
+{
+  Processor *proc;
+  if (outputDocument && (proc = S.getProcessor()))
+    {
+      E( proc -> startDocument(S, outputDocument) );
+      out = NZ( outputDocument -> getOutputter() );
+    }
+  else
+    out = NULL;
+  return OK;
+}
 
+eFlag Vertex::finishDocument(Sit S)
+{
+  Processor *proc;
+  if (outputDocument && (proc = S.getProcessor()))
+    {
+      E( proc -> finishDocument(S, outputDocument, FALSE) );
+    }
+  return OK;
+}
 
 /****************************************
 V e r t e x L i s t
 ****************************************/
 
 VertexList::VertexList(int logBlockSize_ /*=LIST_SIZE_SMALL*/)
-: SList<Vertex*>(logBlockSize_)
+  : SList<Vertex*>(logBlockSize_)
 {
 }
 
@@ -572,9 +610,9 @@ void VertexList::makeStamps(int& stamp_)
 
 eFlag VertexList::getMatchingList(Sit S, Expression& match, Context& result)
 {
-    for (int i = 0; i < number(); i++)
-        E( (*this)[i] -> getMatchingList(S, match, result) );
-	return OK;
+  for (int i = 0; i < number(); i++)
+    E( (*this)[i] -> getMatchingList(S, match, result) );
+  return OK;
 }
 
 /****************************************
@@ -652,10 +690,10 @@ RootNode::~RootNode()
 
 eFlag RootNode::execute(Sit S, Context *c, Bool resolvingGlobals)
 {
-    //! process prolog
-    //
-    E( Daddy::execute(S, c, resolvingGlobals) );
-    return OK;
+  //! process prolog
+  //
+  E( Daddy::execute(S, c, resolvingGlobals) );
+  return OK;
 };
 
 eFlag RootNode::checkChildren(Sit S)
@@ -691,10 +729,12 @@ eFlag RootNode::endCopy(Sit S, OutputterObj& out)
 
 eFlag RootNode::copy(Sit S, OutputterObj& out)
 {
-    E( startCopy(S, out) );
-    E( contents.copy(S, out) );
-    E( endCopy(S, out) );
-    return OK;
+  //RootNode shouldn't need to case for outputDocument
+
+  E( startCopy(S, out) );
+  E( contents.copy(S, out) );
+  E( endCopy(S, out) );
+  return OK;
 }
 
 eFlag RootNode::serialize(Sit S, OutputterObj &out)
@@ -717,9 +757,13 @@ Element::Element(Tree& owner_, QName& aqname, VTYPE avt /*= VT_ELEMENT_WF*/)
 {
     name = aqname;
     //origContext = NULL;
-    exNS = NULL;
-    extNS = NULL;
+    //exNS = NULL;
+    //extNS = NULL;
     attsNames = NULL;
+    preserveSpace = FALSE;
+#ifdef SABLOT_DEBUGGER
+    breakpoint = NULL;
+#endif
 };
 
 Element::~Element()
@@ -727,14 +771,6 @@ Element::~Element()
     // not doing freeall since it's all in the arena
     // namespaces.freeall(FALSE);
     // atts.freeall(FALSE);
-    if (exNS) {
-      exNS -> freeall(FALSE);
-      delete exNS;
-    }
-    if (extNS) {
-      extNS -> freeall(FALSE);
-      delete extNS;
-    }
     if (attsNames) {
       attsNames -> freeall(FALSE);
       delete attsNames;
@@ -743,27 +779,46 @@ Element::~Element()
     atts.destructMembers();
 };
 
+int Element::strip() 
+{
+  if (!preserveSpace)
+    return contents.strip();
+  else
+    return 0;
+}
+
 eFlag Element::execute(Sit S, Context *c, Bool resolvingGlobals)
 {
+#ifdef SABLOT_DEBUGGER
+  if (debugger) E( debugger -> breakOnElement(S, this, c) );
+#endif
+
     EQName expandedName;
-
-    //setOrigContext(c);
-
     owner.expandQ(name, expandedName);
-    E( NZ(S.getProcessor()) -> outputter() -> eventElementStart(S, 
-        expandedName) );
-    if (exNS) {
-      E( NZ(S.getProcessor()) -> outputter() -> eventExcludeNS(S,
-				       *excludedNS(FALSE)) );
-    }
-    E( namespaces.execute(S, c, resolvingGlobals) );
+    OutputterObj * out = NZ(S.getProcessor()) -> outputter();
+
+    //name aliasing
+    Bool aliased = false;
+    NZ(S.getProcessor()) -> getAliasedName(expandedName, aliased);
+
+    //set subtree for exclusions
+    E( out -> eventElementStart(S, expandedName) );
+    //output namespaces (and extra aliased namespace if needed)
+    E( namespaces.executeSkip(S, c, resolvingGlobals, expandedName, aliased) );
+    if (aliased)
+      E( out -> eventNamespace(S, expandedName.getPrefix(), 
+			       expandedName.getUri() ) )
+    //attribute sets
     if (attSetNames(FALSE))
       E( executeAttributeSets(S, c, resolvingGlobals) );
+    //plain attributes
     E( atts.execute(S, c, resolvingGlobals) );
+    //content
     E( Daddy::execute(S, c, resolvingGlobals) );
     removeBindings(S);
-    E( NZ(S.getProcessor()) -> outputter() -> eventElementEnd(S, 
-        expandedName));
+    //end tag
+    E( out -> eventElementEnd(S, expandedName));
+
     return OK;
 }
 
@@ -863,28 +918,36 @@ void Element::removeBindings(Sit S)
 
 eFlag Element::startCopy(Sit S, OutputterObj& out)
 {
-    EQName expandedName;
-    owner.expandQ(name, expandedName);
-    E( out.eventElementStart(S, expandedName ));
-    E( namespaces.copy(S, out) );
-    return OK;
+  EQName expandedName;
+  owner.expandQ(name, expandedName);
+  E( out.eventElementStart(S, expandedName ));
+  E( namespaces.copy(S, out) );
+  return OK;
 }
 
 eFlag Element::endCopy(Sit S, OutputterObj& out)
 {
-    EQName expandedName;
-    owner.expandQ(name, expandedName);
-    E( out.eventElementEnd(S, expandedName ));
-    return OK;
+  EQName expandedName;
+  owner.expandQ(name, expandedName);
+  E( out.eventElementEnd(S, expandedName ));
+  return OK;
 }
 
 eFlag Element::copy(Sit S, OutputterObj& out)
 {
+  OutputterObj *aux;
+  E( startDocument(S, aux) );
+
+  OutputterObj &useOut = aux ? *aux : out;
+
   S.setCurrSAXLine(lineno); //for SAX copying - be careful
-  E( startCopy(S, out) );
-  E( atts.copy(S, out) );
-  E( contents.copy(S, out) );
-  E( endCopy(S, out) );
+  E( startCopy(S, useOut) );
+  E( atts.copy(S, useOut) );
+  E( contents.copy(S, useOut) );
+  E( endCopy(S, useOut) );
+
+  E( finishDocument(S) );
+
   return OK;
 };
 
@@ -986,17 +1049,7 @@ eFlag Element::getMatchingList(Sit S, Expression& match, Context& result)
 	return OK;
 }
 
-UriList *Element::excludedNS(Bool canCreate) {
-  if (!exNS && canCreate) exNS = new UriList();
-  return exNS;
-}
-
-UriList *Element::extensionNS(Bool canCreate) {
-  if (!extNS && canCreate) extNS = new UriList();
-  return extNS;
-}
-
-QNameList *Element::attSetNames(Bool canCreate) {
+QNameList* Element::attSetNames(Bool canCreate) {
   if (!attsNames && canCreate) attsNames = new QNameList;
   return attsNames;
 }
@@ -1198,7 +1251,7 @@ eFlag Attribute::serialize(Sit S, OutputterObj &out)
 A t t L i s t
 ****************************************/
 
-Attribute *AttList::find(XSL_ATT what)
+Attribute* AttList::find(XSL_ATT what)
 {
     int i;
     Attribute *a;
@@ -1232,7 +1285,7 @@ int AttList::findNdx(const QName& attName)
     return -1;
 }
 
-Attribute *AttList::find(const QName& attName)
+Attribute* AttList::find(const QName& attName)
 {
     int ndx = findNdx(attName);
     // need to use a temporary variable
@@ -1251,13 +1304,17 @@ Attribute *AttList::find(const QName& attName)
 /*****************************************************************
     N m S p a c e
 *****************************************************************/
-
-NmSpace::NmSpace(Tree& owner_, Phrase _prefix, Phrase _uri)
+//NmSpace::NmSpace(Tree& owner_, Phrase _prefix, Phrase _uri, NsKind _kind)
+NmSpace::NmSpace(Tree& owner_, Phrase _prefix, Phrase _uri, 
+		 Bool _excluded, NsKind _kind)
 : Vertex(owner_, VT_NAMESPACE)
 {
     prefix = _prefix;
     uri = _uri;
     name.setLocal(prefix);
+    kind = _kind;
+    excluded = _excluded;
+    usageCount = 0;
 }
 
 NmSpace::~NmSpace()
@@ -1267,17 +1324,28 @@ NmSpace::~NmSpace()
 eFlag NmSpace::execute(Sit S, Context *c, Bool resolvingGlobals)
 {
     assert(parent);
-/*
-    made unnecessary by the shadowing inside OutputterObj:
-    if (isXSLElement(parent) || 
-        uri == proc -> stdPhrase(PHRASE_XSL_NAMESPACE) || 
-        uri == proc -> stdPhrase(PHRASE_XML_NAMESPACE))
-        return OK;
-*/
+
     E( NZ(S.getProcessor()) -> outputter() -> 
-        eventNamespace(S,
-            getOwner().dict().getKey(prefix), //was: proc -> aliasXL(prefix)),
-            getOwner().dict().getKey(uri)) );
+       eventNamespace(S,
+		      //was: proc -> aliasXL(prefix)),
+		      getOwner().dict().getKey(prefix), 
+		      getOwner().dict().getKey(uri),
+		      excluded
+		      ));
+    return OK;
+}
+
+eFlag NmSpace::executeSkip(Sit S, Context *c, Bool resolvingGlobals,
+			   EQName & exName, Bool aliased)
+{
+    assert(parent);
+
+    const Str & pStr = getOwner().dict().getKey(prefix);
+    const Str & uStr = getOwner().dict().getKey(uri);
+
+    if (!aliased || !(exName.getPrefix() == pStr))
+      E( NZ(S.getProcessor()) -> outputter() -> 
+	 eventNamespace(S, pStr, uStr, excluded) );
     return OK;
 }
 
@@ -1303,8 +1371,9 @@ eFlag NmSpace::value(Sit S, DStr &s, Context *c)
 eFlag NmSpace::startCopy(Sit S, OutputterObj& out)
 {
     E( out.eventNamespace(S,
-        getOwner().dict().getKey(prefix), 
-        getOwner().dict().getKey(uri)) );
+			  getOwner().dict().getKey(prefix), 
+			  getOwner().dict().getKey(uri),
+			  excluded) );
     return OK;
 }
 
@@ -1323,6 +1392,10 @@ eFlag NmSpace::serialize(Sit S, OutputterObj &out)
     return OK;
 }
 
+NsKind NmSpace::setKind(NsKind kind_) {
+    if (kind < kind_) kind = kind_;
+    return kind;
+}
 /*****************************************************************
     N S L i s t
 *****************************************************************/
@@ -1331,7 +1404,7 @@ NSList::~NSList()
 {
 }
 
-NmSpace *NSList::find(Phrase prefix) const
+NmSpace* NSList::find(Phrase prefix) const
 {
     int ndx = findNdx(prefix);
     return (ndx == -1) ? NULL : toNS((*this)[ndx]);
@@ -1340,9 +1413,9 @@ NmSpace *NSList::find(Phrase prefix) const
 int NSList::findNdx(Phrase prefix) const
 {
     int i;
-    for (i = 0; i < number(); i++)
-        if (toNS((*this)[i]) -> prefix == prefix)
-           return i;
+    for (i = 0; i < number(); i++) 
+      if (toNS((*this)[i]) -> prefix == prefix)
+	return i;
     return -1;
 }
 
@@ -1383,15 +1456,56 @@ appends all current namespace declarations as vertices to tree t
 'other' is the list where they are being appended in t
 */
 
-void NSList::giveCurrent(Sit S, NSList &other, Tree *t) const
+void NSList::giveCurrent(Sit S, NSList &other, Tree *t, int nscount) const
 {
-    const NmSpace *currNS;
-    for (int i = number() - 1; i >= 0; i--)
+    const NmSpace *currNS, *addedNS;
+    const int imax = number() - 1;
+    UriList &excludedNS = t -> getCurrentInfo() -> getExcludedNS();
+    for (int i = imax; i >= 0; i--)
     {
         currNS = toNS((*this)[i]);
-        if (!other.find(currNS -> prefix))
-            t -> appendVertex(S,
-                new(&(NZ(t) -> getArena())) NmSpace(*t, currNS -> prefix, currNS -> uri));
+        if (other.findNdx(currNS -> prefix) == -1) {
+	    addedNS = new(&(NZ(t) -> getArena())) 
+	      NmSpace(*t, currNS -> prefix, currNS -> uri, 
+		      excludedNS.findNdx(currNS->uri)!=-1 || currNS->excluded,
+		      i > (imax - nscount) ?  NSKIND_DECLARED : NSKIND_PARENT);
+            t -> appendVertex(S, toV(addedNS));
+	    // _JP_ ??? why not other.append(toV(addedNS));
+	};
+    }
+}
+
+void NSList::swallow(Sit S, NSList &other, Tree* srcTree, Tree *t)
+{
+  for (int i = 0; i < other.number(); i++)
+    {
+      NmSpace *currNS = toNS(other[i]);
+
+      Phrase p, u;
+      //translate between dictionaries
+      if (srcTree && srcTree != t)
+	{
+	  if (currNS -> prefix == UNDEF_PHRASE)
+	    {
+	      p = UNDEF_PHRASE;
+	    }
+	  else {
+	    const Str &aux1 = srcTree -> expand(currNS -> prefix);
+	    p = t -> unexpand(aux1);
+	  }
+
+	  const Str &aux2 = srcTree -> expand(currNS -> uri);
+	  u = t -> unexpand(aux2);
+	}
+      else
+	{
+	  assert(!srcTree);
+	  p = currNS -> prefix;
+	  u = currNS -> uri;
+	}
+
+      NmSpace *nm = new (&(NZ(t) -> getArena())) NmSpace(*t, p, u, NSKIND_PARENT);
+      append(nm);
     }
 }
 
@@ -1414,6 +1528,41 @@ void NSList::report(Sit S, MsgType type, MsgCode code,
     S.message(type, code, arg1, arg2);
 }
 
+void NSList::setPrefixKind(Phrase prefix_, NsKind kind_) const
+{
+    int ndx = findNdx(prefix_);
+    if (ndx != -1) {
+      toNS((*this)[ndx]) -> setKind(kind_);
+    };
+}
+
+void NSList::incPrefixUsage(Phrase prefix_) const
+{
+    int ndx = findNdx(prefix_);
+    if (ndx != -1) {
+      toNS((*this)[ndx]) -> usageCount++;
+    };
+}
+
+void NSList::decPrefixUsage(Phrase prefix_) const
+{
+    int ndx = findNdx(prefix_);
+    if (ndx != -1) {
+      toNS((*this)[ndx]) -> usageCount--;
+    };
+}
+
+eFlag NSList::executeSkip(Sit S, Context * c, Bool resolvingGlobals,
+			  EQName & exName, Bool aliased)
+{
+  int i;
+  
+  for (i=0; i<number(); i++)
+    E( toNS((*this)[i]) -> executeSkip(S, c, resolvingGlobals,
+				 exName, aliased) );
+  return OK;
+}
+
 /****************************************
 T e x t
 ****************************************/
@@ -1617,6 +1766,10 @@ eFlag XSLElement::execute(Sit S, Context *c, Bool resolvingGlobals)
     //e.g. the current() function needs it
     //setOrigContext(c);
 
+#ifdef SABLOT_DEBUGGER
+    if (debugger) E( debugger -> breakOnElement(S, this, c) );
+#endif
+
     switch(op)
     {
     case XSL_APPLY_TEMPLATES:
@@ -1677,7 +1830,7 @@ eFlag XSLElement::execute(Sit S, Context *c, Bool resolvingGlobals)
 
     case XSL_APPLY_IMPORTS:
     {
-	E( proc -> execApplyImports(S, c, resolvingGlobals) );
+	E( proc -> execApplyImports(S, c, subtree, resolvingGlobals) );
     }; break;
     
     case XSL_IF:
@@ -1705,10 +1858,13 @@ eFlag XSLElement::execute(Sit S, Context *c, Bool resolvingGlobals)
 	    //x -> setOrigContext(c); 
 	    if (x -> op == XSL_WHEN)
 	    {
-		E( NZ(NZ(x -> atts.find(XSLA_TEST)) 
-		      -> expr) -> trueFor(S, c,done)     );
-		if (done)
-		    E( x -> execute(S, c, resolvingGlobals) );
+	      a = NZ( x -> atts.find(XSLA_TEST) );
+	      Expression boolexpr(*this);
+	      E( NZ(a -> expr) -> eval(S, boolexpr, c) );
+	      done = boolexpr.tobool();
+
+	      if (done) 
+		E( x -> execute(S, c, resolvingGlobals) );
 	    }
 	    else
 	    {
@@ -1821,6 +1977,7 @@ eFlag XSLElement::execute(Sit S, Context *c, Bool resolvingGlobals)
 	if (nsuri != UNDEF_PHRASE && 
 	    (q.getPrefix() == UNDEF_PHRASE || prefix == "xmlns"))
 	  {
+
 	    Str pfx = S.getProcessor() -> getNextNSPrefix();
 	    q.setPrefix(owner.unexpand(pfx));
 	  }
@@ -1858,15 +2015,12 @@ eFlag XSLElement::execute(Sit S, Context *c, Bool resolvingGlobals)
     case XSL_TRANSFORM:
     {
 	GP( Context ) newc = c -> copy();
-
-	//uff, I hope, I'm right _ph_
-	if (exNS) {
-	  E( NZ(S.getProcessor()) -> outputter() -> eventExcludeNS(S,
-		         *excludedNS(FALSE)) );
-	}
+	
+	Processor *proc = NZ(S.getProcessor());
 	
 	// process globals
-	E( getOwner().resolveGlobals(S, c, S.getProcessor()) );
+	//_PH_ globals
+	//E( getOwner().resolveGlobals(S, c, S.getProcessor()) );
 	// process other top level elements
 	for (int i = 0; i < contents.number(); i++)
 	{
@@ -1880,7 +2034,12 @@ eFlag XSLElement::execute(Sit S, Context *c, Bool resolvingGlobals)
 		    // global vars have been executed before
 		case XSL_VARIABLE:
 		case XSL_PARAM:
-		    break;
+		  {
+		    //_PH_ globals
+		    QName temp;
+		    temp.empty();
+		    E( proc -> resolveGlobal(S, c, temp, toX(son)) );
+		  }; break;
 		default: E( son -> execute(S, c, resolvingGlobals) );
 		}
 	    }
@@ -1891,10 +2050,13 @@ eFlag XSLElement::execute(Sit S, Context *c, Bool resolvingGlobals)
 	}
 
 	proc -> vars -> startCall();
+
 	E( proc -> execute(S, (Vertex*) NULL, newc, resolvingGlobals ));
+
 	// newc has been deleted by now
 	newc.keep();
 	proc -> vars -> endCall();
+
     }; break;
     case XSL_VALUE_OF:
     {
@@ -1915,14 +2077,16 @@ eFlag XSLElement::execute(Sit S, Context *c, Bool resolvingGlobals)
     {
 	a = NZ( atts.find(XSLA_SELECT) );
 	Expression temp(*this);
-	E( NZ( a -> expr ) -> eval(S, temp, c) );
+	E( NZ( a -> expr ) -> eval(S, temp, c, resolvingGlobals) );
 	if (temp.type == EX_NODESET)
 	{
 	    const Context& ctxt = temp.tonodesetRef();
 	    int k, kLimit = ctxt.getSize();
+
 	    for (k = 0; k < kLimit; k++)
 		// assuming here the context holds physical vertices
-		E( toPhysical(ctxt[k]) -> copy(S, *(proc -> outputter())) );
+	      //E( toPhysical(ctxt[k]) -> copy(S, *(proc -> outputter())) );
+	      E( S.dom().copyNode(S, ctxt[k], *(proc -> outputter())) );
 	}
 	else
 	{
@@ -1933,14 +2097,15 @@ eFlag XSLElement::execute(Sit S, Context *c, Bool resolvingGlobals)
     }; break;
     case XSL_COPY:
     {
-	// assuming here the context holds physical vertices
-	Vertex *curr = toPhysical(c -> current());
-	E( curr -> startCopy(S, *(proc -> outputter())) );
-	//execute attribute sets
-	E( executeAttributeSets(S, c, resolvingGlobals) );
-
-	E( contents.execute(S, c, resolvingGlobals) );
-	E( curr -> endCopy(S, *(proc -> outputter())) );
+      NodeHandle curr = c -> current();
+      E( S.dom().startCopy(S, curr, *(proc -> outputter())) );
+      //E( curr -> startCopy(S, *(proc -> outputter())) );
+      //execute attribute sets
+      E( executeAttributeSets(S, c, resolvingGlobals) );
+      
+      E( contents.execute(S, c, resolvingGlobals) );
+      E( S.dom().endCopy(S, curr, *(proc -> outputter())) );
+      //E( curr -> endCopy(S, *(proc -> outputter())) );
     }; break;
     case XSL_FOR_EACH:
     {
@@ -2000,7 +2165,8 @@ eFlag XSLElement::execute(Sit S, Context *c, Bool resolvingGlobals)
 	{
 	    proc -> vars -> startNested();
 	    TreeConstructer *newTC;
-	    E( proc -> pushTreeConstructer(S, newTC, (*expr).setFragment()) );
+	    E( proc -> pushTreeConstructer(S, newTC, (*expr).setFragment(), 
+					   SAXOUTPUT_INT_PHYSICAL) );
 	    E( contents.execute(S, c, resolvingGlobals) );
 	    E( proc -> outputter() -> eventEndOutput(S) );
 	    E( proc -> popTreeConstructer(S, newTC) );
@@ -2045,12 +2211,12 @@ eFlag XSLElement::execute(Sit S, Context *c, Bool resolvingGlobals)
 	// if there's a 'select', use the expression it gives
 	a = atts.find(XSLA_SELECT);
 	if (a)
-	    expr.assign(NZ(a -> expr));
+	  expr.assign(NZ(a -> expr));
 
 	// otherwise, construct a new expression; it may be an empty string
 	// if this element has void content, or a result tree fragment otherwise 
 	else
-	{
+	  {
 	    expr = new Expression(*this);
 	    if (contents.isEmpty())
 		(*expr).setAtom((const char*)"");
@@ -2059,24 +2225,30 @@ eFlag XSLElement::execute(Sit S, Context *c, Bool resolvingGlobals)
 		// starting a nesting will make the current prebindings
 		// invisible
 		// don't do it if we are resolving globals
-		if (!resolvingGlobals)
-		    proc -> vars -> startNested();
-		TreeConstructer *newTC;
-		E( proc -> pushTreeConstructer(S, newTC, (*expr).setFragment()) );
-		// execute the inside to create the fragment
-		E( contents.execute(S, c, resolvingGlobals) );
-		E( proc -> outputter() -> eventEndOutput(S) );
-		E( proc -> popTreeConstructer(S, newTC) );
+	      
+	      proc -> pushInBinding(TRUE);
 
-		// end the shadowing of preexisting bindings
-		if (!resolvingGlobals)
-		    proc -> vars -> endNested();
+	      if (!resolvingGlobals)
+		proc -> vars -> startNested();
+	      TreeConstructer *newTC;
+	      E( proc -> pushTreeConstructer(S, newTC, (*expr).setFragment(),
+					     SAXOUTPUT_INT_PHYSICAL) );
+	      // execute the inside to create the fragment
+	      E( contents.execute(S, c, resolvingGlobals) );
+	      E( proc -> outputter() -> eventEndOutput(S) );
+	      E( proc -> popTreeConstructer(S, newTC) );
+	      
+	      // end the shadowing of preexisting bindings
+	      if (!resolvingGlobals)
+		proc -> vars -> endNested();
+	      
+	      proc -> popInBinding();
 	    };
 	}
 	//
 	// evaluate the expression
 	GP( Expression ) temp = new Expression(*this);  
-	E( (*expr).eval(S, *temp,c, resolvingGlobals) );
+	E( (*expr).eval(S, *temp, c, resolvingGlobals) );
 	(*temp).pTree = (*expr).pTree;
 	(*expr).pTree = NULL;
 
@@ -2164,7 +2336,7 @@ eFlag XSLElement::execute(Sit S, Context *c, Bool resolvingGlobals)
 	    E( xslNumberCount(S, level, 
 			      getAttExpr(XSLA_COUNT),
 			      getAttExpr(XSLA_FROM),
-			      toV(v), theList) );
+			      v, theList) );
 	}
 	Str format = "1",
 	    lang,
@@ -2231,7 +2403,7 @@ eFlag XSLElement::execute(Sit S, Context *c, Bool resolvingGlobals)
     case XSL_STRIP_SPACE:
     case XSL_PRESERVE_SPACE:
       //processed during parse
-      Warn1(S, W1_UNSUPP_XSL, xslOpNames[op]);
+      //Warn1(S, W1_UNSUPP_XSL, xslOpNames[op]);
 	break;
     default: 
 	Err1(S, E1_UNSUPP_XSL, xslOpNames[op]);
@@ -2244,7 +2416,7 @@ eFlag XSLElement::execute(Sit S, Context *c, Bool resolvingGlobals)
     return OK;
 }
 
-Expression *XSLElement::getAttExpr(XSL_ATT code)
+Expression* XSLElement::getAttExpr(XSL_ATT code)
 {
     Attribute *a = atts.find(code);
     return a ? (a -> expr) : NULL;
@@ -2252,12 +2424,11 @@ Expression *XSLElement::getAttExpr(XSL_ATT code)
 
 int XSLElement::strip()
 {
-    if (op != XSL_TEXT)
+    if (op != XSL_TEXT && !preserveSpace)
         return /* defs.strip() + */ contents.strip();
     else return 0;
 }
 
-
 /*................................................................
 findAttNdx()
     returns the index of given attribute in the attribute table
@@ -2514,7 +2685,7 @@ eFlag XSLElement::makeSortDefs(Sit S, SortDefList &sortDefs, Context *c)
             SortDef *newDef;
             E( toX(child) -> make1SortDef(S, newDef, c) );
             sortDefs.append(newDef);
-#           if !defined(HAVE_WCSXFRM) && !defined(_MSC_VER)
+#           if !defined(SAB_WCS_COLLATE)
             // warn that string sorting may be incorrect - 
             // sorting as ascii
             if (newDef -> asText)
@@ -2530,13 +2701,20 @@ eFlag XSLElement::makeSortDefs(Sit S, SortDefList &sortDefs, Context *c)
 /****************************************
 ExtensionElement
 ****************************************/
-const char *extNSUri[] = {
+const char* extNSUri[] = {
+  "http://exslt.org/functions",
   "http://www.exslt.org/functions",
+  "http://exslt.org/common",
   NULL
 };
 
-const char *exsltElements[] = {
-  "script",
+const char* exsltElementsFunctions[] = {
+  "script", 
+  NULL
+};
+
+const char* exsltElementsCommon[] = {
+  "document", 
   NULL
 };
 
@@ -2549,75 +2727,244 @@ Element(owner_, aqname, VT_EXT_ELEMENT_WF)
   lookupExt(owner, aqname, extns, op);
 };
 
+#ifdef ENABLE_JS
+eFlag ExtensionElement::executeEXSLTScript(Sit S, Context *c, 
+					   Bool resolvingGlobals)
+{
+#ifdef SABLOT_DEBUGGER
+  if (debugger) E( debugger -> breakOnElement(S, this, c) );
+#endif
+
+  QName attName;
+  //why no uri?????
+  //attName.setUri(getName().getUri());
+  attName.setLocal(owner.unexpand("language"));
+  Attribute *a = NZ( atts.find(attName) );
+  
+  QName lang;
+  E( setLogical(S, lang, a -> cont, false) );
+  //debug
+  EQName elang;
+  getOwner().expandQ(lang, elang);
+  //debug - end
+  if (lang.getUri() == UNDEF_PHRASE || elang.getUri() == extNSUri[0])
+    {
+      if (! (elang.getLocal() == (const char*)"javascript") &&
+	  ! (elang.getLocal() == (const char*)"ecmascript"))
+	return OK;
+    }
+  else
+    return OK;
+  
+
+  //if (!(a -> cont == (const char*) "javascript") && 
+  //    !(a -> cont == (const char*) "ecmascript"))
+  //  return OK;
+  
+  //implements-prefix
+  attName.setLocal(owner.unexpand("implements-prefix"));
+  a = NZ( atts.find(attName) );
+  
+  Phrase uri = owner.unexpand(a -> cont);
+  E( namespaces.resolve(S, uri, TRUE) );
+  Str suri = owner.expand(uri);
+  //check uri validity!!!
+  
+  //script
+  DStr script;
+  
+  attName.setLocal(owner.unexpand("src"));
+  a = atts.find(attName);
+  if (a)
+    {
+      DataLine src;
+      StrStrList &args = S.getProcessor() -> getArgList();
+      
+      const Str &base = S.findBaseURI(getSubtreeInfo() -> getBaseURI());
+      Str absolute;
+      makeAbsoluteURI(S, a -> cont, base, absolute);
+      
+      eFlag status = src.open(S, absolute, DLMODE_READ,  &args);
+      
+      if (status == OK)
+	{
+	  char buff[1024];
+	  int read;
+	  while ((read = src.get(S, buff, 1024)))
+	    {
+	      script.nadd(buff, read);
+	    }
+	}
+      else
+	{
+	  E( status );
+	}
+      E( src.close(S) );
+    }
+  else
+    {
+      for (int i = 0; i < contents.number(); i++) 
+	{
+	  Vertex *v = contents[i];
+	  if ( isText(v) ) 
+	    {
+	      script = script + toText(v) -> cont;
+	    }
+	}
+    }
+  E( NZ( S.getProcessor() ) -> evaluateJavaScript(S, suri, script) );
+  return OK;
+}
+#endif
+
+const char* _exsltDocAtts[] = {
+  "version", 
+  "encoding",
+  "omit-xml-declaration",
+  "standalone",
+  "doctype-public",
+  "doctype-system",
+  "indent",
+  "media-type",
+  NULL
+};
+
+XSL_ATT _exsltDocMapping[] = {
+  XSLA_VERSION,
+  XSLA_ENCODING,
+  XSLA_OMIT_XML_DECL,
+  XSLA_STANDALONE,
+  XSLA_DOCTYPE_PUBLIC,
+  XSLA_DOCTYPE_SYSTEM,
+  XSLA_INDENT,
+  XSLA_MEDIA_TYPE,
+  XSLA_NONE
+};
+
+XSL_ATT _lookupEXSLTDocumentAttr(const char* name)
+{
+  int idx = lookup(name, _exsltDocAtts);
+  return _exsltDocMapping[idx];
+}
+
+eFlag ExtensionElement::exsltDocGetOutputterDef(Sit S, Context *c,
+						OutputDefinition &def)
+{
+  int i, attsNumber = atts.number();
+  Attribute *theAtt;
+  for (i = 0; i < attsNumber; i++)
+    {
+      theAtt = toA(atts[i]);
+      Str attName = getOwner().expand(theAtt -> getName().getLocal());
+      Str attUri = getOwner().expand(theAtt -> getName().getUri());
+
+      //skip xslt attributes
+      if ( attUri == theXSLTNamespace )
+	continue;
+
+      if (attName == (const char*) "method")
+	{
+	  QName q;
+	  EQName eq;
+	  DStr val;
+	  theAtt -> value(S, val, c);
+	  E( setLogical(S, q, val, FALSE) );
+	  getOwner().expandQ(q, eq);
+	  E( def.setItemEQName(S, XSLA_METHOD, eq, theAtt, 0) );
+	}
+      else if (attName == (const char*) "cdata-section-elements")
+	{
+	  QName q;
+	  Bool someRemains;
+	  Str listPart;
+	  DStr val;
+	  theAtt -> value(S, val, c);
+	  char *p = (char*)val;
+	  do
+	    {
+	      someRemains = getWhDelimString(p, listPart);
+	      if (someRemains)
+		{
+		  E( setLogical(S, q, listPart, TRUE) );
+		  EQName expanded;
+		  getOwner().expandQ(q, expanded);
+		  E( def.setItemEQName(S, XSLA_CDATA_SECT_ELEMS,
+				       expanded, theAtt, 0) );
+		};
+	    }
+	  while (someRemains);
+	}
+      else if ( !(attName == (const char*) "href") )
+	{
+	  DStr val;
+	  theAtt -> value(S, val, c);
+	  XSL_ATT attType = _lookupEXSLTDocumentAttr(attName);
+	  if (attType == XSLA_NONE)
+	    Err1(S, ET_BAD_ATTR, (char*)attName);
+	  E( def.setItemStr(S, attType, val, theAtt, 0) );
+	  
+	};
+    };
+  return OK;
+}
+
+eFlag ExtensionElement::executeEXSLTDocument(Sit S, Context *c, 
+					     Bool resolvingGlobals)
+{
+  Processor *proc = NZ( S.getProcessor() );
+
+  //get output uri
+  QName attName;
+  attName.setLocal(owner.unexpand("href"));
+  Attribute *a = NZ( atts.find(attName) );
+  
+  DStr href;
+  E( a -> value(S, href, c) );
+
+  GP(OutputDefinition) def = new OutputDefinition();
+  E( exsltDocGetOutputterDef(S, c, *def) );
+
+  OutputDocument *doc;
+  E( proc -> getOutputDocument(S, href, doc, def.keep()) );
+  //we do not need to keep the definition, since document
+  //initialized the outputter with it (I hope)
+  
+  Bool handleOutputter = !proc -> isInBinding();
+  
+  if (handleOutputter)
+    {
+      S.message(MT_LOG, L2_SUBDOC_STARTED, href, "output");
+      E( proc -> startDocument(S, doc) );
+    }
+  else
+    {
+      S.message(MT_LOG, L2_SUBDOC_STARTED, href, "variable");
+      E( proc -> outputter() -> setDocumentForLevel(S, doc) );
+    }
+
+  E( contents.execute(S, c, resolvingGlobals) );
+
+  if (handleOutputter)
+    {
+      E( proc -> finishDocument(S, doc, TRUE) );
+    }
+
+  return OK;
+}
+
 eFlag ExtensionElement::execute(Sit S, Context *c, Bool resolvingGlobals) 
 {
   switch (op) {
   case EXTE_EXSLT_SCRIPT:
-    {
-      QName attName;
-      //why no uri?????
-      //attName.setUri(getName().getUri());
-      attName.setLocal(owner.unexpand("language"));
-      Attribute *a = atts.find(attName);
-      
-      if (!a) Err1(S, E_ATTR_MISSING, "language");
-      if (!(a -> cont == "javascript") && !(a -> cont == "ecmascript"))
-	return OK;
-
-      //implements-prefix
-      attName.setLocal(owner.unexpand("implements-prefix"));
-      a = atts.find(attName);
-      if (!a) Err1(S, E_ATTR_MISSING, "implements-prefix");
-
-      Phrase uri = owner.unexpand(a -> cont);
-      E( namespaces.resolve(S, uri, TRUE) );
-      Str suri = owner.expand(uri);
-      //check uri validity!!!
-
-      //script
-      DStr script;
-
-      attName.setLocal(owner.unexpand("src"));
-      a = atts.find(attName);
-      if (a)
-	{
-	  DataLine src;
-	  StrStrList &args = S.getProcessor() -> getArgList();
-	  
-	  const Str &base = S.findBaseURI(getSubtreeInfo() -> getBaseURI());
-	  Str absolute;
-	  makeAbsoluteURI(a -> cont, base, absolute);
-
-	  eFlag status = src.open(S, absolute, DLMODE_READ,  &args);
-
-	  if (status == OK)
-	    {
-	      char buff[1024];
-	      int read;
-	      while (read = src.get(S, buff, 1024))
-		{
-		  script.nadd(buff, read);
-		}
-	    }
-	  else
-	    {
-	      E( status );
-	    }
-	  E( src.close(S) );
-	}
-      else
-	{
-	  for (int i = 0; i < contents.number(); i++) 
-	    {
-	      Vertex *v = contents[i];
-	      if ( isText(v) ) 
-		{
-		  script = script + toText(v) -> cont;
-		}
-	    }
-	}
-      E( NZ( S.getProcessor() ) -> evaluateJavaScript(S, suri, script) );
-    }; break;
+#ifdef ENABLE_JS
+    E( executeEXSLTScript(S, c, resolvingGlobals) );
+#else
+    assert(!"JSExtension not built");
+#endif
+    break;
+  case EXTE_EXSLT_DOCUMENT:
+    E( executeEXSLTDocument(S, c, resolvingGlobals) );
+    break;
   case EXTE_UNKNOWN:
     {
       //do fallback
@@ -2640,7 +2987,8 @@ eFlag ExtensionElement::execute(Sit S, Context *c, Bool resolvingGlobals)
 eFlag ExtensionElement::checkChildren(Sit S) 
 {
   switch (op) {
-  case EXTNS_EXSLT:
+  case EXTNS_EXSLT_FUNCTIONS:
+  case EXTNS_EXSLT_FUNCTIONS_2:
     {
       int num = contents.number();
       for (int i = 0; i < num; i++) 
@@ -2658,8 +3006,32 @@ eFlag ExtensionElement::checkChildren(Sit S)
   return OK;
 }
 
+eFlag ExtensionElement::checkHasAttr(Sit S, const char *name)
+{
+  QName attName;
+  attName.setLocal(owner.unexpand(name));
+  Attribute *a = atts.find(attName);    
+  if (!a) Err1(S, E_ATTR_MISSING, name);
+  return OK;
+}
+
 eFlag ExtensionElement::checkAtts(Sit S) 
 {
+  switch (op) {
+  case EXTE_EXSLT_SCRIPT:
+    {
+      E( checkHasAttr(S, "implements-prefix") );
+      E( checkHasAttr(S, "language") );
+    }; break;
+  case EXTE_EXSLT_DOCUMENT:
+    {
+      E( checkHasAttr(S, "href") );
+      //all attributes are value templates
+      int num = atts.number();
+      for (int i = 0; i < num; i++)
+	E( toA(atts[i]) -> buildExpr(S, TRUE, EX_NONE) );
+    }; break;
+  }
   return OK;
 }
 
@@ -2670,11 +3042,21 @@ void ExtensionElement::lookupExt(Tree &tree, QName &name,
   Str mylocal = tree.expand(name.getLocal());
   extns_ = (ExtNamespace) lookup(myns, extNSUri);
   switch (extns_) {
-  case EXTNS_EXSLT:
+  case EXTNS_EXSLT_FUNCTIONS:
+  case EXTNS_EXSLT_FUNCTIONS_2:
     {
-      int i = lookup(mylocal, exsltElements);
-      if ( exsltElements[i] ) {
-	op_ = (ExtElement) (i + (int)EXTE_EXSLT);
+      int i = lookup(mylocal, exsltElementsFunctions);
+      if ( exsltElementsFunctions[i] ) {
+	op_ = (ExtElement) (i + (int)EXTE_EXSLT_FUNCTIONS);
+      } else {
+	op_ = EXTE_UNKNOWN;
+      }
+    }; break;
+  case EXTNS_EXSLT_COMMON:
+    {
+      int i = lookup(mylocal, exsltElementsCommon);
+      if ( exsltElementsCommon[i] ) {
+	op_ = (ExtElement) (i + (int)EXTE_EXSLT_COMMON);
       } else {
 	op_ = EXTE_UNKNOWN;
       }
@@ -2685,7 +3067,7 @@ void ExtensionElement::lookupExt(Tree &tree, QName &name,
     }
   }
   //now we have to check, whether the element is really supported
-#ifndef HAVE_JSAPI_H
+#ifndef ENABLE_JS
   if (op_ == EXTE_EXSLT_SCRIPT) op_ = EXTE_UNKNOWN;
 #endif
 }
@@ -2695,7 +3077,7 @@ Bool ExtensionElement::elementAvailable(Tree &t, QName &name)
   ExtNamespace ns;
   ExtElement op;
   lookupExt(t, name, ns, op);
-#ifndef HAVE_JSAPI_H
+#ifndef ENABLE_JS
   if (op == EXTE_EXSLT_SCRIPT) op = EXTE_UNKNOWN;
 #endif
   return op != EXTE_UNKNOWN;
